/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gtk/gtksignal.h>
#include <gtk/gtkmarshal.h>
#include <math.h>
#include <glade/glade.h>
#include "gribouy-bg.h"
#include "gribouy-triplet.h"

#define eps 0.00001

enum {
	LAST_SIGNAL
};

enum {
	PROP_0
};

static void gribouy_bg_init (GribouyBg * bg);
static void gribouy_bg_class_init (GribouyBgClass * klass);
static void gribouy_bg_set_property (GObject *object, guint property_id,
					const GValue *value,
					GParamSpec *pspec);
static void gribouy_bg_get_property (GObject *object, guint property_id,
					GValue *value, GParamSpec *pspec);
static void gribouy_bg_select (GribouyObject *obj, gboolean v);
static void gribouy_bg_update_path (GribouyBg *bg);
static void gribouy_bg_open_menu (GribouyObject *obj, GdkEvent *event);
static void gribouy_bg_editable_cb (GribouyBg *bg, guint action, GtkWidget *item);
static void gribouy_bg_properties_cb (GribouyBg *bg, guint action, GtkWidget *item);
static void gribouy_bg_transfer_cb (GribouyObject *bg, guint action, GtkWidget *item);
static void gribouy_bg_erase_cb (GribouyBg *bg, guint action, GtkWidget *item);
static gboolean bezier_exists (int *ax, int *ay, int n, double *bezier);
static gint gribouy_bg_event (GribouyBg *bg, GdkEvent *event, gpointer data);
static void gribouy_bg_move_image (GribouyObject *bg, gdouble x, gdouble y);

static GnomeCanvasPixbufClass *parent_class = NULL;

//static guint bg_signals[LAST_SIGNAL] = { 0 };
static void gribouy_bg_destroy (GtkObject *object);

GType
gribouy_bg_get_type (void)
{
	static GType gribouy_bg_type = 0;

	if (!gribouy_bg_type) {
		static const GTypeInfo gribouy_bg_info = {
			sizeof (GribouyBgClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_bg_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyBg),
			0,
			(GInstanceInitFunc) gribouy_bg_init,
			NULL
		};

		gribouy_bg_type = g_type_register_static (
				gribouy_object_get_type (),
				"GribouyBg", &gribouy_bg_info,
				(GTypeFlags)0);
	}

	return gribouy_bg_type;
}

static void 
gribouy_bg_class_init (GribouyBgClass * class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	GribouyObjectClass *obj_class = (GribouyObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_bg_set_property;
	g_object_class->get_property = gribouy_bg_get_property;

	object_class->destroy = gribouy_bg_destroy;
	obj_class->select = gribouy_bg_select;
	obj_class->open_menu = gribouy_bg_open_menu;
	obj_class->move_selection = gnome_canvas_item_move;
}

static void 
gribouy_bg_init (GribouyBg *bg)
{
}

char
gribouy_pixbuf_get_alpha (GdkPixbuf *pixbuf)
{
	guchar *pixels;

	g_return_if_fail (GDK_IS_PIXBUF (pixbuf));
	if (gdk_pixbuf_get_n_channels (pixbuf) != 4)
		return 0;

	if (gdk_pixbuf_get_width (pixbuf) == 0 ||
			gdk_pixbuf_get_height (pixbuf) == 0)
		return 0;

	pixels = gdk_pixbuf_get_pixels (pixbuf) + 3;
	return *pixels;
}

void
gribouy_pixbuf_set_alpha (GdkPixbuf *pixbuf, char alpha)
{
	guchar *pixels, *p;
	guint w, h;

	g_return_if_fail (GDK_IS_PIXBUF (pixbuf));
	if (gdk_pixbuf_get_n_channels (pixbuf) != 4)
		return;

	if (gdk_pixbuf_get_width (pixbuf) == 0 ||
			gdk_pixbuf_get_height (pixbuf) == 0)
		return;

	pixels = gdk_pixbuf_get_pixels (pixbuf);
	h = gdk_pixbuf_get_height (pixbuf);
	while (h--) {
		w = gdk_pixbuf_get_width (pixbuf);
		p = pixels;

		while (w--) {
			p[3] = alpha;
			p += 4;
		}
		pixels += gdk_pixbuf_get_rowstride (pixbuf);
	}
}

GribouyObject *
gribouy_bg_new (GribouyLetterWindow *window, gchar *filename,
		double x, double y)
{
	GribouyObject *bg, *triplet;
	GnomeCanvasItem *center;
	GnomeCanvasGroup *root;
	GdkPixbuf *pb, *pb1;
	GError *error = NULL;
	double w, h;

	root = gnome_canvas_root (GNOME_CANVAS (window->canvas));
	bg = GRIBOUY_OBJECT (gnome_canvas_item_new (
				root,
				gribouy_bg_get_type (),
				"cursor", GDK_ARROW,
				NULL));

	pb = gdk_pixbuf_new_from_file (filename, &error);
	pb1 = gdk_pixbuf_add_alpha (pb, FALSE, 0, 0, 0);
	w = gdk_pixbuf_get_width (pb1);
	h = gdk_pixbuf_get_height (pb1);
	gdk_pixbuf_unref (pb);
	gribouy_pixbuf_set_alpha (pb1, 128);

	center = gnome_canvas_item_new (GNOME_CANVAS_GROUP (bg),
			gnome_canvas_pixbuf_get_type (),
			"pixbuf", pb1,
			"x", x,
			"y", y,
			"width_set", TRUE,
			"height_set", TRUE,
			"width", w,
			"height", h,
			NULL);

	gdk_pixbuf_unref (pb1);

	g_object_set (G_OBJECT (bg),
			"center", center,
			"window", window,
			NULL);

	triplet = gribouy_triplet_new_with_tangents (bg, 0, x, y + h, x, y, x + w, y);
	g_signal_connect_swapped (G_OBJECT (triplet),
			"changed",
			G_CALLBACK (gribouy_bg_update_path),
			bg);

	gnome_canvas_item_lower_to_bottom (GNOME_CANVAS_ITEM (bg));
	return bg;
}

static void
gribouy_bg_select (GribouyObject *obj, gboolean v)
{
	GList *lst;

	lst = GNOME_CANVAS_GROUP (obj)->item_list->next;
	g_object_set (G_OBJECT (lst->data), "selected", v, NULL);
}

static void
gribouy_bg_update_path (GribouyBg *bg)
{
	GList *lst;
	GnomeCanvasItem *center;
	GribouyTriplet *triplet;
	double x1, y1, x2, y2, x, y;

	lst = GNOME_CANVAS_GROUP (bg)->item_list;
	center = lst->data;
	lst = lst->next;
	triplet = lst->data;
	g_object_get (G_OBJECT (triplet),
			"x", &x, "y", &y, NULL);
	g_object_get (G_OBJECT (triplet->ctrl1), "x", &x1, "y", &y1, NULL);
	g_object_get (G_OBJECT (triplet->ctrl2), "x", &x2, "y", &y2, NULL);
	g_object_set (G_OBJECT (center),
			"x", x,
			"width", x2,
			"y", y,
			"height", y1, NULL);
}

static void
gribouy_bg_move_image (GribouyObject *bg, gdouble x, gdouble y)
{
}

static void
gribouy_bg_get_property (GObject *object,
		guint property_id,
		GValue *value,
		GParamSpec *pspec)
{
	GnomeCanvasPixbuf *pixbuf;
	GribouyBg *bg;

	pixbuf = GNOME_CANVAS_PIXBUF (object);
	bg = GRIBOUY_BG (object);

	switch (property_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
	return;
}

static void
gribouy_bg_set_property (GObject *object,
			    guint param_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	GnomeCanvasPixbuf *pixbuf;
	GribouyBg *bg;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GRIBOUY_IS_BG (object));

	pixbuf = GNOME_CANVAS_PIXBUF (object);
	bg = GRIBOUY_BG (object);

	switch (param_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
		break;
	}
}

static void
gribouy_bg_transfer_cb (GribouyObject *bg, guint action, GtkWidget *item)
{
	GList *lst, *curves;
	GdkPixbuf *pb;
	int w, h;
	double W, H, x, y;
	GribouyObject *letter;
	double affine[6], affine1[6];;

	g_object_get (G_OBJECT (bg->center),
			"pixbuf", &pb,
			"width", &W,
			"x", &x, "y", &y,
			"height", &H,
			NULL);
	w = gdk_pixbuf_get_width (pb);
	h = gdk_pixbuf_get_height (pb);

	gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (bg), &x, &y);

	affine[0] = W / w;
	affine[3] = H / h;
	affine[1] = affine[2] = 0;
	affine[4] = x;
	affine[5] = y+H;

	g_print ("x = %f\n", x);
	lst = gribouy_bg_get_bounds ((GnomeCanvasPixbuf *)bg->center);
	curves = clear_bounds (lst);

	letter = GRIBOUY_OBJECT (bg->window->letter);
	gnome_canvas_item_i2w_affine (GNOME_CANVAS_ITEM (letter), affine1);
	art_affine_invert (affine1, affine1);
	art_affine_multiply (affine, affine, affine1);
	optimize_bounds (letter, curves, affine);
}
	
#if 0
	GList *lst, *tmp;
	GList *curves;
	GnomeCanvasPathDef *gpp;
	GribouyLetter *letter;
	GdkPixbuf *pb;
	int w, h, *p;
	double W, H;

	lst = gribouy_bg_get_bounds ((GnomeCanvasPixbuf *)bg);
	curves = clear_bounds (lst);

	g_object_get (G_OBJECT (bg), "pixbuf", &pb, NULL);
	w = gdk_pixbuf_get_width (pb);
	h = gdk_pixbuf_get_height (pb);

	g_object_get (G_OBJECT (bg),
			"width", &W,
			"height", &H,
			NULL);

	W /= w;
	H /= h;

//	gribouy_bg_transfer_new (bg, W, H);

	gpp = optimize_bounds (curves, W, H);

	tmp = curves;
	while (tmp) {
		lst = tmp->data;
		while (lst) {
			p = lst->data;
			g_free (p);
			lst = lst->next;
		}
		g_list_free (lst);
		tmp = tmp->next;
	}
	g_list_free (curves);

	letter = GRIBOUY_LETTER (GNOME_CANVAS_ITEM (bg)->parent->parent);

	gribouy_letter_add_gpp (letter, gpp);
	gnome_canvas_path_def_unref (gpp);
}
#endif

static void
gribouy_bg_properties_cb (GribouyBg *bg, guint action, GtkWidget *item)
{
	gribouy_bg_properties_new (bg);
}

static int *
create_path (int x1, int y1, int x2, int y2)
{
	int *retval;

	retval = g_new (int, 4);
	retval[0] = x1;
	retval[1] = y1;
	retval[2] = x2;
	retval[3] = y2;
	return retval;
}

static GList *
add_path (GList *lst, int *path)
{
	GList *l;
	int *p;

	l = lst;
	while (l) {
		p = l->data;
		if (p[0] == path[2] && p[1] == path[3] &&
				p[2] == path[0] && p[3] == path[1]) {
			g_free (p);
			if (l->prev)
				l->prev->next = l->next;
			if (l->next)
				l->next->prev = l->prev;

			if (lst == l)
				lst = lst->next;

			g_free (path);
			return lst;
		}
		l = l->next;
	}
	lst = g_list_append (lst, path);
	return lst;
}
	
GList *
gribouy_bg_get_bounds (GnomeCanvasPixbuf *bg)
{
	GList *lst;
	GdkPixbuf *pb;
	int i, j;
	guint w, h;
	guchar *p;
	int *path;

	g_return_val_if_fail (GNOME_IS_CANVAS_PIXBUF (bg), NULL);
	lst = NULL;
	g_object_get (G_OBJECT (bg), "pixbuf", &pb, NULL);
	w = gdk_pixbuf_get_width (pb);
	h = gdk_pixbuf_get_height (pb);
	p = gdk_pixbuf_get_pixels (pb);
	for (j = h - 1; j >= 0; j --) {
		for (i = 0; i < w; i ++) {
			if (p[0]*p[0]+p[1]*p[1]+p[2]*p[2] < 49152) {
				/* Dark !! We take the pixel */
				path = create_path (i, j, i, j+1);
				lst = add_path (lst, path);
				path = create_path (i, j+1, i+1, j+1);
				lst = add_path (lst, path);
				path = create_path (i+1, j+1, i+1, j);
				lst = add_path (lst, path);
				path = create_path (i+1, j, i, j);
				lst = add_path (lst, path);
			}
			p += 4;
		}
	}
	return lst;
}

/* We create a GList that contains others GLists. The second GLists are broken lines given by there points. All those lines are closed. */
GList *
clear_bounds (GList *lst)
{
	GList *retval, *l, *curve;
	int *A;
	int *p, x, y;

	retval = NULL;
	do {
		curve = NULL;
		p = lst->data;
		A = g_new (int, 2);
		x = A[0] = p[0];
		y = A[1] = p[1];
		do {
			l = lst;
			do {
				p = l->data;
				if (p[0] == A[0] && p[1] == A[1]) {
					lst = g_list_remove (lst, p);
					break;
				}
				l = l->next;
			} while (l);

			g_return_val_if_fail (l, NULL);

			curve = g_list_append (curve, A);
			A = g_new (int, 2);
			A[0] = p[2];
			A[1] = p[3];
			g_free (p);
		} while (A[0] != x || A[1] != y);
		curve = g_list_append (curve, A);
		retval = g_list_append (retval, curve);
	} while (lst);
	return retval;
}

void
optimize_bounds (GribouyObject *let, GList *curves, double affine[6])
{
	GList *cv, *tmp;
	int *p;
	int n, size;
	int *ax;
	int *ay;
	double bezier[9];
	ArtPoint pt, pt1, pt2;
	GribouyLetter *letter;

	letter = GRIBOUY_LETTER (let);
	size = 10;
	ax = g_new (int, size);
	ay = g_new (int, size);
	tmp = curves;
	while (tmp) {
		cv = tmp->data;		/* cv is a curve */
		p = cv->data;		/* p is a point of cv */
		ax[0] = p[0];
		ay[0] = p[1];

		pt.x = ax[0];
		pt.y = -ay[0];
		art_affine_point (&pt, &pt, affine);
		gribouy_letter_moveto (letter, &pt);
		cv = cv->next;
		p = cv->data;
		ax[1] = p[0];
		ay[1] = p[1];
		n = 2;
		cv = cv->next;
		do {
			if (n >= size) {
				size += 10;
				ax = g_realloc (ax, size * sizeof (int));
				ay = g_realloc (ay, size * sizeof (int));
			}
			p = cv->data;
			cv = cv->next;
			ax[n] = p[0];
			ay[n] = p[1];
			n ++;
			if (!bezier_exists (ax, ay, n, bezier)) {
				/* bezier has not changed */

				if (bezier[8] > 0) {
					pt.x = bezier[6];
					pt.y = -bezier[7];
					art_affine_point (&pt, &pt, affine);
					gribouy_letter_lineto (letter, &pt);
				}
				else {
					pt.x = bezier[2];
					pt.y = -bezier[3];
					art_affine_point (&pt, &pt, affine);
					pt1.x = bezier[4];
					pt1.y = -bezier[5];
					art_affine_point (&pt1, &pt1, affine);
					pt2.x = bezier[6];
					pt2.y = -bezier[7];
					art_affine_point (&pt2, &pt2, affine);
					gribouy_letter_curveto (letter,
							pt.x, pt.y,
							pt1.x, pt1.y,
							pt2.x, pt2.y);
				}
				ax[0] = ax[n-2];
				ay[0] = ay[n-2];
				ax[1] = ax[n-1];
				ay[1] = ay[n-1];
				n = 2;
			} /* else bezier contains the new curve */
		} while (cv);
		if (n > 2) {
			if (bezier[8] > 0) {
				pt.x = bezier[6];
				pt.y = -bezier[7];
				art_affine_point (&pt, &pt, affine);
				gribouy_letter_lineto (letter, &pt);
			}
			else {
				pt.x = bezier[2];
				pt.y = -bezier[3];
				art_affine_point (&pt, &pt, affine);
				pt1.x = bezier[4];
				pt1.y = -bezier[5];
				art_affine_point (&pt1, &pt1, affine);
				pt2.x = bezier[6];
				pt2.y = -bezier[7];
				art_affine_point (&pt2, &pt2, affine);
				gribouy_letter_curveto (letter,
						pt.x, pt.y,
						pt1.x, pt1.y,
						pt2.x, pt2.y);
			}
		}
		gribouy_letter_closepath (letter);
		tmp = tmp->next;
	}
	g_free (ax);
	g_free (ay);
}

#if 0
void
optimize_bounds (GribouyObject *let, GList *curves, double W, double H)
{
	GList *cv, *tmp;
	int *p;
	int n, size;
	int *ax;
	int *ay;
	double bezier[9];
	ArtPoint pt;
	GribouyLetter *letter;

	letter = GRIBOUY_LETTER (let);
	size = 10;
	ax = g_new (int, size);
	ay = g_new (int, size);
	tmp = curves;
	while (tmp) {
		cv = tmp->data;		/* cv is a curve */
		p = cv->data;		/* p is a point of cv */
		ax[0] = p[0];
		ay[0] = p[1];
		pt.x = ax[0]*W;
		pt.y = -ay[0]*H;
		gribouy_letter_moveto (letter, &pt);
		cv = cv->next;
		p = cv->data;
		ax[1] = p[0];
		ay[1] = p[1];
		n = 2;
		cv = cv->next;
		do {
			if (n >= size) {
				size += 10;
				ax = g_realloc (ax, size * sizeof (int));
				ay = g_realloc (ay, size * sizeof (int));
			}
			p = cv->data;
			cv = cv->next;
			ax[n] = p[0];
			ay[n] = p[1];
			n ++;
			if (!bezier_exists (ax, ay, n, bezier)) {
				/* bezier has not changed */

				if (bezier[8] > 0) {
					pt.x = bezier[6]*W;
					pt.y = -bezier[7]*H;
					gribouy_letter_lineto (letter, &pt);
				}
				else
					gribouy_letter_curveto (letter,
							bezier[2]*W, -bezier[3]*H,
							bezier[4]*W, -bezier[5]*H,
							bezier[6]*W, -bezier[7]*H);
				ax[0] = ax[n-2];
				ay[0] = ay[n-2];
				ax[1] = ax[n-1];
				ay[1] = ay[n-1];
				n = 2;
			} /* else bezier contains the new curve */
		} while (cv);
		if (n > 2) {
			if (bezier[8] > 0) {
				pt.x = bezier[6]*W;
				pt.y = -bezier[7]*H;
				gribouy_letter_lineto (letter, &pt);
			}
			else
				gribouy_letter_curveto (letter,
						bezier[2]*W, -bezier[3]*H,
						bezier[4]*W, -bezier[5]*H,
						bezier[6]*W, -bezier[7]*H);
		}
		gribouy_letter_closepath (letter);
		tmp = tmp->next;
	}
	g_free (ax);
	g_free (ay);
}
#endif

static double *
obtain_spline (double *t, int *y, int n)
{
	int i, j, k;
	double tt;
	double *b;
	double max, buf;
	double a[4][4];
	double p[7] = {0, 0, 0, 0, 0, 0, 0};

	/* System writing */

	b = g_new0(double, 4);
	for (i = 0; i < n; i ++) {
		tt = 1.;
		for (j = 6; j >= 0; j --) {
			p[j] += 2 * tt;
			tt *= t[i];
		}
		tt = y[i];
		for (j = 3; j >= 0; j --) {
			b[j] += 2 * tt;
			tt *= t[i];
		}
	}
	for (i = 0; i < 4; i ++)
		for (j = 0; j < 4; j ++)
			a[i][j] = p[i+j];

	/* System triangulation */
	for (i = 0; i < 4; i ++) {
		max = a[i][i];
		if (max < 0)
			max = -max;
		k = i;
		for (j = i + 1; j < 4; j ++) {
			if (a[j][i] > max) {
				max = a[j][i];
				k = j;
			}
			else if (-a[j][i] > max) {
				max = -a[j][i];
				k = j;
			}
		}
		if (max > eps) {
			if (k != i) {
				for (j = i; j < 4; j ++) {
					buf = a[i][j];
					a[i][j] = a[k][j];
					a[k][j] = buf;
				}
				buf = b[i];
				b[i] = b[k];
				b[k] = buf;
			}
			for (j = i + 1; j < 4; j ++) {
				buf = a[j][i] / a[i][i];
				for (k = i; k < 4; k ++)
					a[j][k] -= a[i][k] * buf;
				b[j] -= b[i] * buf;
			}
		}
	}

	/* System resolution */
	for (i = 3; i >= 0; i --) {
		buf = b[i];
		for (j = i + 1; j < 4; j ++)
			buf -= a[i][j] * b[j];
		if (a[i][i] > eps || a[i][i] < -eps)
			b[i] = buf / a[i][i];
		else if (buf <= eps && buf >= -eps)
			b[i] = 0.;
		else {
			g_free (b);
			return NULL;
		}
	}
	return b;
}

static gboolean
bezier_exists (int *ax, int *ay, int n, double *bezier)
{
	double *t, tt;
	int i, j;
	double *bx, *by;
	double dist, dist1, buf1, buf2;
	double a, b;
	int ix[4], iy[4];

	/* Initialization of (t_i) */
	t = g_new (double, n);	/* We can do better */
	t[0] = 0.;
	for (i = 1; i < n; i ++)
		t[i] = t[i-1]
			+ sqrt ((ax[i]-ax[i-1]) * (ax[i]-ax[i-1])
					+ (ay[i]-ay[i-1])*(ay[i]-ay[i-1]));
	for (i = 1; i < n; i ++)
		t[i] /= t[n-1];

/*	for (i = 0; i < n ; i ++)
		t[i] = (double)i / n;*/

	bx = obtain_spline (t, ax, n);
	by = obtain_spline (t, ay, n);

	if (!bx || !by) {
		if (bx)
			g_free (bx);
		if (by)
			g_free (by);
		return FALSE;
	}
	dist = 0;
	for (i = 0; i < n; i ++) {
		tt = 1;
		buf1 = buf2 = 0;
		for (j = 0; j < 4; j ++) {
			buf1 += bx[3-j] * tt;
			buf2 += by[3-j] * tt;
			tt *= t[i];
		}
		buf1 -= ax[i];
		buf2 -= ay[i];
		dist1 = buf1*buf1 + buf2*buf2;
		if (dist1 > dist)
			dist = dist1;
	}
	g_free (t);
	if (dist < 1) {
		bezier[0] = bx[3];
		bezier[1] = by[3];
		bezier[2] = (bx[2]/3 + bx[3]);
		bezier[3] = (by[2]/3 + by[3]);
		bezier[4] = ((bx[1] + 2*bx[2])/3 + bx[3]);
		bezier[5] = ((by[1] + 2*by[2])/3 + by[3]);
		bezier[6] = (bx[0] + bx[1] + bx[2] + bx[3]);
		bezier[7] = (by[0] + by[1] + by[2] + by[3]);

		g_free (bx);
		g_free (by);
		if (fabs (bezier[6] - bezier[0]) > 1) {
			a = (bezier[7] - bezier[1]) / (bezier[6] - bezier[0]);
			b = bezier[1] - a * bezier[0];
			if (fabs (a * bezier[2] + b - bezier[3]) < 1 &&
					fabs (a * bezier[4] + b - bezier[5]) < 1)
				bezier[8] = 1;
			else
				bezier[8] = -1;
		}
		else {
			if (fabs (bezier[0] - bezier[2]) < 1 &&
					fabs (bezier[0] - bezier[4]) < 1)
				bezier[8] = 1;
			else
				bezier[8] = -1;
		}
		return TRUE;
	}
	else
		return FALSE;
	
}

static void
gribouy_bg_destroy (GtkObject *object)
{
	g_return_if_fail (GRIBOUY_IS_BG (object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gribouy_bg_open_menu (GribouyObject *obj, GdkEvent *event)
{
	GtkItemFactory *factory;
	GtkAccelGroup *accel;
	GtkWidget *menu;
	GtkItemFactoryEntry menu_items[] = {
		{ "/_Properties",
			"<control>B",
			gribouy_bg_properties_cb, 1,
			"<StockItem>", GTK_STOCK_PROPERTIES },
		{ "/_Editable",
			NULL,
			gribouy_bg_editable_cb, 1, "<CheckItem>" },
		{ "/_Transfer",
			NULL,
			gribouy_bg_transfer_cb, 1,
			"<StockItem>", GTK_STOCK_JUMP_TO },
		{ "/Erase",
			NULL,
			gribouy_bg_erase_cb, 1,
			"<StockItem>", GTK_STOCK_DELETE }
	};

	accel = gtk_accel_group_new ();
	factory = gtk_item_factory_new (GTK_TYPE_MENU,
			"<bg>", accel);
	gtk_item_factory_create_items (factory, 4, menu_items, obj);
	menu = gtk_item_factory_get_widget (factory, "<bg>");

	if (menu) {
		gtk_widget_show (menu);
		gtk_menu_popup (GTK_MENU (menu),
				NULL, NULL, NULL, NULL,
				2, event->button.time);
	}
}

static void
gribouy_bg_erase_cb (GribouyBg *bg, guint action, GtkWidget *item)
{
	gtk_object_destroy (GTK_OBJECT (bg));
}

static void
gribouy_bg_editable_cb (GribouyBg *bg, guint action, GtkWidget *item)
{
}
