/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include "gribouy-curve.h"
#include "gribouy-art.h"
#include "gribouy-triplet.h"

enum {
	CHANGED,
	LAST_SIGNAL
};

enum {
	PROP_0,
};

static void gribouy_curve_init (GribouyCurve *curve);
static void gribouy_curve_class_init (GribouyCurveClass * class);
static void gribouy_curve_edit (GribouyObject *obj, gboolean v);
static void gribouy_curve_select (GribouyObject *obj, gboolean v);
static void gribouy_curve_move_selection (GribouyObject *obj,
		gdouble x, gdouble y);
static void gribouy_curve_open_menu (GribouyObject *obj, GdkEvent *event);
static void gribouy_curve_erase_cb (GribouyCurve *curve, guint action,
		GtkWidget *item);
static void gribouy_curve_update_path (GribouyCurve *curve);
static gint compare_triplets (GribouyTriplet *a, GribouyTriplet *b);
static gint gribouy_curve_wind (GribouyCurve *curve, ArtPoint *point);
static void gribouy_curve_insert_point (GribouyCurve *curve, ArtPoint *point);
static void gribouy_curve_insert_control (GribouyCurve *curve, ArtPoint *point);
static void gribouy_curve_insert_control_cb (GribouyCurve *curve, guint action,
		GtkWidget *widget);
static void gribouy_curve_set_property (GObject *object, guint property_id,
					const GValue *value,
					GParamSpec *pspec);
static void gribouy_curve_get_property (GObject *object, guint property_id,
					GValue *value, GParamSpec *pspec);

static void gribouy_curve_destroy (GtkObject *object);

static GnomeCanvasGroupClass *parent_class = NULL;
static guint curve_signals[LAST_SIGNAL] = {0};

GType
gribouy_curve_get_type (void)
{
	static GType gribouy_curve_type = 0;

	if (!gribouy_curve_type) {
		static const GTypeInfo gribouy_curve_info = {
			sizeof (GribouyCurveClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_curve_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyCurve),
			0,
			(GInstanceInitFunc) gribouy_curve_init,
			NULL
		};

		gribouy_curve_type = g_type_register_static (
				gribouy_object_get_type (),
				"GribouyCurve", &gribouy_curve_info,
				(GTypeFlags)0);
	}
	return gribouy_curve_type;
}

static void 
gribouy_curve_class_init (GribouyCurveClass *class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	GribouyObjectClass *obj_class = (GribouyObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_curve_set_property;
	g_object_class->get_property = gribouy_curve_get_property;

	object_class->destroy = gribouy_curve_destroy;
	obj_class->select = gribouy_curve_select;
	obj_class->edit = gribouy_curve_edit;
	obj_class->move_selection = gribouy_curve_move_selection;
	obj_class->open_menu = gribouy_curve_open_menu;
}

static void 
gribouy_curve_init (GribouyCurve *curve)
{
	curve->closed = 0;
}

static void
gribouy_curve_get_property (GObject *object,
			guint property_id,
			GValue *value,
			GParamSpec *pspec)
{
	GribouyCurve *curve;

	curve = GRIBOUY_CURVE (object);
	g_return_if_fail (GRIBOUY_IS_CURVE (curve));

	switch (property_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

GribouyObject *
gribouy_curve_get_triplet_index (GribouyCurve *curve, gint index)
{
	GList *lst;

	if (index == -1)
		return GNOME_CANVAS_GROUP (curve)->item_list_end->data;

	lst = GNOME_CANVAS_GROUP (curve)->item_list->next;
	while ((index-- > 0)) {
		lst = lst->next;
		if (!lst)
			lst = GNOME_CANVAS_GROUP (curve)->item_list->next;
	}
	return lst->data;
}

static void
gribouy_curve_set_property (GObject *object,
			    guint param_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	GribouyCurve *curve;
	ArtPoint *point;
	double *points;
	guint v;

	curve = GRIBOUY_CURVE (object);
	g_return_if_fail (GRIBOUY_IS_CURVE (curve));

	switch (param_id) {
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
			break;
	}
}

gboolean
gribouy_curve_is_closed (GribouyCurve *curve)
{
	return curve->closed;
}

static void
gribouy_curve_select (GribouyObject *obj, gboolean v)
{
	if (v)
		g_object_set (G_OBJECT (obj->center),
				"outline_color", "red", NULL);
	else
		g_object_set (G_OBJECT (obj->center),
				"outline_color", "black", NULL);
}

static void
gribouy_curve_edit (GribouyObject *obj, gboolean v)
{
	if (v) {
		// Hiding the object permits a total redrawing
		gnome_canvas_item_hide (obj->center);
		g_object_set (G_OBJECT (obj->center),
				"fill_color_rgba", 0,
				"fill_color", NULL,
				NULL);
		gnome_canvas_item_show (obj->center);
	}
	else
		g_object_set (G_OBJECT (obj->center),
				"fill_color_rgba", 0xddddff80,
				NULL);
}

GribouyObject * 
gribouy_curve_new (GribouyObject *letter)
{
	GribouyObject *curve;
	GnomeCanvasItem *center;
	GnomeCanvasPathDef *gpp;

	g_return_if_fail (GRIBOUY_IS_LETTER_WINDOW (letter->window));
	curve = GRIBOUY_OBJECT (gnome_canvas_item_new (
				GNOME_CANVAS_GROUP (letter),
				gribouy_curve_get_type (),
				"cursor", GDK_CROSSHAIR,
				NULL));
	gpp = gnome_canvas_path_def_new ();
	center = gnome_canvas_item_new (GNOME_CANVAS_GROUP (curve),
			gnome_canvas_bpath_get_type (),
			"bpath", gpp,
			"outline_color", "black",
			"width_pixels", 1,
			NULL);
	g_object_set (G_OBJECT (curve),
			"center", center,
			"window", letter->window,
			NULL);
	gnome_canvas_path_def_unref (gpp);
	return curve;
}
			
static gint
gribouy_curve_wind (GribouyCurve *curve, ArtPoint *point)
{
	ArtBpath *bpath, *bp;
	ArtVpath *vp;
	GnomeCanvasPathDef *gpp;
	ArtSVP *svp;
	int index;
	double pixels_per_unit;

	pixels_per_unit = GNOME_CANVAS_ITEM (curve)->canvas->pixels_per_unit;
	bpath = art_new (ArtBpath, 3);
	g_object_get (G_OBJECT (GRIBOUY_OBJECT (curve)->center), "bpath", &gpp, NULL);
	bp = gnome_canvas_path_def_bpath (gpp);
	vp = NULL;
	svp = NULL;
	index = 0;

	bpath[0].code = ART_MOVETO_OPEN;
	bpath[0].x3 = bp->x3;
	bpath[0].y3 = bp->y3;
	bp++;

	do {
		index++;

		/* we cut the curve in little parts to indentify
		 * where is our point... */
		switch (bp->code) {
		case ART_LINETO:
			bpath[1].code = ART_LINETO;
			bpath[1].x3 = bp->x3;
			bpath[1].y3 = bp->y3;
			bpath[2].code = ART_END;
			break;
		case ART_CURVETO:
			bpath[1].code = ART_CURVETO;
			bpath[1].x1 = bp->x1;
			bpath[1].y1 = bp->y1;
			bpath[1].x2 = bp->x2;
			bpath[1].y2 = bp->y2;
			bpath[1].x3 = bp->x3;
			bpath[1].y3 = bp->y3;
			bpath[2].code = ART_END;
			break;
		case ART_MOVETO:
		case ART_MOVETO_OPEN:
			bpath[0].code = ART_MOVETO_OPEN;
			bpath[0].x3 = bp->x3;
			bpath[0].y3 = bp->y3;
			bp++;
			goto next_one;
			break;
		default:
			g_warning ("it is strange, isn't it ?\n");
		}
		if (vp)
			art_free (vp);
		vp = (ArtVpath *) art_bez_path_to_vec (bpath, 0.1);

		if (svp) {
			art_free (svp);
		}
		svp = art_svp_vpath_stroke (vp,
					    ART_PATH_STROKE_JOIN_MITER,
					    ART_PATH_STROKE_CAP_BUTT,
					    7,
					    4,
					    0.1);
		if (art_svp_point_wind (svp, point->x, point->y)) {
			art_free (vp);
			art_free (bpath);
			art_free (svp);
			return index;
		}
		else {
			bpath[0].x3 = bpath[1].x3;
			bpath[0].y3 = bpath[1].y3;
			bp++;
		}
	next_one:
	} while (bp->code != ART_END);
	art_free (bpath);
	if (vp)
		art_free (vp);
	if (svp) {
		art_free (svp);
	}
	return -1;
}

void
gribouy_curve_moveto (GribouyCurve *curve, ArtPoint *point)
{
	GList *lst;
	GribouyObject *triplet;

	g_object_get (G_OBJECT (curve), "children", &lst, NULL);
	g_return_if_fail (lst == NULL);

	triplet = gribouy_triplet_new (GRIBOUY_OBJECT (curve), 0,
			point->x, point->y);
	g_signal_connect_swapped (G_OBJECT (triplet),
			"changed",
			G_CALLBACK (gribouy_curve_update_path),
			curve);
}

void
gribouy_curve_lineto (GribouyCurve *curve, ArtPoint *point)
{
	GribouyObject *triplet;
	GList *lst;
	double x, y;

	g_object_get (G_OBJECT (curve), "children", &lst, NULL);
	g_return_if_fail (lst);

	/* x and y are coord. of the first triplet of the curve */
	g_object_get (G_OBJECT (lst->data),
			"x", &x, "y", &y, NULL);

	if (fabs (x - point->x) >= 1 || fabs (y - point->y) >= 1) {
		triplet = gribouy_triplet_new (GRIBOUY_OBJECT (curve), -1,
				point->x, point->y);
		g_signal_connect_swapped (G_OBJECT (triplet),
				"changed",
				G_CALLBACK (gribouy_curve_update_path),
				curve);
	}
	else
		curve->closed = 1;
	gribouy_curve_update_path (curve);
}

void
gribouy_curve_curveto (GribouyCurve *curve, double x1, double y1, double x2, double y2, double x3, double y3)
{
	GribouyObject *triplet;
	GList *lst;
	gint index;
	double x, y;
	ArtPoint p;

	g_object_get (G_OBJECT (curve), "children", &lst, NULL);
	g_return_if_fail (lst);

	/* x and y are coord. of the first triplet of the curve */
	g_object_get (G_OBJECT (lst->data),
			"x", &x, "y", &y, NULL);

	if (fabs (x - x3) >= 1 || fabs (y - y3) >= 1) {
		triplet = gribouy_triplet_new (GRIBOUY_OBJECT (curve), -1,
				x3, y3);
		g_signal_connect_swapped (G_OBJECT (triplet),
				"changed",
				G_CALLBACK (gribouy_curve_update_path),
				curve);
	}
	else {
		curve->closed = 1;
		triplet = GRIBOUY_OBJECT (GNOME_CANVAS_GROUP (curve)->item_list->next->data);
	}

	g_object_get (G_OBJECT (triplet), "index", &index, NULL);
	p.x = x2;
	p.y = y2;
	g_object_set (G_OBJECT (triplet), "tangent1", &p, NULL);
	triplet = gribouy_curve_get_triplet_index (curve, index - 1);
	p.x = x1;
	p.y = y1;
	g_object_set (G_OBJECT (triplet), "tangent2", &p, NULL);

	gribouy_curve_update_path (curve);
}

void
gribouy_curve_append_point (GribouyCurve *curve, ArtPoint *point)
{
	GList *lst;
	GribouyObject *triplet;

	/* We are going to the first triplet */
	g_object_get (G_OBJECT (curve), "children", &lst, NULL);

	if (lst) {
		gribouy_curve_lineto (curve, point);
	}
	else {
		triplet = gribouy_triplet_new (GRIBOUY_OBJECT (curve), -1, point->x, point->y);
		g_signal_connect_swapped (G_OBJECT (triplet),
				"changed",
				G_CALLBACK (gribouy_curve_update_path),
				curve);
	}
}

static void
gribouy_curve_insert_point (GribouyCurve *curve, ArtPoint *point)
{
	gint index;
	GribouyObject *triplet;
	GribouyObject *triplet1;
	ArtPoint *t1, *t2;

	gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (curve)->parent,
			&point->x, &point->y);
	gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (curve),
			&point->x, &point->y);
	index = gribouy_curve_wind (curve, point);
	if (index >= 0) {
		triplet1 = gribouy_curve_get_triplet_index (GRIBOUY_CURVE (curve), index - 1);
		g_object_get (G_OBJECT (triplet1), "tangent2", &t1, NULL);
		triplet1 = gribouy_curve_get_triplet_index (GRIBOUY_CURVE (curve), index);
		g_object_get (G_OBJECT (triplet1), "tangent1", &t2, NULL);
		triplet = gribouy_triplet_new (GRIBOUY_OBJECT (curve), index,
				point->x, point->y);
		if (t1) {
			g_object_set (G_OBJECT (triplet), "tangent1", t1, NULL);
			art_free (t1);
		}
		if (t2) {
			g_object_set (G_OBJECT (triplet), "tangent2", t2, NULL);
			art_free (t2);
		}
		g_signal_connect_swapped (G_OBJECT (triplet),
				"changed",
				G_CALLBACK (gribouy_curve_update_path),
				curve);
		gribouy_curve_update_path (curve);
	}
}

static void
gribouy_curve_insert_control (GribouyCurve *curve, ArtPoint *point)
{
	gint index;
	GribouyObject *triplet1, *triplet2;
	ArtPoint *t1, *t2;

	gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (curve)->parent,
			&point->x, &point->y);

	gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (curve),
			&point->x, &point->y);
	index = gribouy_curve_wind (curve, point);
	if (index >= 0) {
		triplet1 = gribouy_curve_get_triplet_index (curve, index - 1);
		triplet2 = gribouy_curve_get_triplet_index (curve, index);
		g_object_get (G_OBJECT (triplet1), "tangent2", &t1, NULL);
		g_object_get (G_OBJECT (triplet2), "tangent1", &t2, NULL);
		if (!t1 && !t2) {
			g_object_set (G_OBJECT (triplet1), "tangent2", point, NULL);
			g_object_set (G_OBJECT (triplet2), "tangent1", point, NULL);
		}
		else if (t1 && t1->x == t2->x && t1->y == t2->y)
			g_object_set (G_OBJECT (triplet2), "tangent1", point, NULL);
		else
			g_print ("Cannot do anything...\n");

		gribouy_curve_update_path (curve);
	}
}

static void
gribouy_curve_destroy (GtkObject *object)
{
	g_return_if_fail (GRIBOUY_IS_CURVE (object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gribouy_curve_insert_control_cb (GribouyCurve *curve, guint action,
		GtkWidget *widget)
{
	ArtPoint p;
	GribouyObject *obj;

	obj = GRIBOUY_OBJECT (curve);
	p.x = obj->clk_x;
	p.y = obj->clk_y;
	switch (action) {
	case 1:
		gribouy_curve_insert_point (curve, &p);
		break;
	case 2:
		gribouy_curve_insert_control (curve, &p);
		break;
	}
}

static void
gribouy_curve_move_selection (GribouyObject *curve, gdouble x, gdouble y)
{
	GnomeCanvasGroup *letter;
	GList *lst;
	GribouyObject *c;

	letter = GNOME_CANVAS_GROUP ((GNOME_CANVAS_ITEM (curve)->parent));
	lst = letter->item_list->next;
	while (lst) {
		c = lst->data;
		lst = lst->next;
		if (c->selected)
			gnome_canvas_item_move (GNOME_CANVAS_ITEM (c),
					x, y);
	}
}

static void
gribouy_curve_open_menu (GribouyObject *obj, GdkEvent *event)
{
	GtkItemFactory *factory;
	GtkAccelGroup *accel;
	GtkWidget *menu;
	GtkItemFactoryEntry menu_items[] = {
		{ "/Add _point",
			NULL,
			gribouy_curve_insert_control_cb, 1,
			"<StockItem>", GTK_STOCK_ADD },
		{ "/Add _control",
			NULL,
			gribouy_curve_insert_control_cb, 2,
			"<StockItem>", GTK_STOCK_REFRESH },
		{ "/_Erase",
			NULL,
			gribouy_curve_erase_cb, 1,
			"<StockItem>", GTK_STOCK_CLEAR },
	};

	accel = gtk_accel_group_new ();
	factory = gtk_item_factory_new (GTK_TYPE_MENU,
			"<control>", accel);
	gtk_item_factory_create_items (factory, 3, menu_items, obj);
	menu = gtk_item_factory_get_widget (factory, "<control>");

	if (menu) {
		gtk_widget_show (menu);
		gtk_menu_popup (GTK_MENU (menu),
				NULL, NULL, NULL, NULL,
				2, event->button.time);
	}
}

static gint
compare_triplets (GribouyTriplet *a, GribouyTriplet *b)
{
	return a->index - b->index;
}

static void
gribouy_curve_update_path (GribouyCurve *curve)
{
	GList *lst, *lst1;
	GnomeCanvasItem *center;
	GnomeCanvasPathDef *gpp;
	ArtPoint *p1, *p2, *p3, *p4;
	double x, y, xstart, ystart;

	lst = GNOME_CANVAS_GROUP (curve)->item_list;
	center = lst->data;
	gpp = gnome_canvas_path_def_new ();

	/* We reorder the triplets upon the indexes' order */
	lst1 = g_list_sort (lst->next, (GCompareFunc) compare_triplets);
	lst->next = lst1;
	lst1->prev = lst;
	lst1 = GNOME_CANVAS_GROUP (curve)->item_list_end;
	while (lst1->next)
		lst1 = lst1->next;
	GNOME_CANVAS_GROUP (curve)->item_list_end = lst1;

	lst = lst->next;
	if (lst) {
		g_object_get (G_OBJECT (lst->data),
				"x", &xstart, "y", &ystart,
				"tangent1", &p4,
				"tangent2", &p3, NULL);
		gnome_canvas_path_def_moveto (gpp, xstart, ystart);
		lst = lst->next;

		while (lst) {
			g_object_get (G_OBJECT (lst->data),
					"x", &x, "y", &y,
					"tangent1", &p1,
					"tangent2", &p2, NULL);
			if (p1 && p3) {
				gnome_canvas_path_def_curveto (gpp, p3->x, p3->y, p1->x, p1->y, x, y);
				art_free (p1);
				art_free (p3);
			}
			else {
				if (p1) {
					gribouy_triplet_destroy_tangent1 (GRIBOUY_TRIPLET (lst->data));
					art_free (p1);
				}
				else if (p3) {
					gribouy_triplet_destroy_tangent2 (GRIBOUY_TRIPLET (lst->prev->data));
					art_free (p3);
				}
				gnome_canvas_path_def_lineto (gpp, x, y);
			}
			p3 = p2;
			lst = lst->next;
		}
		if (curve->closed) {
			if (p3 && p4) {
				gnome_canvas_path_def_curveto (gpp, p3->x, p3->y,
						p4->x, p4->y, xstart, ystart);
				gnome_canvas_path_def_closepath_current (gpp);
				art_free (p4);
				art_free (p3);
			}
			else {
				if (p3) {
					gribouy_triplet_destroy_tangent2 (GRIBOUY_TRIPLET (GNOME_CANVAS_GROUP (curve)->item_list_end->data));
					art_free (p3);
				}
				if (p4) {
					gribouy_triplet_destroy_tangent1 (GRIBOUY_TRIPLET (GNOME_CANVAS_GROUP (curve)->item_list->next->data));
					art_free (p4);
				}
				gnome_canvas_path_def_lineto (gpp, xstart, ystart);
				gnome_canvas_path_def_closepath_current (gpp);
			}
		}
	}
	g_object_set (G_OBJECT (center),
			"bpath", gpp, NULL);
	gnome_canvas_path_def_unref (gpp);
	g_signal_emit_by_name (G_OBJECT (curve), "changed", 0);
}

static void
gribouy_curve_erase_cb (GribouyCurve *curve, guint action,
		GtkWidget *item)
{
	GList *lst;
	GnomeCanvasGroup *letter;
	GribouyObject *crv;

	letter = GNOME_CANVAS_GROUP (GNOME_CANVAS_ITEM (curve)->parent);
	lst = letter->item_list->next;
	while (lst) {
		crv = lst->data;
		lst = lst->next;
		if (crv->selected)
			gtk_object_destroy (GTK_OBJECT (crv));
	}
	lst = letter->item_list->next;
	if (lst)
		g_signal_emit_by_name (G_OBJECT (lst->data), "changed", 0);
}
