/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "gribouy-dialog.h"

static void gribouy_grid_dialog_y_value_changed (GtkSpinButton *entry, GObject *grid);
static void gribouy_grid_dialog_x_value_changed (GtkSpinButton *entry, GObject *grid);
static void on_contrast_changed (GtkWidget *ruler, GnomeCanvasPixbuf *img);
static void gribouy_bg_transfer_apply (GladeXML *xml, GtkWidget *button);
static void gribouy_bg_transfer_ok (GladeXML *xml, GtkWidget *button);
static void on_alpha_changed (GtkWidget *widget, GribouyObject *bg);
static void gribouy_grid_dialog_visible_toggled (GtkCheckButton *entry,
		GObject *grid);
static void gribouy_grid_dialog_magnetic_toggled (GtkCheckButton *entry,
		GObject *grid);

GtkWidget *
gribouy_bg_properties_new (GribouyObject *bg)
{
	GladeXML *xml;
	GtkWidget *dialog, *widget;
	GdkPixbuf *pb;
	guchar alpha;
	GtkAdjustment *adj;

	g_object_get (G_OBJECT (bg->center), "pixbuf", &pb, NULL);
	alpha = gribouy_pixbuf_get_alpha (pb);
	xml = glade_xml_new (GLADE_FILE1, "bg_dialog", NULL);
	if (!xml)
		xml = glade_xml_new (GLADE_FILE, "bg_dialog", NULL);
	dialog = glade_xml_get_widget (xml, "bg_dialog");
	widget = glade_xml_get_widget (xml, "alpha_scale");
	adj = gtk_range_get_adjustment (GTK_RANGE (widget));
	gtk_adjustment_set_value (adj, (gfloat)alpha);

	g_signal_connect (G_OBJECT (widget), "value_changed",
			G_CALLBACK (on_alpha_changed), bg);
	widget = glade_xml_get_widget (xml, "okbutton3");
	g_signal_connect_swapped (GTK_OBJECT (widget), "clicked",
			G_CALLBACK (gtk_widget_destroy), dialog);
	gtk_widget_show (dialog);
	return dialog;
}

static void
on_alpha_changed (GtkWidget *widget, GribouyObject *bg)
{
	GdkPixbuf *pb;
	GtkAdjustment *adj;
	char val;

	g_return_if_fail (GRIBOUY_IS_BG (bg));
	g_object_get (G_OBJECT (bg->center), "pixbuf", &pb, NULL);
	adj = gtk_range_get_adjustment (GTK_RANGE (widget));
	val = adj->value;
	gribouy_pixbuf_set_alpha (pb, val);
	gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (bg->center));
}

GtkWidget *
gribouy_bg_transfer_new (GribouyObject *bg)
{
	GladeXML *xml;
	GtkWidget *dialog, *button, *ruler;
	GnomeCanvas *canvas;
	GnomeCanvasItem *item;
	GdkPixbuf *pb;
	GribouyLetter *letter;

	letter = GRIBOUY_LETTER (GNOME_CANVAS_ITEM (bg)->parent->parent);
	g_object_get (G_OBJECT (bg->center), "pixbuf", &pb, NULL);
	xml = glade_xml_new (GLADE_FILE1, "bg_transfer", NULL);
	if (!xml)
		xml = glade_xml_new (GLADE_FILE, "bg_transfer", NULL);

	dialog = glade_xml_get_widget (xml, "bg_transfer");
	canvas = GNOME_CANVAS (glade_xml_get_widget (xml, "canvas1"));
	item = gnome_canvas_item_new (gnome_canvas_root (canvas),
			gnome_canvas_pixbuf_get_type (),
			"pixbuf", pb,
			"x", 0.,
			"y", 0.,
			NULL);
	g_object_set_data (G_OBJECT (item), "orig", pb);
	ruler = glade_xml_get_widget (xml, "hscale1");
	on_contrast_changed (ruler, GNOME_CANVAS_PIXBUF (item));
	g_signal_connect (G_OBJECT (ruler), "value_changed",
			G_CALLBACK (on_contrast_changed), item);

	button = glade_xml_get_widget (xml, "cancelbutton3");
	g_signal_connect_swapped (G_OBJECT (button), "clicked",
			G_CALLBACK (gtk_widget_destroy), dialog);

	button = glade_xml_get_widget (xml, "applybutton2");
	g_signal_connect_swapped (G_OBJECT (button), "clicked",
			G_CALLBACK (gribouy_bg_transfer_apply), xml);

	button = glade_xml_get_widget (xml, "okbutton4");
	g_object_set_data (G_OBJECT (button), "letter", letter);
	g_signal_connect_swapped (G_OBJECT (button), "clicked",
			G_CALLBACK (gribouy_bg_transfer_ok), xml);

	return dialog;
}

static void
on_contrast_changed (GtkWidget *ruler, GnomeCanvasPixbuf *img)
{
	GdkPixbuf *pb, *orig;
	GtkAdjustment *adj;
	int val;
	guint w, h;
	guchar *p;
	int i, j;

	adj = gtk_range_get_adjustment (GTK_RANGE (ruler));
	val = adj->value * 51 / 20;
	val *= val;
	orig = GDK_PIXBUF (g_object_get_data (G_OBJECT (img), "orig"));
	pb = gdk_pixbuf_copy (orig);
	w = gdk_pixbuf_get_width (pb);
	h = gdk_pixbuf_get_height (pb);
	p = gdk_pixbuf_get_pixels (pb);
	for (j = 0; j < h; j ++) {
		for (i = 0; i < w; i ++) {
			if ((int)(p[0]*p[0]+p[1]*p[1]+p[2]*p[2]) < val)
				p[0] = p[1] = p[2] = 0;
			else
				p[0] = p[1] = p[2] = 255;
			p += 4;
		}
	}
			
	g_object_set (G_OBJECT (img), "pixbuf", pb, NULL);
	g_object_unref (pb);
	gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (img));
}

static void
gribouy_bg_transfer_apply (GladeXML *xml, GtkWidget *button)
{
	GList *lst, *curves, *tmp;
	GnomeCanvas *canvas;
	int *p;
	GnomeCanvasPixbuf *item;
	static GnomeCanvasBpath *bpath = NULL;
	GnomeCanvasPathDef *gpp;

	if (bpath) {
		gtk_object_destroy (GTK_OBJECT (bpath));
		bpath = NULL;
	}
	canvas = GNOME_CANVAS (glade_xml_get_widget (xml, "canvas1"));
	item = gnome_canvas_root (canvas)->item_list->data;

	lst = gribouy_bg_get_bounds (item);
	curves = clear_bounds (lst);
	tmp = curves;
	while (tmp) {
		lst = tmp->data;
		while (lst) {
			p = lst->data;
			g_free (p);
			lst = lst->next;
		}
		g_list_free (lst);
		tmp = tmp->next;
	}
	g_list_free (curves);
	bpath = GNOME_CANVAS_BPATH (gnome_canvas_item_new (
				gnome_canvas_root (canvas),
				gnome_canvas_bpath_get_type (),
				"bpath", gpp,
				"fill_color_rgba", 0xddddff80,
				"outline_color", "black",
				"width_units", 1.0,
				NULL));
	gnome_canvas_path_def_unref (gpp);
}

static void
gribouy_bg_transfer_ok (GladeXML *xml, GtkWidget *button)
{
#if 0
	GnomeCanvas *canvas;
	GnomeCanvasItem *img, *bpath;
	GnomeCanvasPathDef *gpp;
	GribouyLetter *letter;
	GtkWidget *dialog;

	gribouy_bg_transfer_apply (xml, NULL);
	letter = GRIBOUY_LETTER (g_object_get_data ((button), "letter"));
	canvas = GNOME_CANVAS (glade_xml_get_widget (xml, "canvas1"));
	img = gnome_canvas_root (canvas)->item_list->data;
	bpath = gnome_canvas_root (canvas)->item_list->next->data;
	if (GNOME_IS_CANVAS_PIXBUF (img))
		g_print ("Ok1 !!\n");
	if (GNOME_IS_CANVAS_BPATH (bpath))
		g_print ("Ok2 !!\n");
	g_object_get (G_OBJECT (bpath), "bpath", &gpp, NULL);
	gribouy_letter_add_gpp (letter, gpp);
//	gnome_canvas_path_def_unref (gpp);
	dialog = glade_xml_get_widget (xml, "bg_transfer");
	gtk_widget_destroy (dialog);
#endif
}

GtkWidget *
gribouy_grid_dialog_new (GribouyLetterWindow *letter_window)
{
	GladeXML *xml;
	GtkWidget *dialog, *entry, *button;
	int grid_x, grid_y;
	gboolean visible, magnetic;

	g_object_get (G_OBJECT (letter_window->grid),
			"grid_x", &grid_x,
			"grid_y", &grid_y,
			"visible", &visible,
			"magnetic", &magnetic,
			NULL);
	xml = glade_xml_new (GLADE_FILE1, "grid", NULL);
	if (!xml)
		xml = glade_xml_new (GLADE_FILE, "grid", NULL);

	entry = glade_xml_get_widget (xml, "spinbutton1");
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (entry),
			(double) grid_x);
	g_signal_connect (G_OBJECT (entry), "value_changed", G_CALLBACK (gribouy_grid_dialog_x_value_changed), letter_window->grid);

	entry = glade_xml_get_widget (xml, "spinbutton2");
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (entry),
			(double) grid_y);
	g_signal_connect (G_OBJECT (entry), "value_changed", G_CALLBACK (gribouy_grid_dialog_y_value_changed), letter_window->grid);

	entry = glade_xml_get_widget (xml, "checkbutton1");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (entry), visible);
	g_signal_connect (G_OBJECT (entry), "toggled", G_CALLBACK (gribouy_grid_dialog_visible_toggled), letter_window->grid);

	entry = glade_xml_get_widget (xml, "checkbutton2");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (entry), magnetic);
	g_signal_connect (G_OBJECT (entry), "toggled", G_CALLBACK (gribouy_grid_dialog_magnetic_toggled), letter_window->grid);

	button = glade_xml_get_widget (xml, "closebutton1");
	dialog = glade_xml_get_widget (xml, "grid");
	g_signal_connect_swapped (G_OBJECT (button), "clicked",
			G_CALLBACK (gtk_widget_destroy), dialog);
	return dialog;
}

static void
gribouy_grid_dialog_visible_toggled (GtkCheckButton *entry,
		GObject *grid)
{
	gboolean visible;

	visible = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (entry));
	g_object_set (grid, "visible", visible, NULL);
}

static void
gribouy_grid_dialog_magnetic_toggled (GtkCheckButton *entry,
		GObject *grid)
{
	gboolean magnetic;

	magnetic = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (entry));
	g_object_set (grid, "magnetic", magnetic, NULL);
}
static void
gribouy_grid_dialog_x_value_changed (GtkSpinButton *entry,
		GObject *grid)
{
	gint grid_x;

	grid_x = gtk_spin_button_get_value_as_int (entry);
	g_object_set (grid, "grid_x", grid_x, NULL);
}

static void
gribouy_grid_dialog_y_value_changed (GtkSpinButton *entry,
		GObject *grid)
{
	gint grid_y;

	grid_y = gtk_spin_button_get_value_as_int (entry);
	g_object_set (grid, "grid_y", grid_y, NULL);
}
