/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "gribouy-grid.h"

enum {
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_VISIBLE,
	PROP_MAGNETIC,
	PROP_WIDTH,
	PROP_LEFT,
	PROP_GRID_X,
	PROP_GRID_Y,
};

static void gribouy_grid_init (GribouyGrid *grid);
static void gribouy_grid_class_init (GribouyGridClass *class);
static void gribouy_grid_set_property (GObject *object, guint property_id,
		const GValue *value,
		GParamSpec *pspec);
static void gribouy_grid_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec);
static void gribouy_grid_destroy (GtkObject *object);
static gint gribouy_grid_width_bar_event (GribouyGrid *grid,
		GdkEvent *event, GnomeCanvasLine *width_bar);
static GnomeCanvasPathDef *gribouy_grid_construct_grid (GribouyGrid *grid,
		GribouyLetterWindow *window, GArray *blue_values);

static GnomeCanvasGroupClass *parent_class = NULL;

GType
gribouy_grid_get_type (void)
{
	static GType gribouy_grid_type = 0;

	if (!gribouy_grid_type) {
		static const GTypeInfo gribouy_grid_info = {
			sizeof (GribouyGridClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_grid_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyGrid),
			0,
			(GInstanceInitFunc) gribouy_grid_init,
			NULL
		};

		gribouy_grid_type = g_type_register_static (
				gnome_canvas_group_get_type (),
				"GribouyGrid", &gribouy_grid_info,
				(GTypeFlags)0);
	}
	return gribouy_grid_type;
}

static void 
gribouy_grid_class_init (GribouyGridClass *class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_grid_set_property;
	g_object_class->get_property = gribouy_grid_get_property;

	g_object_class_install_property
		(g_object_class,
		 PROP_VISIBLE,
		 g_param_spec_boolean ("visible", NULL, NULL,
			 FALSE,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_MAGNETIC,
		 g_param_spec_boolean ("magnetic", NULL, NULL,
			 FALSE,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_WIDTH,
		 g_param_spec_uint ("width", NULL, NULL,
			 0, 10000, 0,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));

	g_object_class_install_property
		(g_object_class,
		 PROP_LEFT,
		 g_param_spec_int ("left", NULL, NULL,
			 -10000, 10000, 0,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));

	g_object_class_install_property
		(g_object_class,
		 PROP_GRID_X,
		 g_param_spec_int ("grid_x", NULL, NULL,
			 1, G_MAXINT, 1,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));

	g_object_class_install_property
		(g_object_class,
		 PROP_GRID_Y,
		 g_param_spec_int ("grid_y", NULL, NULL,
			 1, G_MAXINT, 1,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	object_class->destroy = gribouy_grid_destroy;
}

static void 
gribouy_grid_init (GribouyGrid *grid)
{
	grid->visible = TRUE;
	grid->magnetic = TRUE;
	grid->grid_x = 1;
	grid->grid_y = 1;
	grid->x1 = grid->y1 = grid->x2 = grid->y2 = 0;
}

static void
gribouy_grid_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec)
{
	GribouyGrid *grid;

	grid = GRIBOUY_GRID (object);
	switch (property_id) {
		case PROP_VISIBLE:
			g_value_set_boolean (value, grid->visible);
			break;
		case PROP_MAGNETIC:
			g_value_set_boolean (value, grid->magnetic);
			break;
		case PROP_WIDTH:
			g_value_set_uint (value, grid->width);
			break;
		case PROP_GRID_X:
			g_value_set_int (value, grid->grid_x);
			break;
		case PROP_GRID_Y:
			g_value_set_int (value, grid->grid_y);
			break;
		case PROP_LEFT:
			g_value_set_int (value, grid->left);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					property_id, pspec);
			break;
	}
}

static void
gribouy_grid_set_property (GObject *object, guint param_id,
		const GValue *value, GParamSpec *pspec)
{
	GribouyGrid *grid;
	ArtPoint *point;

	grid = GRIBOUY_GRID (object);

	switch (param_id) {
		case PROP_VISIBLE:
			grid->visible = g_value_get_boolean (value);
			gribouy_grid_construct (grid);
			break;
		case PROP_MAGNETIC:
			grid->magnetic = g_value_get_boolean (value);
			break;
		case PROP_WIDTH:
			grid->width = g_value_get_uint (value);
			break;
		case PROP_LEFT:
			grid->left = g_value_get_int (value);
			break;
		case PROP_GRID_X:
			grid->grid_x = g_value_get_int (value);
			gribouy_grid_construct (grid);
			break;
		case PROP_GRID_Y:
			grid->grid_y = g_value_get_int (value);
			gribouy_grid_construct (grid);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					param_id, pspec);
			break;
	}
}

static GnomeCanvasPathDef *
gribouy_grid_construct_grid (GribouyGrid *grid, GribouyLetterWindow *window,
		GArray *blue_values)
{
	GnomeCanvasPathDef *gpp;
	gint i, j;
	double y1 = -750, y2 = 0;

	gpp = gnome_canvas_path_def_new ();
	if (grid->width == 0)
		grid->width = 500;

	gnome_canvas_path_def_moveto (gpp, grid->left, -10000.);
	gnome_canvas_path_def_lineto (gpp, grid->left, 10000.);
	gnome_canvas_path_def_moveto (gpp, grid->left + grid->width, -10000.);
	gnome_canvas_path_def_lineto (gpp, grid->left + grid->width, 10000.);
	for (i = 0; i < blue_values->len; i ++) {
		if (-g_array_index (blue_values, int, i) < y1)
			y1 = (double)(-g_array_index (blue_values, int, i));
		if (-g_array_index (blue_values, int, i) > y2)
			y2 = (double)(-g_array_index (blue_values, int, i));

		gnome_canvas_path_def_moveto (gpp, grid->left, -g_array_index (blue_values, int, i));
		gnome_canvas_path_def_lineto (gpp, grid->left + grid->width, -g_array_index (blue_values, int, i));
	}
	g_print ("y1=%f, y2=%f\n", y1, y2);
	gnome_canvas_set_scroll_region (
			GNOME_CANVAS_ITEM (window->letter)->canvas,
			grid->left, y1 - 10,
			grid->left + grid->width, y2 + 10); 
	return gpp;
}

void
gribouy_grid_update (GribouyGrid *grid, double x1, double y1,
		double x2, double y2)
{
	GnomeCanvasPoints *pts;
	gdouble x, y;
	int X, Y;
	gint gx, gy;
	double pu, pux;
	GnomeCanvas *canvas;

	if (x1 == grid->x1 && y1 == grid->y1 && x2 == grid->x2 && y2 == grid->y2)
		return;
	else {
		grid->x1 = x1;
		grid->y1 = y1;
		grid->x2 = x2;
		grid->y2 = y2;
	}

	if (grid->grid) {
		gtk_object_destroy (GTK_OBJECT (grid->grid));
		grid->grid = NULL;
	}

	canvas = GNOME_CANVAS_ITEM (grid)->canvas;
	pu = canvas->pixels_per_unit;

	if (!grid->visible)
		return;

	grid->grid = GNOME_CANVAS_GROUP (
			gnome_canvas_item_new (GNOME_CANVAS_GROUP (grid),
				gnome_canvas_group_get_type (),
				"x", 0,
				"y", 0, NULL));
	gnome_canvas_item_lower_to_bottom (GNOME_CANVAS_ITEM (grid->grid));
	gx = grid->grid_x;
	gy = grid->grid_y;

	pux = 10. / pu;

	while (gx < pux)
		gx <<= 1;
	while (gy < pux)
		gy <<= 1;

	pts = gnome_canvas_points_new (2);

	X = ((int)x1) / gx * gx;
	for (x = X; x < x2; x += gx) {
		pts->coords[0] = x;
		pts->coords[1] = y1;
		pts->coords[2] = x;
		pts->coords[3] = y2;

		gnome_canvas_item_new (
				GNOME_CANVAS_GROUP (grid->grid),
				gnome_canvas_line_get_type (),
				"points", pts,
				"fill_color", "gray",
				"width_pixels", 0,
				NULL);
	}

	Y = ((int)y1) / gy * gy;
	for (y = Y; y < y2; y += gy) {
		pts->coords[0] = x1;
		pts->coords[1] = y;
		pts->coords[2] = x2;
		pts->coords[3] = y;

		gnome_canvas_item_new (
				GNOME_CANVAS_GROUP (grid->grid),
				gnome_canvas_line_get_type (),
				"points", pts,
				"fill_color", "gray",
				"width_pixels", 0,
				NULL);
	}
	gnome_canvas_points_unref (pts);
}

GnomeCanvasItem *
gribouy_grid_new (GribouyLetterWindow *window, GArray *blue_value, gint left, guint width)
{
	GnomeCanvasGroup *group;
	GribouyGrid *grid;
	GnomeCanvasPathDef *gpp;
	GnomeCanvasPoints *pts;
	int i, j;

	group = gnome_canvas_root (GNOME_CANVAS (window->canvas));
	grid = GRIBOUY_GRID (
		gnome_canvas_item_new (group,
				       gribouy_grid_get_type (),
				       "x", 0.,
				       "y", 0.,
				       NULL));
	gnome_canvas_item_lower_to_bottom ((GnomeCanvasItem *) grid);

	grid->width = width;
	grid->left = left;
	grid->window = window;
	if (grid->width == 0)
		grid->width = 500;

	gpp = gribouy_grid_construct_grid (grid, window, blue_value);
	grid->path = GNOME_CANVAS_BPATH (
			gnome_canvas_item_new (GNOME_CANVAS_GROUP (grid),
				gnome_canvas_bpath_get_type (),
				"bpath", gpp,
				"fill_color", NULL,
				"outline_color", "black",
				"width_pixels", 0,
				NULL));
	grid->grid = NULL;

	gribouy_grid_construct (grid);

	pts = gnome_canvas_points_new (2);

	pts->coords[0] = grid->left + grid->width;
	pts->coords[1] = -10000;
	pts->coords[2] = grid->left + grid->width;
	pts->coords[3] = 10000;

	grid->width_bar = GNOME_CANVAS_LINE (gnome_canvas_item_new (
				GNOME_CANVAS_GROUP (grid),
				gnome_canvas_line_get_type (),
				"points", pts,
				"fill_color", "gray",
				"width_pixels", 4,
				NULL));

	gnome_canvas_points_unref (pts);
	g_signal_connect_swapped (G_OBJECT (grid->width_bar),
			"event",
			G_CALLBACK (gribouy_grid_width_bar_event),
			grid);
	g_signal_connect_swapped (G_OBJECT (grid->window),
			"need-grid",
			G_CALLBACK (gribouy_grid_update),
			grid);

	return (GnomeCanvasItem *) grid;
}

static gint
gribouy_grid_width_bar_event (GribouyGrid *grid, GdkEvent *event,
		GnomeCanvasLine *width_bar)
{
	static gboolean dragging = FALSE;
	static double x;
	GnomeCanvasPoints *pts;
	double item_x, item_y;
	GdkCursor *cur;

	g_return_val_if_fail (GRIBOUY_IS_GRID (grid), 0);

	item_x = event->button.x;
	item_y = event->button.y;
	gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (width_bar), &item_x, &item_y);

	switch (event->type) {
		case GDK_ENTER_NOTIFY:
			cur = gdk_cursor_new (GDK_SB_H_DOUBLE_ARROW);
			gnome_canvas_item_grab (
					GNOME_CANVAS_ITEM (width_bar),
					GDK_POINTER_MOTION_MASK
					| GDK_BUTTON_PRESS_MASK
					| GDK_BUTTON_RELEASE_MASK
					| GDK_LEAVE_NOTIFY_MASK,
					cur, event->button.time);
			gdk_cursor_destroy (cur);
			break;
		case GDK_BUTTON_PRESS:
			if (event->button.button == 1) {
				g_object_get (G_OBJECT (width_bar),
						"points", &pts, NULL);
				x = pts->coords[0];
				gnome_canvas_points_unref (pts);
				dragging = TRUE;
			}
			break;
		case GDK_MOTION_NOTIFY:
			if (dragging && (event->motion.state & GDK_BUTTON1_MASK)) {
				g_object_get (G_OBJECT (width_bar),
						"points", &pts, NULL);
				pts->coords[0] = pts->coords[2] = item_x;
				g_object_set (G_OBJECT (width_bar),
						"points", pts, NULL);
				grid->width = item_x - grid->left;
				g_object_set (G_OBJECT (grid->window->letter),
						"width", grid->width, NULL);
				x = item_x;
			}
			break;
		case GDK_BUTTON_RELEASE:
			dragging = FALSE;
			break;
		case GDK_LEAVE_NOTIFY:
			gnome_canvas_item_ungrab ((GnomeCanvasItem *)width_bar,
					event->button.time);
			break;
		default:
			break;
	}
	return FALSE;
}

/* Destroy handler for Gribouy grid */
static void
gribouy_grid_destroy (GtkObject *object)
{
	GribouyGrid *grid;

	g_return_if_fail (GRIBOUY_IS_GRID (object));

	grid = GRIBOUY_GRID (object);

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

void
gribouy_grid_construct (GribouyGrid *grid)
{
	GnomeCanvas *canvas;
	double x1, y1, x2, y2;
	int canvas_height, canvas_width;
	double w2c[6], inv[6];
	ArtPoint w, c;

	canvas = GNOME_CANVAS_ITEM (grid)->canvas;
	canvas_height = GTK_WIDGET (canvas)->allocation.height;
	canvas_width = GTK_WIDGET (canvas)->allocation.width;
	grid->x1 = grid->y1 = grid->x2 = grid->y2 = 0;
	x1 = canvas->layout.hadjustment->value - canvas->zoom_xofs;
	y1 = canvas->layout.vadjustment->value - canvas->zoom_yofs;
	x2 = x1 + canvas_width;
	y2 = y1 + canvas_height;
	gnome_canvas_window_to_world (canvas, x1, y1, &x1, &y1);
	gnome_canvas_window_to_world (canvas, x2, y2, &x2, &y2);

	gribouy_grid_update (grid, x1, y1, x2, y2);
}
