/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gtk/gtksignal.h>
#include <gtk/gtkmarshal.h>
#include <gtk/gtk.h>

#include "gribouy-letter.h"
#include "gribouy-curve.h"
#include "gribouy-cursor.h"

enum {
	PROP_0,
	PROP_WIDTH,
	PROP_LEFT
};

static void gribouy_letter_init (GribouyLetter *letter);
static void gribouy_letter_class_init (GribouyLetterClass *class);
static void gribouy_letter_update_path (GribouyLetter *letter);
static void gribouy_letter_edit (GribouyObject *obj, gboolean v);
static void gribouy_letter_move_letter (GribouyObject *obj,
		gdouble x, gdouble y);

static void gribouy_letter_set_property (GObject *object, guint property_id,
					const GValue *value,
					GParamSpec *pspec);
static void gribouy_letter_get_property (GObject *object, guint property_id,
					GValue *value, GParamSpec *pspec);
static void gribouy_letter_destroy (GtkObject *object);

static GribouyObjectClass *parent_class = NULL;

GType
gribouy_letter_get_type (void)
{
	static GType gribouy_letter_type = 0;

	if (!gribouy_letter_type) {
		static const GTypeInfo gribouy_letter_info = {
			sizeof (GribouyLetterClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_letter_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyLetter),
			0,
			(GInstanceInitFunc) gribouy_letter_init
		};

		gribouy_letter_type = g_type_register_static (
				gribouy_object_get_type (),
				"GribouyLetter", &gribouy_letter_info,
				(GTypeFlags)0);
	}
	return gribouy_letter_type;
}

static void 
gribouy_letter_class_init (GribouyLetterClass *class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	GribouyObjectClass *obj_class = (GribouyObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_letter_set_property;
	g_object_class->get_property = gribouy_letter_get_property;

	g_object_class_install_property
		(g_object_class,
		 PROP_WIDTH,
		 g_param_spec_uint ("width", NULL, NULL,
			 0, 1000, 500,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));

	g_object_class_install_property
		(g_object_class,
		 PROP_LEFT,
		 g_param_spec_int ("left", NULL, NULL,
			 -1000, 1000, 0,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));

	object_class->destroy = gribouy_letter_destroy;
	obj_class->edit = gribouy_letter_edit;
	obj_class->move_selection = gribouy_letter_move_letter;
}

static void
gribouy_letter_move_letter (GribouyObject *obj, gdouble x, gdouble y)
{
	gnome_canvas_item_move (GNOME_CANVAS_ITEM (obj),
			x, y);
}

static void 
gribouy_letter_init (GribouyLetter *letter)
{
}

static void
gribouy_letter_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec)
{
	GribouyLetter *letter;

	letter = GRIBOUY_LETTER (object);
	switch (property_id) {
		case PROP_WIDTH:
			g_value_set_uint (value, letter->width);
			break;
		case PROP_LEFT:
			g_value_set_int (value, letter->left);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					property_id, pspec);
			break;
	}
}

static void
gribouy_letter_set_property (GObject *object, guint param_id,
		const GValue *value, GParamSpec *pspec)
{
	GribouyLetter *letter;

	letter = GRIBOUY_LETTER (object);

	switch (param_id) {
		case PROP_WIDTH:
			letter->width = g_value_get_uint (value);
			break;
		case PROP_LEFT:
			letter->left = g_value_get_int (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					param_id, pspec);
			break;
	}
}

GribouyObject *
gribouy_letter_new (GribouyLetterWindow *window)
{
	GribouyObject *letter;
	GnomeCanvasItem *center;
	GnomeCanvasPathDef *gpp;
	GnomeCanvasGroup *root;

	root = gnome_canvas_root (GNOME_CANVAS (window->canvas));

	letter = GRIBOUY_OBJECT (gnome_canvas_item_new (root,
				gribouy_letter_get_type (),
				"cursor", GDK_PLUS,
				NULL));
	gpp = gnome_canvas_path_def_new ();
	center = gnome_canvas_item_new (GNOME_CANVAS_GROUP (letter),
			gnome_canvas_bpath_get_type (),
			"bpath", gpp,
			"fill_color_rgba", 0xddddff80,
			"width_pixels", 0,
			NULL);
	g_object_set (G_OBJECT (letter),
			"center", center,
			"window", window,
			NULL);
	gnome_canvas_path_def_unref (gpp);
	return letter;
}

static void
read_art (GribouyLetter *letter, gpointer prog, GPtrArray *subroutines, gboolean rg, gint *left, guint *width)
{
	GList *prg;
	GribouyArt *art;
	static ArtPoint current;

	if (!rg)
		current.x = current.y = 0.;

	for (prg = prog; prg != NULL; prg = prg->next) {
		art = prg->data;
		switch (art->code) {
		case HSTEM:
/*			gribouy_hstem_new (group, -1. * art->nb[0],
					  -1. * (art->nb[0] + art->nb[1]));
*/			break;
		case VSTEM:
/*			gribouy_vstem_new (group, *left + art->nb[0],
					  *left + (art->nb[0] + art->nb[1]));
*/			break;
		case VMOVETO:
			current.y -= art->nb[0];
			gribouy_letter_moveto (letter, &current);
			break;
		case RMOVETO:
			current.x += art->nb[0];
			current.y -= art->nb[1];
			gribouy_letter_moveto (letter, &current);
			break;
		case HMOVETO:
			current.x += art->nb[0];
			gribouy_letter_moveto (letter, &current);
			break;
		case RLINETO:
			current.x += art->nb[0];
			current.y -= art->nb[1];
			gribouy_letter_lineto (letter, &current);
			break;
		case HLINETO:
			current.x += art->nb[0];
			gribouy_letter_lineto (letter, &current);
			break;
		case VLINETO:
			current.y -= art->nb[0];
			gribouy_letter_lineto (letter, &current);
			break;
		case RRCURVETO:
			gribouy_letter_curveto (letter,
					current.x + art->nb[0],
					current.y - art->nb[1],
					current.x + art->nb[0] + art->nb[2],
					current.y - art->nb[1] - art->nb[3],
					current.x + art->nb[0] + art->nb[2] + art->nb[4],
					current.y - art->nb[1] - art->nb[3] - art->nb[5]);
			current.x += art->nb[0] + art->nb[2]+ art->nb[4];
			current.y -= (art->nb[1] + art->nb[3]+ art->nb[5]);
			break;
		case CLOSEPATH:
			gribouy_letter_closepath (letter);
//			gribouy_letter_moveto (letter, &current);
			break;

		case CALLSUBR:
			read_art (letter, g_ptr_array_index (subroutines, art->nb[0]), subroutines, TRUE, left, width);
			break;
		case RETURN:
			return;
			break;
		case CALLOTHERSUBR:
			if (art->nb[1] == 1 && art->nb[2] == 3) {
				read_art (letter, g_ptr_array_index (subroutines, art->nb[0]), subroutines, TRUE, left, width);
				prg = prg->next;
			}
			else {
				g_print ("\n*** problem...%d, %d, %d\n ***\n",
					   art->nb[0],
					   art->nb[1],
					   art->nb[2]);
			}
			break;
		case HSBW:
			current.y = 0;
			*left = current.x = art->nb[0];
			*width = art->nb[1];
			gribouy_letter_moveto (letter, &current);
			break;
		case VHCURVETO:
			gribouy_letter_curveto(letter,
					current.x,
					current.y - art->nb[0],
					current.x + art->nb[1],
					current.y - art->nb[0] - art->nb[2],
					current.x + art->nb[1] + art->nb[3],
					current.y - art->nb[0] - art->nb[2]);
			current.x += art->nb[1] + art->nb[3];
			current.y -= (art->nb[0] + art->nb[2]);
			break;
		case HVCURVETO:
			gribouy_letter_curveto (letter,
					current.x + art->nb[0],
					current.y,
					current.x + art->nb[0] + art->nb[1],
					current.y - art->nb[2],
					current.x + art->nb[0] + art->nb[1],
					current.y - art->nb[2] - art->nb[3]);
			current.x += art->nb[0] + art->nb[1];
			current.y -= (art->nb[2] + art->nb[3]);
			break;
		case ENDCHAR:
			break;
		default:
			g_print ("%d not implemented...\n", art->code);
		}
	}
}


GribouyObject *
gribouy_letter_new_with_art (GribouyLetterWindow *window, gpointer prog,
		GribouyType1 *font)
{
	GribouyLetter *letter;
	GribouyObject *obj;
	GnomeCanvasItem *point;
	gint left;
	guint width;

	letter = GRIBOUY_LETTER (gribouy_letter_new (window));
	
	if (prog) {
		read_art (letter, prog, font->subroutines,
				FALSE, &left, &width);
		letter->width = width;
		letter->left = left;
	}
	return (GribouyObject *)letter;
}

/* Destroy handler for Gribouy letter */
static void
gribouy_letter_destroy (GtkObject *object)
{
	GribouyLetter *letter;

	g_return_if_fail (GRIBOUY_IS_LETTER (object));

	letter = GRIBOUY_LETTER (object);

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

void
gribouy_letter_moveto (GribouyLetter *letter, ArtPoint *p)
{
	GribouyCurve *curve;

	if (g_list_length (GNOME_CANVAS_GROUP (letter)->item_list) > 1) {
		curve = GRIBOUY_CURVE (GNOME_CANVAS_GROUP (letter)->item_list_end->data);
		if (g_list_length (GNOME_CANVAS_GROUP (curve)->item_list) > 2) {
			curve = (GribouyCurve *) gribouy_curve_new (GRIBOUY_OBJECT (letter));
			g_signal_connect_swapped (G_OBJECT (curve),
					"changed",
					G_CALLBACK (gribouy_letter_update_path),
					letter);
		}
		else
			gtk_object_destroy (GTK_OBJECT (GNOME_CANVAS_GROUP (curve)->item_list_end->data));
	}
	else {
		curve = (GribouyCurve *) gribouy_curve_new (GRIBOUY_OBJECT (letter));
		g_signal_connect_swapped (G_OBJECT (curve),
				"changed",
				G_CALLBACK (gribouy_letter_update_path),
				letter);
	}
	gribouy_curve_moveto (curve, p);
}

void
gribouy_letter_lineto (GribouyLetter *letter, ArtPoint *p)
{
	GribouyCurve *curve;

	curve = GRIBOUY_CURVE (GNOME_CANVAS_GROUP (letter)->item_list_end->data);
	gribouy_curve_lineto (curve, p);
}

void
gribouy_letter_curveto (GribouyLetter *letter,
		double x1, double y1, double x2, double y2,
		double x3, double y3)
{
	GribouyCurve *curve;

	curve = GRIBOUY_CURVE (GNOME_CANVAS_GROUP (letter)->item_list_end->data);
	gribouy_curve_curveto (curve, x1, y1, x2, y2, x3, y3);
}

void
gribouy_letter_closepath (GribouyLetter *letter)
{
	GribouyCurve *curve;
	ArtPoint p;
	GList *lst;

	curve = GRIBOUY_CURVE (GNOME_CANVAS_GROUP (letter)->item_list_end->data);
	g_object_get (G_OBJECT (curve), "children", &lst, NULL);
	g_object_get (G_OBJECT (lst->data), "x", &p.x, "y", &p.y, NULL);
	gribouy_curve_lineto (curve, &p);
}

void
gribouy_letter_append_point (GribouyObject *letter, ArtPoint *p)
{
	GribouyCurve *curve;

	if (!GNOME_CANVAS_GROUP (letter)->item_list->next) {
		curve = (GribouyCurve *) gribouy_curve_new (letter);
		g_signal_connect_swapped (G_OBJECT (curve),
				"changed",
				G_CALLBACK (gribouy_letter_update_path),
				letter);
	
		gribouy_curve_append_point (curve, p);
	}
	else {
		curve = GRIBOUY_CURVE (GNOME_CANVAS_GROUP (letter)->item_list_end->data);
		if (gribouy_curve_is_closed (curve)) {
			curve = (GribouyCurve *) gribouy_curve_new (letter);
			g_signal_connect_swapped (G_OBJECT (curve),
					"changed",
					G_CALLBACK (gribouy_letter_update_path),
					letter);
		}			
		gribouy_curve_append_point (curve, p);
	}
}

static void
gribouy_bpath_affine_transform (ArtBpath *dst, const ArtBpath *src, double affine[6])
{
	int i;
	double x, y;
	int code;

	for (i = 0; src[i].code != ART_END; i ++) {
		code = src[i].code;
		dst[i].code = code;
		if (code == ART_CURVETO) {
			x = src[i].x1;
			y = src[i].y1;
			dst[i].x1 = affine[0] * x + affine[2] * y + affine[4];
			dst[i].y1 = affine[1] * x + affine[3] * y + affine[5];

			x = src[i].x2;
			y = src[i].y2;
			dst[i].x2 = affine[0] * x + affine[2] * y + affine[4];
			dst[i].y2 = affine[1] * x + affine[3] * y + affine[5];
		}
/*		else {
			dst[i].x1 = 0;
			dst[i].y1 = 0;
			dst[i].x2 = 0;
			dst[i].y2 = 0;
		}*/
		x = src[i].x3;
		y = src[i].y3;
		dst[i].x3 = affine[0] * x + affine[2] * y + affine[4];
		dst[i].y3 = affine[1] * x + affine[3] * y + affine[5];
	}
	dst[i].code = ART_END;
/*	dst[i].x1 = 0;
	dst[i].y1 = 0;
	dst[i].x2 = 0;
	dst[i].y2 = 0;
	dst[i].x3 = 0;
	dst[i].y3 = 0;*/
}

static void
gribouy_letter_update_path (GribouyLetter *letter)
{
	GList *lst;
	GnomeCanvasItem *center;
	GnomeCanvasPathDef *gpp;
	guint length;
	ArtBpath *bpath, *bp;
	double c2w[6], w2l[6], c2l[6];

	lst = GNOME_CANVAS_GROUP (letter)->item_list;
	center = lst->data;
	length = 1;
	for (lst = lst->next; lst; lst = lst->next) {
		g_object_get (G_OBJECT (GRIBOUY_OBJECT (lst->data)->center),
				"bpath", &gpp, NULL);
		length += gnome_canvas_path_def_length (gpp) - 1;
		gnome_canvas_path_def_unref (gpp);
	}
	bpath = art_new (ArtBpath, length);
	bp = bpath;

	lst = GNOME_CANVAS_GROUP (letter)->item_list;
	for (lst = lst->next; lst; lst = lst->next) {
		g_object_get (G_OBJECT (GRIBOUY_OBJECT (lst->data)->center),
				"bpath", &gpp, NULL);
		length = gnome_canvas_path_def_length (gpp) - 1;
		gnome_canvas_item_i2w_affine (GNOME_CANVAS_ITEM (lst->data),
				c2w);
		gnome_canvas_item_i2w_affine (GNOME_CANVAS_ITEM (letter),
				w2l);
		art_affine_invert (w2l, w2l);
		art_affine_multiply (c2l, c2w, w2l);
		gribouy_bpath_affine_transform (bp,
				gnome_canvas_path_def_bpath (gpp),
				c2l);
		gnome_canvas_path_def_unref (gpp);
		bp += length;
	}
	gpp = gnome_canvas_path_def_new_from_bpath (bpath);
	g_object_set (G_OBJECT (center), "bpath", gpp, NULL);
	gnome_canvas_path_def_unref (gpp);
}

#if 0
static void
gribouy_letter_update_path (GribouyLetter *letter)
{
	GList *lst;
	GSList *gpps, *sl;
	GnomeCanvasItem *center;
	GnomeCanvasPathDef *gpp;
	ArtBpath *bpath, *bp;
	guint length;

	lst = GNOME_CANVAS_GROUP (letter)->item_list;
	center = lst->data;
	gpps = NULL;

	lst = lst->next;
	length = 1;
	while (lst) {
		g_object_get (G_OBJECT (GRIBOUY_OBJECT (lst->data)->center),
				"bpath", &gpp, NULL);
		length += gnome_canvas_path_def_length (gpp) - 1;
		gpps = g_slist_append (gpps, gpp);
		gnome_canvas_path_def_unref (gpp);
		lst = lst->next;
	}
	bpath = art_new (ArtBpath, length);
	bp = bpath;
	sl = gpps;
	while (sl) {
		gpp = sl->data;
		length = gnome_canvas_path_def_length (gpp) - 1;
		memcpy (bp, gnome_canvas_path_def_bpath (gpp), length * sizeof (ArtBpath));
		bp += length;
		sl = sl->next;
	}
	bp->code = ART_END;
	gpp = gnome_canvas_path_def_new_from_bpath (bpath);
	g_slist_free (gpps);
	g_object_set (G_OBJECT (center), "bpath", gpp, NULL);
	gnome_canvas_path_def_unref (gpp);
}
#endif

#if 0
static void
gribouy_letter_update_path (GribouyLetter *letter)
{
	GList *lst;
	GSList *gpps;
	GnomeCanvasItem *center;
	GnomeCanvasPathDef *gpp;

	lst = GNOME_CANVAS_GROUP (letter)->item_list;
	center = lst->data;
	gpps = NULL;

	lst = lst->next;
	while (lst) {
		g_object_get (G_OBJECT (GRIBOUY_OBJECT (lst->data)->center),
				"bpath", &gpp, NULL);
		gpps = g_slist_append (gpps, gpp);
		lst = lst->next;
	}
	gpp = gnome_canvas_path_def_concat (gpps);
	g_slist_free (gpps);
	g_object_set (G_OBJECT (center), "bpath", gpp, NULL);
}
#endif

GList *
gribouy_letter_save (GribouyLetter *letter)
{
	GList *retval, *lst;
	GnomeCanvasPathDef *gpp;
	ArtBpath *bpath, *bp;
	GribouyArt *art;
	gint XP3, YP3, X1, Y1, X2, Y2, X3, Y3, xstart, ystart;
	double x1, x2, y1, y2;

	retval = NULL;
/*	for (lst = curve->plan3->item_list;
	     lst != NULL;
	     lst = g_list_next (lst)) {
		if (GRIBOUY_IS_HSTEM (lst->data)) {
			gtk_object_get (GTK_OBJECT (lst->data),
					"y1", &y1,
					"y2", &y2,
					NULL);
			art = g_new0 (GribouyArt, 1);
			art->code = HSTEM;
			art->nb[0] = - y1;
			art->nb[1] = y1 - y2;

			retval = g_list_append (retval, art);
		}
		if (GRIBOUY_IS_VSTEM (lst->data)) {
			gtk_object_get (GTK_OBJECT (lst->data),
					"x1", &x1,
					"x2", &x2,
					NULL);
			art = g_new0 (GribouyArt, 1);
			art->code = VSTEM;
			art->nb[0] = x1;
			art->nb[1] = x2 - x1;

			retval = g_list_append (retval, art);
		}
	}
*/
	g_object_get (G_OBJECT (GRIBOUY_OBJECT (letter)->center),
			"bpath", &gpp, NULL);
	bpath = gnome_canvas_path_def_bpath (gpp);

	if (bpath->code == ART_END)
		return retval;

	art = g_new0 (GribouyArt, 1);

	art->code = HSBW;
	art->nb[0] = letter->left;
	art->nb[1] = letter->width;
	retval = g_list_append (retval, art);

	bp = bpath;
	YP3 = 0;
	XP3 = letter->left;
	art = g_new0 (GribouyArt, 1);

	do {
		X3 = bp->x3;
		Y3 = bp->y3;
		switch (bp->code) {
		case ART_MOVETO_OPEN:
			g_warning ("ART_MOVETO_OPEN ????\n");
		case ART_MOVETO:
			if (X3 == XP3) {
				art->code = VMOVETO;
				art->nb[0] = YP3 - Y3;
				YP3 = Y3;
			}
			else if (Y3 == YP3) {
				art->code = HMOVETO;
				art->nb[0] = X3 - XP3;
				XP3 = X3;
			}
			else {
				art->code = RMOVETO;
				art->nb[0] = X3 - XP3;
				art->nb[1] = YP3 - Y3;
				XP3 = X3;
				YP3 = Y3;
			}
			xstart = XP3;
			ystart = YP3;
			break;
		case ART_LINETO:
			if (X3 == xstart && Y3 == ystart) {
				art->code = CLOSEPATH;
				g_print ("CLOSEPATH\n");
			}
			else if (X3 == XP3) {
				art->code = VLINETO;
				art->nb[0] = YP3 - Y3;
				YP3 = Y3;
			}
			else if (Y3 == YP3) {
				art->code = HLINETO;
				art->nb[0] = X3 - XP3;
				XP3 = X3;
			}
			else {
				art->code = RLINETO;
				art->nb[0] = X3 - XP3;
				art->nb[1] = YP3 - Y3;
				XP3 = X3;
				YP3 = Y3;
			}
			break;
		case ART_CURVETO:
			X1 = bp->x1;
			Y1 = bp->y1;
			X2 = bp->x2;
			Y2 = bp->y2;
			if (XP3 == X1 && Y2 == Y3) {
				art->code = VHCURVETO;
				art->nb[0] = YP3 - Y1;
				art->nb[1] = X2 - XP3;
				art->nb[2] = Y1 - Y2;
				art->nb[3] = X3 - X2;
				XP3 = X3;
				YP3 = Y3;
			}
			else if (YP3 == Y1 && X2 == X3) {
				art->code = HVCURVETO;
				art->nb[0] = X1 - XP3;
				art->nb[1] = X2 - X1;
				art->nb[2] = YP3 - Y2;
				art->nb[3] = Y2 - Y3;
				XP3 = X3;
				YP3 = Y3;
			}
			else {
				art->code = RRCURVETO;
				art->nb[0] = X1 - XP3;
				art->nb[1] = YP3 - Y1;
				art->nb[2] = X2 - X1;
				art->nb[3] = Y1 - Y2;
				art->nb[4] = X3 - X2;
				art->nb[5] = Y2 - Y3;
				XP3 = X3;
				YP3 = Y3;
			}
			if (XP3 == xstart && YP3 == ystart) {
				retval = g_list_append (retval, art);
				art = g_new0 (GribouyArt, 1);
				art->code = CLOSEPATH;
				g_print ("CLOSEPATH\n");
			}
			break;
		default :
			g_print ("not yet implemented...\n");
		}
		bp ++;
		retval = g_list_append (retval, art);
		art = g_new0 (GribouyArt, 1);
	}
	while (bp->code != ART_END);
	art->code = ENDCHAR;
	retval = g_list_append (retval, art);
	return retval;
}

static void
gribouy_letter_edit (GribouyObject *obj, gboolean v)
{
}

