/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gnome.h>
#include <config.h>
#include <bonobo/bonobo-main.h>
#include <bonobo/bonobo-window.h>
#include <bonobo/bonobo-ui-util.h>
#include <glade/glade.h>

#include "gribouy-letter.h"
#include "gribouy-letter-window.h"
#include "gribouy-type1.h"
#include "gribouy-triplet.h"
#include "gribouy-bg.h"
#include "gribouy-cursor.h"

static void gribouy_die (GnomeClient *client, gpointer user_data);
static gboolean gribouy_save_yourself (GnomeClient *client,
		gint phase,
		GnomeSaveStyle save_style,
		gboolean shutdown,
		GnomeInteractStyle interact_style,
		gboolean fast,
		gpointer user_data);
static gboolean gribouy_main_get_file (GtkFileSelection *sel);
static gboolean gribouy_main_open_file (char *filename);
static void gribouy_main_set_current (char *str, GtkTreeIter *iter);
static void gribouy_main_save_file (GtkFileSelection *sel);
static GtkTreeIter *model_append (const gchar *key, gpointer *value,
		GtkListStore **user_data);
static void construct_model (gchar *key, gpointer *value,
		GtkListStore **user_data);
static void construct_model_with_standard_encoding (GHashTable *symbols,
		GtkListStore **model);
static void gribouy_main_create_model ();
static void gribouy_main_select_cursors (GnomeCanvasRect *sel, gboolean keep);
static gboolean gribouy_main_save (gchar *file);
static void gribouy_main_init_all ();
static void gribouy_main_update_model (gboolean standard_encoding);

static gboolean on_treeview_doubleclick (GtkTreeView *treeview, GdkEvent *event, gpointer user_data);
static void on_letter_window_changed (GtkWidget *window, gboolean changed, gpointer user_data);

gboolean on_treeview_button_press_event (GtkWidget *treeview,
		GdkEventButton *event,
		gpointer user_data);
void on_treeview_changed (GtkTreeSelection *selection, gpointer user_data);
void on_font_properties_okbutton_clicked (GtkObject *dial, gpointer *object);

static void verb_FileNew_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_FileSave_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_FileSaveAs_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_FileOpen_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_FileExit_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_EditProperties_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_HelpAbout_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static gboolean event_LettersEncoding_cb (GtkCheckMenuItem *item,
		gpointer user_data);
static void verb_LettersNew_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);
static void verb_LettersDelete_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname);

gboolean on_letter_name_okbutton_clicked (GtkWidget *entry);
void on_clear_activate (GtkMenuItem *item, gpointer user_data);

static char *filename = NULL;

static BonoboUIVerb gribouy_app_verbs[] = {
	BONOBO_UI_VERB ("FileNew", verb_FileNew_cb),
	BONOBO_UI_VERB ("FileOpen", verb_FileOpen_cb),
	BONOBO_UI_VERB ("FileSave", verb_FileSave_cb),
	BONOBO_UI_VERB ("FileSaveAs", verb_FileSaveAs_cb),
	BONOBO_UI_VERB ("FileExit", verb_FileExit_cb),
	BONOBO_UI_VERB ("EditProperties", verb_EditProperties_cb),
	BONOBO_UI_VERB ("LettersNew", verb_LettersNew_cb),
	BONOBO_UI_VERB ("LettersDelete", verb_LettersDelete_cb),
	BONOBO_UI_VERB ("HelpAbout", verb_HelpAbout_cb),
	BONOBO_UI_VERB_END
};

typedef struct _GribouyApp GribouyApp;

struct _GribouyApp {
	BonoboWindow *win;
	gchar *fullname;
	gchar *filename;
	ArtPoint pt;
	GribouyType1 *font;
	GtkListStore *model;
	GtkTreeView *treeview;
};

GribouyApp *gribouy_app;

gboolean
on_treeview_button_press_event (GtkWidget *treeview,
		   GdkEventButton *event,
		   gpointer user_data)
{
	ArtPoint point;
	GtkWidget *item;
	GtkWidget *menu = gtk_menu_new ();
	gboolean se;
	BonoboUIEngine *engine;
	BonoboUINode *node;

	point.x = event->x;
	point.y = event->y;

	switch (event->type) {
	case GDK_BUTTON_PRESS:
		if (event->button == 3) {
			bonobo_window_add_popup (gribouy_app->win,
					GTK_MENU (menu), "/popups/letters");
			gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, gribouy_app, 3, event->time);
			gtk_widget_show (menu);
			engine = bonobo_window_get_ui_engine (gribouy_app->win);
			g_object_get (G_OBJECT (gribouy_app->font), "standard_encoding", &se, NULL);
			node = bonobo_ui_engine_get_path (engine, "/popups/letters/LettersEncoding");
			item = bonobo_ui_engine_node_get_widget (engine, node);
			gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), se);
			g_signal_connect (G_OBJECT (item), "toggled", G_CALLBACK (event_LettersEncoding_cb), NULL);
			node = bonobo_ui_engine_get_path (engine, "/popups/letters/LettersNew");
			item = bonobo_ui_engine_node_get_widget (engine, node);
			gtk_widget_set_sensitive (item, !se);
			node = bonobo_ui_engine_get_path (engine, "/popups/letters/LettersDelete");
			item = bonobo_ui_engine_node_get_widget (engine, node);
			gtk_widget_set_sensitive (item, !se);
		}
		break;
	}
	return FALSE;
}
		   
void
on_treeview_changed (GtkTreeSelection *selection, gpointer user_data)
{
}

/* Opens a new window containing the letter given by the string str.
 * iter corresponds to this letter iterator */
static void
gribouy_main_set_current (char *str, GtkTreeIter *iter)
{
	GtkWidget *letter_window;

	letter_window = gribouy_letter_window_new (str, gribouy_app->font);
	g_signal_connect (G_OBJECT (letter_window), "changed",
			G_CALLBACK (on_letter_window_changed), iter);

	gtk_widget_show_all (letter_window);
}

void
on_font_properties_okbutton_clicked (GtkObject *dial, gpointer *object)
{
	GtkWidget *entry;
	GladeXML *xml;
	gchar *txt;
	xmlNodePtr cur;
	GtkTextBuffer *buffer;
	GtkTextIter start, end;

	xml = glade_get_widget_tree (GTK_WIDGET (dial));
	cur = xmlDocGetRootElement (gribouy_app->font->xmlfont);
	while (cur && xmlStrcmp (cur->name, "Type1Font"))
		cur = cur->next;
	if (!cur)
		cur = xmlNewChild (cur, NULL, "Type1Font", NULL);

	entry = glade_xml_get_widget (xml, "fontname_entry");
	txt = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
	xmlSetProp (cur, "FontName", txt);
	g_free (txt);

	cur = cur->xmlChildrenNode;
	while (cur && xmlStrcmp (cur->name, "FontInfo"))
		cur = cur->next;
	if (!cur)
		cur = xmlNewChild (cur, NULL, "Type1Font", NULL);

	entry = glade_xml_get_widget (xml, "version_entry");
	txt = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
	xmlSetProp (cur, "version", txt);
	g_free (txt);

	entry = glade_xml_get_widget (xml, "notice_entry");
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (entry));

	gtk_text_buffer_get_start_iter (buffer, &start);
	gtk_text_buffer_get_end_iter (buffer, &end);

	txt = gtk_text_buffer_get_text (buffer, &start, &end, TRUE);
	xmlSetProp (cur, "Notice", txt);
	g_free (txt);

	entry = glade_xml_get_widget (xml, "copyright_entry");
	txt = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
	xmlSetProp (cur, "Copyright", txt);
	g_free (txt);

	entry = glade_xml_get_widget (xml, "fullname_entry");
	txt = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
	xmlSetProp (cur, "FullName", txt);
	g_free (txt);

	entry = glade_xml_get_widget (xml, "family_entry");
	txt = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
	xmlSetProp (cur, "FamilyName", txt);
	g_free (txt);

	gtk_widget_destroy (GTK_WIDGET (dial));
}

static void
verb_FileNew_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	gribouy_main_init_all ();
}

static void
verb_FileSave_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	gchar *error_str, *file;
	GtkDialog *dialog;
	gint answer;

	if (gribouy_app->filename) {
		file = g_strdup (gribouy_app->filename);
		gribouy_main_save (file);
		g_free (file);
	}
	else
		verb_FileSaveAs_cb (uic, user_data, cname);
}

static void
verb_FileSaveAs_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	GtkFileSelection *file_dlg;

	file_dlg = GTK_FILE_SELECTION (gtk_file_selection_new (_("Save the file...")));
	gtk_signal_connect (GTK_OBJECT (file_dlg), "delete_event",
			    (GtkSignalFunc)gtk_widget_destroy, NULL);
	gtk_signal_connect (GTK_OBJECT (file_dlg), "destroy",
			    (GtkSignalFunc)gtk_widget_destroy, NULL);

	gtk_signal_connect_object (GTK_OBJECT (file_dlg->cancel_button),
				   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT (file_dlg));

	gtk_signal_connect_object (GTK_OBJECT (file_dlg->ok_button),
				   "clicked",
				   (GtkSignalFunc) gribouy_main_save_file,
				   GTK_OBJECT (file_dlg));

	gtk_widget_show (GTK_WIDGET (file_dlg));
}

static void
verb_FileOpen_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	GtkFileSelection *file_dlg;

	file_dlg = GTK_FILE_SELECTION (gtk_file_selection_new (_("Open the file...")));
	gtk_signal_connect (GTK_OBJECT (file_dlg), "delete_event",
			    (GtkSignalFunc)gtk_widget_destroy, NULL);
	gtk_signal_connect (GTK_OBJECT (file_dlg), "destroy",
			    (GtkSignalFunc)gtk_widget_destroy, NULL);

	gtk_signal_connect_object (GTK_OBJECT (file_dlg->cancel_button),
				   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT (file_dlg));

	gtk_signal_connect_object (GTK_OBJECT (file_dlg->ok_button),
				   "clicked",
				   (GtkSignalFunc) gribouy_main_get_file,
				   GTK_OBJECT (file_dlg));

	gtk_widget_show (GTK_WIDGET (file_dlg));
	gtk_object_destroy (GTK_OBJECT (gribouy_app->model));
	gribouy_app->model = NULL;
}

static void
verb_FileExit_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	bonobo_main_quit ();
}

gboolean
on_letter_name_okbutton_clicked (GtkWidget *entry)
{
	GtkWidget *dial;
	char *txt;

	dial = entry->parent;
	while (dial->parent != NULL)
		dial = dial->parent;
	txt = gtk_entry_get_text (GTK_ENTRY (entry));
	if (!g_hash_table_lookup (gribouy_app->font->symbols, txt))
		model_append (txt, NULL, &gribouy_app->model);
	gtk_widget_destroy (dial);
	return TRUE;
}

static gboolean
gribouy_main_save (gchar *file)
{
	gchar *error_str;
	GtkDialog *dialog;
	gint answer;

	if (g_file_test (file, G_FILE_TEST_IS_REGULAR)) {
		error_str = g_strdup_printf (_("The file '%s' already exists. Do you want to replace it ?"), file);
		dialog = gtk_message_dialog_new (GTK_WINDOW (gribouy_app->win),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR,
						 GTK_BUTTONS_YES_NO,
						 error_str);

		g_free (error_str);
		gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_NO);

		gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

		answer = gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (GTK_WIDGET (dialog));
	}
	else
		answer = GTK_RESPONSE_YES;
	
	if (answer == GTK_RESPONSE_YES) {
		if (!gribouy_type1_save_file (file, gribouy_app->font)) {
			error_str = g_strdup_printf (_("I cannot save this font."));
			dialog = gtk_message_dialog_new (GTK_WINDOW (gribouy_app->win),
							 GTK_DIALOG_MODAL,
							 GTK_MESSAGE_ERROR,
							 GTK_BUTTONS_OK,
							 error_str);

			g_free (error_str);
			gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

			gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

			gtk_dialog_run (GTK_DIALOG (dialog));
			gtk_widget_destroy (GTK_WIDGET (dialog));
		}
		else {
			if (gribouy_app->filename)
				g_free (gribouy_app->filename);
			gribouy_app->filename = g_strdup (file);
		}
	}
}

static void
gribouy_main_save_file (GtkFileSelection *sel)
{
	gchar *file;
	GtkDialog *dialog;
	gchar *error_str;
	gint answer;

	file = g_strdup (gtk_file_selection_get_filename (sel));

	gtk_widget_destroy (GTK_WIDGET (sel));

	gribouy_main_save (file);
	g_free (file);
}

static void
verb_EditProperties_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	GladeXML *xml;
	GtkWidget *dialog, *entry;
	xmlNodePtr cur;
	char *txt;

	xml = glade_xml_new (gribouy_app->fullname, "font_properties", NULL);
	dialog = glade_xml_get_widget (xml, "font_properties");
	glade_xml_signal_autoconnect (xml);
	cur = xmlDocGetRootElement (gribouy_app->font->xmlfont);
	txt = xmlGetProp (cur, "FontName");
	entry = glade_xml_get_widget (xml, "fontname_entry");
	if (txt) {
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}
	else
		gtk_entry_set_text (GTK_ENTRY (entry), "No name");

	cur = cur->xmlChildrenNode;
	while (xmlStrcmp (cur->name, (const xmlChar *) "FontInfo"))
		cur = cur->next;

	txt = xmlGetProp (cur, "FullName");
	if (txt) {
		entry = glade_xml_get_widget (xml, "fullname_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "FamilyName");
	if (txt) {
		entry = glade_xml_get_widget (xml, "family_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "version");
	if (txt) {
		entry = glade_xml_get_widget (xml, "version_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "Notice");
	if (txt) {
		entry = GTK_WIDGET (gtk_text_view_get_buffer (GTK_TEXT_VIEW (glade_xml_get_widget (xml, "notice_entry"))));
		gtk_text_buffer_set_text (GTK_TEXT_BUFFER (entry), txt, -1);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "Copyright");
	if (txt) {
		entry = glade_xml_get_widget (xml, "copyright_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	cur = cur->parent;
	cur = cur->xmlChildrenNode;
	while (cur && xmlStrcmp (cur->name, (const xmlChar *) "FontBBox"))
		cur = cur->next;
	if (cur) {
		txt = xmlGetProp (cur, "v1");
		if (txt) {
			entry = glade_xml_get_widget (xml, "bbox1_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v2");
		if (txt) {
			entry = glade_xml_get_widget (xml, "bbox2_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v3");
		if (txt) {
			entry = glade_xml_get_widget (xml, "bbox3_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v4");
		if (txt) {
			entry = glade_xml_get_widget (xml, "bbox4_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}
		cur = cur->parent;
		cur = cur->xmlChildrenNode;
	}

	while (cur && xmlStrcmp (cur->name, (const xmlChar *) "FontMatrix"))
		cur = cur->next;

	if (cur) {
		txt = xmlGetProp (cur, "v1");
		if (txt) {
			entry = glade_xml_get_widget (xml, "matrix1_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v2");
		if (txt) {
			entry = glade_xml_get_widget (xml, "matrix2_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v3");
		if (txt) {
			entry = glade_xml_get_widget (xml, "matrix3_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v4");
		if (txt) {
			entry = glade_xml_get_widget (xml, "matrix4_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v5");
		if (txt) {
			entry = glade_xml_get_widget (xml, "matrix5_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}

		txt = xmlGetProp (cur, "v6");
		if (txt) {
			entry = glade_xml_get_widget (xml, "matrix6_entry");
			gtk_entry_set_text (GTK_ENTRY (entry), txt);
			g_free (txt);
		}
		cur = cur->parent;
		cur = cur->xmlChildrenNode;
	}

	while (xmlStrcmp (cur->name, (const xmlChar *) "Private"))
		cur = cur->next;

	cur = cur->xmlChildrenNode;
	while (xmlStrcmp (cur->name, (const xmlChar *) "BlueValues"))
		cur = cur->next;

	txt = xmlGetProp (cur, "v1");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bv1_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "v2");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bv2_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "v3");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bv3_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "v4");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bv4_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "v5");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bv5_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "v6");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bv6_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}
	cur = cur->parent;

	txt = xmlGetProp (cur, "BlueScale");
	if (txt) {
		entry = glade_xml_get_widget (xml, "bs_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "StdHW");
	if (txt) {
		entry = glade_xml_get_widget (xml, "stdhw_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	txt = xmlGetProp (cur, "StdVW");
	if (txt) {
		entry = glade_xml_get_widget (xml, "stdvw_entry");
		gtk_entry_set_text (GTK_ENTRY (entry), txt);
		g_free (txt);
	}

	entry = glade_xml_get_widget (xml, "okbutton1");
	gtk_signal_connect_object (GTK_OBJECT (entry), "clicked",
			    (GtkSignalFunc)on_font_properties_okbutton_clicked,
			    dialog);
}

static void
verb_HelpAbout_cb (BonoboUIComponent *uic, gpointer user_data,
		const char *cname)
{
	GtkWidget *dialog;
	GdkPixbuf *logo;
	char *filename;

	gchar *authors[] = {
		"David Boucher <bouda1@wanadoo.fr>",
		NULL
	};

	gchar *documenters[] = {
		NULL
	};

	filename = g_strdup (PACKAGE_ICONDIR"/gribouy-logo.jpg");
	logo = gdk_pixbuf_new_from_file (filename, NULL);
	g_free (filename);

	if (!logo)
		g_print ("No logo found...\n");
	dialog = gnome_about_new (_("Gribouy"),
			VERSION,
			"Copyright \xc2\xa9 2003 David Boucher",
			_("A type1 font editor"),
			authors,
			documenters,
			NULL,
			logo);
	g_signal_connect (G_OBJECT (dialog), "destroy",
			G_CALLBACK (gtk_widget_destroy), dialog);
	gtk_widget_show (dialog);
}

/* Append the given key to the model.
 * Returns the new iterator. */
static GtkTreeIter *
model_append (const gchar *key, gpointer *value, GtkListStore **user_data)
{
	g_return_if_fail (GTK_IS_LIST_STORE (*user_data));
	GtkTreeIter *iter;
	GtkTreeSelection *selection;

	iter = g_new0 (GtkTreeIter, 1);
	gtk_list_store_append (GTK_LIST_STORE (*user_data), iter);
	gtk_list_store_set (GTK_LIST_STORE (*user_data),
			    iter,
			    0, FALSE,
			    1, key,
			    -1);
	selection = gtk_tree_view_get_selection (gribouy_app->treeview);
	gtk_tree_selection_select_iter (selection, iter);
	return iter;
}

static void
construct_model (gchar *key, gpointer *value, GtkListStore **user_data)
{
	g_return_if_fail (GTK_IS_LIST_STORE (*user_data));
	GtkTreeIter iter;

	gtk_list_store_append (GTK_LIST_STORE (*user_data), &iter);
	gtk_list_store_set (GTK_LIST_STORE (*user_data),
				&iter,
				0, TRUE,
				1, key,
				-1);
}
       
static void
construct_model_with_standard_encoding (GHashTable *symbols,
		GtkListStore **model)
{
	GtkTreeIter iter;
	char **w;
	char *std_enc[] = {"exclam", "quotedbl", "numbersign", "dollar", "percent", "ampersand", "quotesingle", "parenleft", "parenright", "asterisk", "plus", "comma", "hyphen", "period", "slash", "zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "colon", "semicolon", "less", "equal", "greater", "question", "at", "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "bracketleft", "backslash", "bracketright", "asciicircum", "underscore", "grave", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "braceleft", "bar", "braceright", "asciitilde", "Adieresis", "Aring", "Ccedilla", "Eacute", "Ntilde", "Odieresis", "Udieresis", "aacute", "agrave", "acircumflex", "adieresis", "atilde", "aring", "ccedilla", "eacute", "egrave", "ecircumflex", "edieresis", "iacute", "igrave", "icircumflex", "idieresis", "ntilde", "oacute", "ograve", "ocircumflex", "odieresis", "otilde", "uacute", "ugrave", "ucircumflex", "udieresis", "dagger", "degree", "cent", "sterling", "section", "bullet", "paragraph", "germandbls", "registered", "copyright", "trademark", "acute", "dieresis", "notequal", "AE", "Oslash", "infinity", "plusminus", "lessequal", "greaterequal", "yen", "mu", "partialdiff", "summation", "product", "pi", "integral", "ordfeminine", "ordmasculine", "Omega", "ae", "oslash", "questiondown", "exclamdown", "logicalnot", "radical", "florin", "approxequal", "Delta", "guillemotleft", "guillemotright", "ellipsis", "nobreakspace", "Agrave", "Atilde", "Otilde", "OE", "oe", "endash", "emdash", "quotedblleft", "quotedblright", "quoteleft", "quoteright", "divide", "lozenge", "ydieresis", "Ydieresis", "fraction", "currency", "guilsinglleft", "guilsinglright", "fi", "fl", "daggerdbl", "periodcentered", "quotesinglbase", "quotedblbase", "perthousand", "Acircumflex", "Ecircumflex", "Aacute", "Edieresis", "Egrave", "Iacute", "Icircumflex", "Idieresis", "Igrave", "Oacute", "Ocircumflex", "apple", "Ograve", "Uacute", "Ucircumflex", "Ugrave", "dotlessi", "circumflex", "tilde", "macron", "breve", "dotaccent", "ring", "cedilla", "hungarumlaut", "ogonek", "caron", "Lslash", "lslash", "Scaron", "scaron", "Zcaron", "zcaron", "brokenbar", "Eth", "eth", "Yacute", "yacute", "Thorn", "thorn", "minus", "multiply", "onesuperior", "twosuperior", "threesuperior", "onehalf", "onequarter", "threequarters", NULL};

	for (w = std_enc; *w != NULL; w++) {
		if (g_hash_table_lookup (symbols, *w)) {
			gtk_list_store_append (*model, &iter);
			gtk_list_store_set (*model, &iter, 0, TRUE, 1, *w, -1);
		}
		else {
			gtk_list_store_append (*model, &iter);
			gtk_list_store_set (*model, &iter, 0, FALSE, 1, *w, -1);
		}
	}
}

static void
gribouy_main_update_model (gboolean standard_encoding)
{
	if (standard_encoding)
		construct_model_with_standard_encoding (
				gribouy_app->font->symbols,
				&gribouy_app->model);
	else
		g_hash_table_foreach (gribouy_app->font->symbols,
				(GFunc)construct_model,
				&gribouy_app->model);
}

static gboolean
gribouy_main_get_file (GtkFileSelection *sel)
{
	char *file;

	file = gtk_file_selection_get_filename (sel);
	gtk_widget_destroy (GTK_WIDGET (sel));

	g_object_unref (G_OBJECT (gribouy_app->font));
	return gribouy_main_open_file (file);
}

static gboolean
gribouy_main_open_file (char *filename)
{
	gchar *error_str;
	GtkDialog *dialog;

	if ((gribouy_app->font = GRIBOUY_TYPE1 (gribouy_type1_new_from_file (filename)))) {
		gribouy_app->model = gtk_list_store_new (2, G_TYPE_BOOLEAN, G_TYPE_STRING);
		if (gribouy_app->treeview)
			gtk_tree_view_set_model (gribouy_app->treeview, GTK_TREE_MODEL (gribouy_app->model));

		gribouy_main_update_model (
				gribouy_app->font->standard_encoding);

		gribouy_app->filename = g_strdup (filename);
	}
	else {
		error_str = g_strdup_printf (_("This file is not a font file."));
		dialog = GTK_DIALOG (gtk_message_dialog_new (GTK_WINDOW (gribouy_app), GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR,
						 GTK_BUTTONS_OK,
						 error_str));

		gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

		gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (GTK_WIDGET (dialog));
		return FALSE;
	}
	return TRUE;
}
static void
verb_LettersDelete_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;
	gint filled;
	gchar *letter;

	selection = gtk_tree_view_get_selection (gribouy_app->treeview);
	gtk_tree_selection_get_selected (selection, &model, &iter);
	gtk_tree_model_get (model, &iter, 0, &filled, 1, &letter, -1);

	if (filled) {
		gribouy_type1_remove (gribouy_app->font, letter);
		gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
	}
}

static void
verb_LettersNew_cb (BonoboUIComponent *uic, gpointer user_data, const char *cname)
{
	GladeXML *xml;

	xml = glade_xml_new (gribouy_app->fullname, "new_letter", NULL);
	glade_xml_signal_autoconnect (xml);
}

/* Creation of all the model */
static void
gribouy_main_create_model ()
{
	GtkCellRenderer *cell;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	GtkWidget *sw;

	/* We create the model : 2 columns containing : boolean / string */
	if (!gribouy_app->treeview) {
		gribouy_app->treeview = (GtkTreeView *)gtk_tree_view_new ();
		sw = gtk_scrolled_window_new (NULL, NULL);
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				GTK_POLICY_AUTOMATIC, 
				GTK_POLICY_AUTOMATIC);

		gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw),
				GTK_WIDGET (gribouy_app->treeview));
		bonobo_window_set_contents (BONOBO_WINDOW (gribouy_app->win),
				sw);

		cell = gtk_cell_renderer_toggle_new ();

		column = gtk_tree_view_column_new_with_attributes (_("Created"), cell,
								   "active", 0,
								   NULL);
		gtk_tree_view_append_column (gribouy_app->treeview,
					     GTK_TREE_VIEW_COLUMN (column));

		cell = gtk_cell_renderer_text_new ();

		column = gtk_tree_view_column_new_with_attributes (_("Letters"), cell,
								   "text", 1,
								   NULL);

		gtk_tree_view_append_column (gribouy_app->treeview,
					     GTK_TREE_VIEW_COLUMN (column));
		selection = gtk_tree_view_get_selection (gribouy_app->treeview);

		g_signal_connect (G_OBJECT (gribouy_app->treeview),
				"button_press_event", G_CALLBACK (on_treeview_button_press_event), NULL);
		g_signal_connect (G_OBJECT (selection), "changed",
				  G_CALLBACK (on_treeview_changed),
				  NULL);
		g_signal_connect (G_OBJECT (gribouy_app->treeview),
				"button_press_event",
				G_CALLBACK (on_treeview_doubleclick),
				selection);
	}

	if (gribouy_app->model) {
		g_object_unref (G_OBJECT (gribouy_app->model));
		gribouy_app->model = NULL;
	}

	gribouy_app->model = gtk_list_store_new (2, G_TYPE_BOOLEAN, G_TYPE_STRING);

	gtk_tree_view_set_model (gribouy_app->treeview, GTK_TREE_MODEL (gribouy_app->model));
	gtk_widget_show_all (GTK_WIDGET (gribouy_app->win));
}


void
on_clear_activate (GtkMenuItem *item, gpointer user_data)
{
}

static gboolean
event_LettersEncoding_cb (GtkCheckMenuItem *item, gpointer user_data)
{
	gboolean se;

	g_object_get (G_OBJECT (gribouy_app->font), "standard_encoding", &se, NULL);
	if (!se) {
		g_object_set (G_OBJECT (gribouy_app->font), "standard_encoding", TRUE, NULL);
		gribouy_main_create_model ();
		gribouy_main_update_model (TRUE);
	}
	else {
		g_object_set (G_OBJECT (gribouy_app->font), "standard_encoding", FALSE, NULL);
		gribouy_main_create_model (); 
		gribouy_main_update_model (FALSE);
	}
	return !se;
}

static void
gribouy_main_init_all ()
{
	if (gribouy_app->font)
		g_object_unref (gribouy_app->font);

	gribouy_app->font = GRIBOUY_TYPE1 (gribouy_type1_new ());

	gribouy_main_create_model ();
	gribouy_main_update_model (gribouy_app->font->standard_encoding);
}

static void
gribouy_die (GnomeClient *client, gpointer user_data)
{
	bonobo_main_quit ();
}

static gboolean gribouy_save_yourself (GnomeClient *client,
		gint phase,
		GnomeSaveStyle save_style,
		gboolean shutdown,
		GnomeInteractStyle interact_style,
		gboolean fast,
		gpointer user_data)
{
	/* to be completed... */

	return TRUE;
}

static gboolean
on_treeview_doubleclick (GtkTreeView *treeview, GdkEvent *event,
		gpointer user_data)
{
	GtkTreeIter *iter;
	GtkTreeModel *model;
	GtkTreeSelection *selection;
	gchar *letter;

	selection = GTK_TREE_SELECTION (user_data);
	if (event->type == GDK_2BUTTON_PRESS) {
		iter = g_new0 (GtkTreeIter, 1);
		if (gtk_tree_selection_get_selected (selection, &model, iter))
		{
			gtk_tree_model_get (model, iter, 1, &letter, -1);

/*			if (!gribouy_app->font->current || strcmp (gribouy_app->font->current, letter)) {*/
				gribouy_main_set_current (letter, iter);
/*			}
			else {
				g_free (iter);
				g_print (_("Already selected.\n"));
			}*/
			g_free (letter);
		}
		else
			g_free (iter);
	}
	return FALSE;
}

static void
on_letter_window_changed (GtkWidget *window, gboolean changed, gpointer user_data)
{
	GtkTreeIter *iter;

	iter = (GtkTreeIter *)user_data;
	gtk_list_store_set (gribouy_app->model,
			iter, 0, changed, -1);
	g_free (iter);
}

static void
create_window (GribouyApp *app)
{
	BonoboUIContainer *ui_container;
	BonoboUIComponent *ui_component;

	ui_component = bonobo_ui_component_new_default ();
	app->win = bonobo_window_new (PACKAGE, GETTEXT_PACKAGE);
	ui_container = bonobo_window_get_ui_container (
			BONOBO_WINDOW (app->win));
	bonobo_ui_component_set_container (
			ui_component,
			BONOBO_OBJREF (ui_container),
			NULL);
	bonobo_ui_util_set_ui (
			ui_component,
			DATA_DIR,
			"gribouy-ui.xml",
			PACKAGE,
			NULL);
	bonobo_ui_component_add_verb_list_with_data (
			ui_component,
			gribouy_app_verbs,
			gribouy_app->win);
	g_signal_connect (G_OBJECT (gribouy_app->win),
			"destroy",
			G_CALLBACK (bonobo_main_quit),
			NULL);
	app->treeview = NULL;
	gribouy_main_init_all ();

	if (gribouy_app->filename)
		gribouy_main_open_file (gribouy_app->filename);
	gtk_widget_show_all (app->win);
}

int
main (int argc, char *argv[])
{
	const char **args;
	poptContext popt_context;
	GnomeClient *client;
	GnomeProgram *gribouy;
	BonoboUIContainer *ui_container;
	BonoboUIComponent *ui_component;
	GValue value = {0,};
	int i;

	static const struct poptOption options [] = {
		{NULL, '\0', 0, NULL, 0, NULL, NULL}
	};

	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	gribouy_app = g_new0 (GribouyApp, 1);

	gribouy = gnome_program_init (GETTEXT_PACKAGE, VERSION,
			LIBGNOMEUI_MODULE, argc, argv,
			GNOME_PARAM_POPT_TABLE, options,
			GNOME_PARAM_APP_DATADIR, PACKAGE_DATA_DIR,
			NULL);

	g_object_get (gribouy, GNOME_PARAM_POPT_CONTEXT, &popt_context, NULL);

	args = poptGetArgs (popt_context);

	if (args) {
		if (args[0] && !args[1])
			gribouy_app->filename = g_strdup (args[0]);
		else {
			gribouy_app->filename = NULL;
			g_error (_("Only one font can be opened."));
		}
	}
	poptFreeContext (popt_context);

	client = gnome_master_client ();
	g_signal_connect (G_OBJECT (client), "save_yourself",
			G_CALLBACK (gribouy_save_yourself), NULL);

	g_signal_connect (G_OBJECT (client), "die",
			G_CALLBACK (gribouy_die), NULL);

	gnome_window_icon_set_default_from_file (PACKAGE_ICONDIR"/gribouy.png");
	gribouy_app->fullname = g_strdup (PACKAGE_DATA_UI"/gribouy.glade");
	create_window (gribouy_app);

	bonobo_main ();
	return 0;
}


