/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gtk/gtksignal.h>
#include <gtk/gtkmarshal.h>
#include <gtk/gtk.h>

#include "gribouy-object.h"

enum {
	PROP_0,
	PROP_EDITABLE,
	PROP_SELECTED,
	PROP_CENTER,
	PROP_WINDOW,
	PROP_CURSOR,
	PROP_CHILDREN
};

enum {
	CHANGED_SIGNAL,
	LAST_SIGNAL
};

static guint gribouy_object_signals[LAST_SIGNAL] = { 0 };

static void gribouy_object_init (GribouyObject *obj);
static void gribouy_object_class_init (GribouyObjectClass *class);
static void gribouy_object_set_property (GObject *object, guint property_id,
		const GValue *value, GParamSpec *pspec);
static void gribouy_object_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec);
static gint gribouy_object_event (GribouyObject *obj, GdkEvent *event,
		GnomeCanvasItem *center);
static void gribouy_object_only_select (GribouyObject *obj);
static void gribouy_object_changed (GribouyObject *obj);
static void gribouy_object_grid_translation (GribouyObject *obj,
		double tx, double ty,
		double *rx, double *ry);

static GnomeCanvasGroupClass *parent_class = NULL;

GType
gribouy_object_get_type (void)
{
	static GType gribouy_object_type = 0;

	if (!gribouy_object_type) {
		static const GTypeInfo gribouy_object_info = {
			sizeof (GribouyObjectClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_object_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyObject),
			0,
			(GInstanceInitFunc) gribouy_object_init
		};

		gribouy_object_type = g_type_register_static (
				GNOME_TYPE_CANVAS_GROUP,
				"GribouyObject", &gribouy_object_info, 0);
	}
	return gribouy_object_type;
}

static void 
gribouy_object_class_init (GribouyObjectClass *class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_object_set_property;
	g_object_class->get_property = gribouy_object_get_property;

	gribouy_object_signals[CHANGED_SIGNAL] =
		g_signal_new ("changed",
				G_TYPE_FROM_CLASS (class),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET (GribouyObjectClass, changed),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE,
				0);

	g_object_class_install_property
		(g_object_class,
		 PROP_EDITABLE,
		 g_param_spec_boolean ("editable", NULL, NULL,
			 TRUE,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_SELECTED,
		 g_param_spec_boolean ("selected", NULL, NULL,
			 FALSE,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_CENTER,
		 g_param_spec_object ("center", NULL, NULL,
			 gnome_canvas_item_get_type (),
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_WINDOW,
		 g_param_spec_object ("window", NULL, NULL,
			 gribouy_letter_window_get_type (),
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_CHILDREN,
		 g_param_spec_pointer ("children", NULL, NULL,
			 (G_PARAM_READABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_CURSOR,
		 g_param_spec_int ("cursor", NULL, NULL,
			 -1, 200, 2,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	class->only_select = gribouy_object_only_select;
	class->changed = gribouy_object_changed;
}

static void 
gribouy_object_init (GribouyObject *obj)
{
	obj->editable = TRUE;
	obj->selected = FALSE;
}

static void
gribouy_object_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec)
{
	GribouyObject *obj;
	ArtPoint *point;

	obj = GRIBOUY_OBJECT (object);
	switch (property_id) {
		case PROP_EDITABLE:
			g_value_set_boolean (value, obj->editable);
			break;
		case PROP_SELECTED:
			g_value_set_boolean (value, obj->selected);
			break;
		case PROP_CENTER:
			g_value_set_object (value, obj->center);
			break;
		case PROP_WINDOW:
			g_value_set_object (value, obj->window);
			break;
		case PROP_CHILDREN:
			g_value_set_pointer (value, GNOME_CANVAS_GROUP (object)->item_list->next);
			break;
		case PROP_CURSOR:
			g_value_set_int (value, obj->cursor);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (obj,
					property_id, pspec);
			break;
	}
}

static void
gribouy_object_set_property (GObject *object, guint param_id,
		const GValue *value, GParamSpec *pspec)
{
	GribouyObject *obj;
	ArtPoint *point;

	obj = GRIBOUY_OBJECT (object);

	switch (param_id) {
		case PROP_EDITABLE:
			obj->editable = g_value_get_boolean (value);
			if (GRIBOUY_OBJECT_GET_CLASS (obj)->edit)
				GRIBOUY_OBJECT_GET_CLASS (obj)->edit (obj, obj->editable);
			break;
		case PROP_SELECTED:
			obj->selected = g_value_get_boolean (value);
			if (GRIBOUY_OBJECT_GET_CLASS (obj)->select)
				GRIBOUY_OBJECT_GET_CLASS (obj)->select (obj, obj->selected);
			break;
		case PROP_CENTER:
			obj->center = g_value_get_object (value);
			if (obj->center)
				g_signal_connect_swapped ((GObject *)obj->center,
						"event",
						G_CALLBACK (gribouy_object_event),
						obj);
			break;
		case PROP_WINDOW:
			obj->window = g_value_get_object (value);
			break;
		case PROP_CURSOR:
			obj->cursor = g_value_get_int (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (obj,
					param_id, pspec);
			break;
	}
}

static gint
gribouy_object_event (GribouyObject *obj, GdkEvent *event, GnomeCanvasItem *center)
{
	static gboolean dragging;
	GnomeCanvasItem *item;
	double item_x, item_y, tx, ty;
	GdkCursor *X;
	gboolean magnetic;

	item = (GnomeCanvasItem *)obj;

	g_return_if_fail (GRIBOUY_IS_LETTER_WINDOW (obj->window));

	item_x = event->button.x;
	item_y = event->button.y;

	gnome_canvas_item_w2i (item->parent, &item_x, &item_y);

	switch (event->type) {
	case GDK_ENTER_NOTIFY:
		X = gdk_cursor_new (obj->cursor);
		gnome_canvas_item_grab (center,
				GDK_POINTER_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_LEAVE_NOTIFY_MASK,
				X,
				event->button.time);
		gdk_cursor_destroy (X);
		break;
	case GDK_LEAVE_NOTIFY:
		gnome_canvas_item_ungrab (center, event->button.time);
		break;
	case GDK_BUTTON_PRESS:
		obj->clk_x = item_x;
		obj->clk_y = item_y;

		if (event->button.button == 1 || event->button.button == 2) {
			if (event->button.state & GDK_CONTROL_MASK)
				g_object_set (G_OBJECT (obj),
						"selected", obj->selected?FALSE:TRUE,
						NULL);
			else
				GRIBOUY_OBJECT_GET_CLASS (obj)->only_select (obj);

			if (obj->editable)
				dragging = TRUE;

		}
		if (event->button.button == 2 &&
				GRIBOUY_OBJECT_GET_CLASS (obj)->open_menu)
			GRIBOUY_OBJECT_GET_CLASS (obj)->open_menu (obj, event);
		break;
	case GDK_MOTION_NOTIFY:
		if (dragging && (event->motion.state & GDK_BUTTON1_MASK)) {
			if (GRIBOUY_OBJECT_GET_CLASS (obj)->move_selection) {
				gribouy_object_grid_translation (obj,
						item_x - obj->clk_x,
						item_y - obj->clk_y,
						&tx, &ty);
				GRIBOUY_OBJECT_GET_CLASS (obj)->move_selection (
						obj,
						tx, ty);
			}
			g_signal_emit (G_OBJECT (obj), gribouy_object_signals[CHANGED_SIGNAL], 0);
			obj->clk_x += tx;
			obj->clk_y += ty;
			return FALSE;
		}
		break;
	case GDK_BUTTON_RELEASE:
		dragging = FALSE;
		break;
	}
	return FALSE;
}

/**
 * gribouy_object_grid_translation:
 * @obj; A gribouy object.
 * @tx: X coordinate of the translation in the object's coordinates system.
 * @ty: Y coordinate of the translation in the object's coordinates system.
 * @rx: X coordinate of the translation to align the object on the grid.
 * @ry: Y coordinate of the translation to align the object on the grid.
 *
 * Optimizes the given translation to align the given object on the grid.
 **/
static void
gribouy_object_grid_translation (GribouyObject *obj, double tx, double ty,
		double *rx, double *ry)
{
	gint grid_x, grid_y;
	ArtPoint p, p1, p2;
	double trx, try;
	double i2w[6], w2i[6];
	gboolean magnetic;

	g_object_get (G_OBJECT (obj->window->grid),
			"grid_x", &grid_x,
			"grid_y", &grid_y,
			"magnetic", &magnetic,
			NULL);

	if (!magnetic) {
		*rx = tx;
		*ry = ty;
		return;
	}

	g_object_get (G_OBJECT (obj),
			"x", &p.x,
			"y", &p.y,
			NULL);
	gnome_canvas_item_i2w_affine (GNOME_CANVAS_ITEM (obj)->parent, i2w);
	art_affine_point (&p1, &p, i2w); /* returns p1 in world's coord. */

	trx = i2w[0] * tx + i2w[2] * ty;
	try = i2w[1] * tx + i2w[3] * ty; /* translation in world's coord. */

	p1.x = ROUND ((p1.x + trx) / grid_x) * grid_x;
	p1.y = ROUND ((p1.y + try) / grid_y) * grid_y;

	art_affine_invert (w2i, i2w);

	art_affine_point (&p2, &p1, w2i);

	*rx = p2.x - p.x;
	*ry = p2.y - p.y;
}

static void
gribouy_object_only_select (GribouyObject *obj)
{
	GribouyObject *obj1;
	GList *lst;

	if (obj->selected)
		return;
	obj1 = (GribouyObject *) GNOME_CANVAS_ITEM (obj)->parent;
	lst = GNOME_CANVAS_GROUP (obj1)->item_list->next;
	while (lst) {
		obj1 = lst->data;
		g_object_set (G_OBJECT (obj1), "selected", (obj1 == obj), NULL);
		lst = lst->next;
	}
}

static void
gribouy_object_changed (GribouyObject *obj)
{
	if (GRIBOUY_IS_OBJECT (GNOME_CANVAS_ITEM (obj)->parent))
		g_signal_emit (G_OBJECT (GNOME_CANVAS_ITEM (obj)->parent),
				gribouy_object_signals[CHANGED_SIGNAL],
				0);
}

void
gribouy_object_select_children (GribouyObject *obj, GnomeCanvasRect *rect, gboolean keep, gboolean descendance)
{
	GList *lst;
	double x1, y1, x2, y2;
	double xr1, yr1, xr2, yr2;

	g_return_if_fail (GRIBOUY_IS_OBJECT (obj));
	g_return_if_fail (GRIBOUY_IS_LETTER_WINDOW (obj->window));
	g_return_if_fail (GNOME_IS_CANVAS_ITEM (obj->center));

	if (rect)
		gnome_canvas_item_get_bounds (GNOME_CANVAS_ITEM (rect),
				&xr1, &yr1, &xr2, &yr2);
	else
		xr1 = xr2 = yr1 = yr2 = 0.;
	gnome_canvas_item_get_bounds (GNOME_CANVAS_ITEM (obj->center),
			&x1, &y1, &x2, &y2);
	gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (obj->center)->parent,
			&x1, &y1);
	gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (obj->center)->parent,
			&x2, &y2);

	if (xr1 <= x1 && xr2 >= x2 && yr1 <= y1 && yr2 >= y2)
		g_object_set (G_OBJECT (obj), "selected",
				(!keep)?TRUE:(!obj->selected), NULL);
	else if (!keep)
		g_object_set (G_OBJECT (obj), "selected", FALSE, NULL);

	if (descendance) {
		g_object_get (G_OBJECT (obj), "children", &lst, NULL);
		while (lst) {
			gribouy_object_select_children (lst->data, rect,
					keep, descendance);
			lst = lst->next;
		}
	}
}
