/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __GRIBOUY_OBJECT_H__
#define __GRIBOUY_OBJECT_H__

#include <gdk/gdk.h>
#include <gtk/gtkobject.h>
#include <libgnomecanvas/libgnomecanvas.h>
#include <math.h>
#include "gribouy-letter-window.h"

#define epsilon		1.
#define near(a, b)	((a)-(b) < 0.5 && (b)-(a) < 0.5)
#define ROUND(x)	floor((x) + 0.5)

#define GRIBOUY_TYPE_OBJECT	(gribouy_object_get_type())
#define GRIBOUY_OBJECT(obj)	G_TYPE_CHECK_INSTANCE_CAST((obj), gribouy_object_get_type(), GribouyObject)
#define GRIBOUY_OBJECT_CONST(obj)	G_TYPE_CHECK_INSTANCE_CAST((obj), gribouy_object_get_type(), GribouyObject const)
#define GRIBOUY_OBJECT_CLASS(klass)	G_TYPE_CHECK_CLASS_CAST((klass), gribouy_object_get_type(), GribouyObjectClass)
#define GRIBOUY_IS_OBJECT(obj)	G_TYPE_CHECK_INSTANCE_TYPE((obj), gribouy_object_get_type ())

#define GRIBOUY_OBJECT_GET_CLASS(obj)	G_TYPE_INSTANCE_GET_CLASS((obj), gribouy_object_get_type(), GribouyObjectClass)

typedef struct _GribouyObject GribouyObject;
typedef struct _GribouyObjectClass GribouyObjectClass;

struct _GribouyObject {
	GnomeCanvasGroup parent;
	GribouyLetterWindow *window;
	GnomeCanvasItem *center;
	GdkCursorType cursor;
	gdouble clk_x, clk_y;

	guint editable : 1;
	guint selected : 1;
};

struct _GribouyObjectClass {
	GnomeCanvasGroupClass parent_class;

	void (* select) (GribouyObject *obj, gboolean v);
	void (* only_select) (GribouyObject *obj);
	void (* edit) (GribouyObject *obj, gboolean v);
	void (* move_selection) (GribouyObject *obj, gdouble x, gdouble y);
//	void (* update_parent) (GribouyObject *obj);
	void (* open_menu) (GribouyObject *obj, GdkEvent *event);
	void (* changed);
};


GType gribouy_object_get_type (void) G_GNUC_CONST;
void gribouy_object_select_children (GribouyObject *obj, GnomeCanvasRect *rect, gboolean keep, gboolean descendance);
#endif
