/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <gtk/gtksignal.h>
#include <gtk/gtkmarshal.h>

#include "gribouy-stem.h"

enum {
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_HORIZ,
};

static void gribouy_stem_init (GribouyStem * stem);
static void gribouy_stem_class_init (GribouyStemClass * klass);
static void gribouy_stem_set_property (GObject *object, guint property_id,
					const GValue *value,
					GParamSpec *pspec);
static void gribouy_stem_get_property (GObject *object, guint property_id,
					GValue *value, GParamSpec *pspec);
static void gribouy_stem_moved (GribouyStem *stem, gint pos, gint index);
static void gribouy_stem_newpos (GribouyStem *stem);

static GnomeCanvasRectClass *parent_class = NULL;
/* static guint stem_signals[LAST_SIGNAL] = { 0 }; */

GtkType
gribouy_stem_get_type (void)
{
	static GtkType stem_type = 0;

	if (!stem_type) {
		static const GtkTypeInfo stem_info = {
			"GribouyStem",
			sizeof (GribouyStem),
			sizeof (GribouyStemClass),
			(GtkClassInitFunc) gribouy_stem_class_init,
			(GtkObjectInitFunc) gribouy_stem_init,
			NULL,
			NULL,
			(GtkClassInitFunc) NULL
		};

		stem_type = gtk_type_unique (gnome_canvas_rect_get_type (), &stem_info);
	}

	return stem_type;
}

static void 
gribouy_stem_init (GribouyStem *stem)
{
}

static void 
gribouy_stem_class_init (GribouyStemClass * klass)
{
	GObjectClass *g_object_class;
	GtkObjectClass *object_class;

	g_object_class = (GObjectClass *) klass;
	object_class = (GtkObjectClass *) klass;

	parent_class = gtk_type_class (gtk_object_get_type ());

	g_object_class->set_property = gribouy_stem_set_property;
	g_object_class->get_property = gribouy_stem_get_property;

	g_object_class_install_property
		(g_object_class,
		 PROP_HORIZ,
		 g_param_spec_boolean ("horiz", NULL, NULL,
				       TRUE,
				       (G_PARAM_READABLE)));
}

GnomeCanvasItem * 
gribouy_stem_new (GnomeCanvasGroup *group, gboolean horiz, double y1, double y2)
{
	GribouyStem *stem;
	double xmin, xmax, ymin, ymax;

/* 	gnome_canvas_get_scroll_region (GNOME_CANVAS_ITEM (group)->canvas, */
/* 					&xmin, &ymin, */
/* 					&xmax, &ymax); */
	xmin = -1000.;
	ymin = -1000;;
	xmax = 1000.;
	ymax = 1000.;

	if (horiz) {
		if (y1 < y2) {
			ymin = y1;
			ymax = y2;
		}
		else {
			ymin = y2;
			ymax = y1;
		}
	}
	else {
		if (y1 < y2) {
			xmin = y1;
			xmax = y2;
		}
		else {
			xmin = y2;
			xmax = y1;
		}
	}
	stem = GRIBOUY_STEM
		(gnome_canvas_item_new (group,
					gribouy_stem_get_type (),
					"x1", xmin,
					"y1", ymin,
					"x2", xmax,
					"y2", ymax,
					"fill_color_rgba", 0xccddff50,
					NULL));
	stem->horiz = horiz;
	return (GnomeCanvasItem *)stem;
}

static void
gribouy_stem_get_property (GObject *object,
			    guint property_id,
			    GValue *value,
			    GParamSpec *pspec)
{
	GnomeCanvasRE *rect;
	GribouyStem *stem;

	rect = GNOME_CANVAS_RE (object);
	stem = GRIBOUY_STEM (object);

	switch (property_id) {
	case PROP_HORIZ:
		g_value_set_boolean (value, stem->horiz);
		break;
/* 	case PROP_X: */
/* 		g_value_set_double (value, stem->x); */
/* 		break; */
/* 	case PROP_Y: */
/* 		g_value_set_double (value, stem->y); */
/* 		break; */
/* 	case PROP_EDITABLE: */
/* 		g_value_set_boolean (value, stem->editable); */
/* 		break; */
/* 	case PROP_POS: */
/* 		g_value_set_uint (value, stem->pos); */
/* 		break; */
/* 	case PROP_INDEX: */
/* 		g_value_set_uint (value, stem->index); */
/* 		break; */
/* 	case PROP_LETTER: */
/* 		g_value_set_pointer (value, stem->letter); */
/* 		break; */
/* 	case PROP_LINE: */
/* 		g_value_set_pointer (value, stem->line); */
/* 		break; */
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
	return;
}

static void
gribouy_stem_set_property (GObject *object,
			    guint param_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	GnomeCanvasRE *rect;
	GribouyStem *stem;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GRIBOUY_IS_STEM (object));

	rect = GNOME_CANVAS_RE (object);
	stem = GRIBOUY_STEM (object);

	switch (param_id) {
/* 	case PROP_X: */
/* 		stem->x = g_value_get_double (value); */
/* 		rect->x1 = stem->x - 5.; */
/* 		rect->x2 = stem->x + 5.; */
/* 		rect->path_dirty = 1; */
/* 		gtk_signal_emit (GTK_OBJECT (stem), stem_signals[MOVED]); */
/* 		gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (stem)); */
/* 		break; */
/* 	case PROP_Y: */
/* 		stem->y = g_value_get_double (value); */
/* 		rect->y1 = stem->y - 5.; */
/* 		rect->y2 = stem->y + 5.; */
/* 		rect->path_dirty = 1; */
/* 		gtk_signal_emit (GTK_OBJECT (stem), stem_signals[MOVED]); */
/* 		gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (stem)); */
/* 		break; */
/* 	case PROP_EDITABLE: */
/* 		if ((stem->editable = g_value_get_boolean (value))) */
/* 			gnome_canvas_item_show (GNOME_CANVAS_ITEM (stem)); */
/* 		else */
/* 			gnome_canvas_item_hide (GNOME_CANVAS_ITEM (stem)); */
/* 		break; */
/* 	case PROP_INDEX: */
/* 		stem->index = g_value_get_uint (value); */
/* 		break; */
/* 	case PROP_LETTER: */
/* 		stem->letter = g_value_get_pointer (value); */
/* 		break; */
/* 	case PROP_LINE: */
/* 		stem->line = g_value_get_pointer (value); */
/* 		break; */
/* 	case PROP_POS: */
/* 		{ */
/* 			int color; */

/* 			stem->pos = g_value_get_uint (value); */
/* 			switch (stem->pos) { */
/* 			case 1: */
/* 			case 2: */
/* 				color = 0xffaaff80; */
/* 				break; */
/* 			case 3: */
/* 				color = 0xaaffff80; */
/* 				break; */
/* 			default: */
/* 				g_print ("bad position for this stem !!!\n"); */
/* 			} */
/* 			gnome_canvas_item_set (GNOME_CANVAS_ITEM (stem), */
/* 					       "fill_color_rgba", color, NULL); */
/* 		} */
/* 		gtk_signal_emit (GTK_OBJECT (stem), stem_signals[NEWPOS]); */
/* 		gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (stem)); */
/* 		break; */
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
		break;
	}
}

/* static void */
/* gribouy_stem_moved (GribouyStem *stem, gint pos, gint index) */
/* { */
/* 	g_return_if_fail (GRIBOUY_IS_STEM (stem)); */
/* } */

/* static void */
/* gribouy_stem_newpos (GribouyStem *stem) */
/* { */
/* 	g_return_if_fail (GRIBOUY_IS_STEM (stem)); */
/* } */
