/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "gribouy-cursor.h"
#include "gribouy-triplet.h"

enum {
	PROP_0,
	PROP_TANGENT1,
	PROP_TANGENT2,
	PROP_LINKED,
	PROP_INDEX
};

static void gribouy_triplet_init (GribouyTriplet *triplet);
static void gribouy_triplet_class_init (GribouyTripletClass *class);
static void gribouy_triplet_edit (GribouyObject *obj, gboolean v);
static void gribouy_triplet_select (GribouyObject *obj, gboolean v);
static void gribouy_triplet_move_selection (GribouyObject *obj,	gdouble x, gdouble y);
static void gribouy_triplet_open_menu (GribouyObject *obj, GdkEvent *event);
static void gribouy_triplet_link_tangents_cb (GribouyTriplet *triplet,
		guint action, GtkWidget *item);
static void gribouy_triplet_erase_cb (GribouyTriplet *triplet,
		guint action, GtkWidget *item);

static void gribouy_triplet_set_property (GObject *object, guint property_id,
					const GValue *value,
					GParamSpec *pspec);
static void gribouy_triplet_get_property (GObject *object, guint property_id,
					GValue *value, GParamSpec *pspec);
static void gribouy_triplet_destroy (GtkObject *object);

static GribouyObjectClass *parent_class = NULL;

GType
gribouy_triplet_get_type (void)
{
	static GType gribouy_triplet_type = 0;

	if (!gribouy_triplet_type) {
		static const GTypeInfo gribouy_triplet_info = {
			sizeof (GribouyTripletClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_triplet_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyTriplet),
			0,
			(GInstanceInitFunc) gribouy_triplet_init
		};

		gribouy_triplet_type = g_type_register_static (
				gribouy_object_get_type (),
				"GribouyTriplet", &gribouy_triplet_info,
				(GTypeFlags)0);
	}
	return gribouy_triplet_type;
}

static void 
gribouy_triplet_class_init (GribouyTripletClass *class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	GribouyObjectClass *obj_class = (GribouyObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_triplet_set_property;
	g_object_class->get_property = gribouy_triplet_get_property;

	g_object_class_install_property
		(g_object_class,
		 PROP_LINKED,
		 g_param_spec_boolean ("linked", NULL, NULL,
			 FALSE,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_INDEX,
		 g_param_spec_int ("index", NULL, NULL,
			 -1, G_MAXINT, 0,
			 (G_PARAM_WRITABLE | G_PARAM_READABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_TANGENT1,
		 g_param_spec_pointer ("tangent1", NULL, NULL,
			 (G_PARAM_WRITABLE | G_PARAM_READABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_TANGENT2,
		 g_param_spec_pointer ("tangent2", NULL, NULL,
			 (G_PARAM_WRITABLE | G_PARAM_READABLE)));

	object_class->destroy = gribouy_triplet_destroy;
	obj_class->select = gribouy_triplet_select;
	obj_class->edit = gribouy_triplet_edit;
	obj_class->move_selection = gribouy_triplet_move_selection;
	obj_class->open_menu = gribouy_triplet_open_menu;
}

static void 
gribouy_triplet_init (GribouyTriplet *triplet)
{
	triplet->ctrl1 = NULL;
	triplet->ctrl2 = NULL;
}

static void
gribouy_triplet_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec)
{
	GribouyTriplet *triplet;
	ArtPoint *point;

	triplet = GRIBOUY_TRIPLET (object);
	switch (property_id) {
		case PROP_LINKED:
			g_value_set_boolean (value, triplet->linked);
			break;
		case PROP_INDEX:
			g_value_set_int (value, triplet->index);
			break;
		case PROP_TANGENT1:
			if (triplet->ctrl1) {
				point = g_new (ArtPoint, 1);
				g_object_get (G_OBJECT (triplet->ctrl1),
						"x", &point->x,
						"y", &point->y, NULL);
				gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (triplet),
						&point->x, &point->y);
				gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (triplet)->parent,
						&point->x, &point->y);
				g_value_set_pointer (value, point);
			}
			else {
				point = NULL;
				g_value_set_pointer (value, point);
			}
			break;
		case PROP_TANGENT2:
			if (triplet->ctrl2) {
				point = g_new (ArtPoint, 1);
				g_object_get (G_OBJECT (triplet->ctrl2),
						"x", &point->x,
						"y", &point->y, NULL);
				gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (triplet),
						&point->x, &point->y);
				gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (triplet)->parent,
						&point->x, &point->y);
				g_value_set_pointer (value, point);
			}
			else {
				point = NULL;
				g_value_set_pointer (value, point);
			}
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					property_id, pspec);
			break;
	}
}

static void
gribouy_triplet_set_property (GObject *object, guint param_id,
		const GValue *value, GParamSpec *pspec)
{
	GribouyTriplet *triplet;
	GnomeCanvasGroup *curve;
	GList *lst, *lst1;
	gint index;
	ArtPoint *point;
	ArtPoint p;

	triplet = GRIBOUY_TRIPLET (object);
	curve = GNOME_CANVAS_GROUP (GNOME_CANVAS_ITEM (object)->parent);

	switch (param_id) {
		case PROP_TANGENT1:
			if (triplet->ctrl1)
				gtk_object_destroy (GTK_OBJECT (triplet->ctrl1));
			point = g_value_get_pointer (value);
			if (point) {
				p.x = point->x;
				p.y = point->y;
				gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (curve),
						&p.x, &p.y);
				gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (triplet),
						&p.x, &p.y);
				triplet->ctrl1 = gribouy_cursor_new (
						GRIBOUY_OBJECT (triplet),
						p.x,
						p.y, 1);
				gnome_canvas_item_lower_to_bottom (GNOME_CANVAS_ITEM (triplet->ctrl1));
			}
			else
				triplet->ctrl1 = NULL;

			break;
		case PROP_TANGENT2:
			if (triplet->ctrl2)
				gtk_object_destroy (GTK_OBJECT (triplet->ctrl2));
			point = g_value_get_pointer (value);
			if (point) {
				p.x = point->x;
				p.y = point->y;
				gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (curve),
						&p.x, &p.y);
				gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (triplet),
						&p.x, &p.y);
				triplet->ctrl2 = gribouy_cursor_new (
						GRIBOUY_OBJECT (triplet),
						p.x,
						p.y, 1);
				gnome_canvas_item_lower_to_bottom (GNOME_CANVAS_ITEM (triplet->ctrl2));
			}
			else
				triplet->ctrl2 = NULL;

			break;
		case PROP_LINKED:
			if (triplet->ctrl1 && triplet->ctrl2)
				if ((triplet->linked = g_value_get_boolean (value))) {
					GRIBOUY_OBJECT_GET_CLASS (G_OBJECT (triplet->ctrl1))->move_selection (GRIBOUY_OBJECT (triplet->ctrl1), 0, 0);
					g_signal_emit_by_name (G_OBJECT (triplet->ctrl1), "changed", 0);
				}
			break;
		case PROP_INDEX:
			index = g_value_get_int (value);
			if (index == -1) {
				lst = curve->item_list->next;
				lst1 = curve->item_list_end;
				if (lst == NULL)
					break;
				if (lst->data == lst1->data)
					triplet->index = 0;
				else {
					lst = lst1->prev;
					triplet->index = GRIBOUY_TRIPLET (lst->data)->index + 1;
				}
			}
			else {
				lst = curve->item_list->next;
				while (lst) {
					if (triplet == lst->data)
						triplet->index = index;
					else if (GRIBOUY_TRIPLET (lst->data)->index >= index)
						((GribouyTriplet *)(lst->data))->index ++;
					lst = lst->next;
				}
				lst = curve->item_list->next;
			}
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					param_id, pspec);
			break;
	}
}

static void
gribouy_triplet_select (GribouyObject *obj, gboolean v)
{
	GribouyTriplet *triplet;

	triplet = (GribouyTriplet *) obj;
	g_object_set (G_OBJECT (obj->center),
			"selected", v, NULL);
	if (triplet->ctrl1)
		g_object_set (G_OBJECT (triplet->ctrl1),
				"selected", v, NULL);
	if (triplet->ctrl2)
		g_object_set (G_OBJECT (triplet->ctrl2),
				"selected", v, NULL);
}

static void
gribouy_triplet_edit (GribouyObject *obj, gboolean v)
{
	if (v)
		gnome_canvas_item_show (GNOME_CANVAS_ITEM (obj));
	else
		gnome_canvas_item_hide (GNOME_CANVAS_ITEM (obj));
}

GribouyObject *
gribouy_triplet_new (GribouyObject *curve, gint index, gdouble x, gdouble y)
{
	GribouyObject *triplet;
	GribouyObject *center;

	g_return_if_fail (GRIBOUY_IS_LETTER_WINDOW (curve->window));
	triplet = GRIBOUY_OBJECT (
		gnome_canvas_item_new (GNOME_CANVAS_GROUP (curve),
				       gribouy_triplet_get_type (),
				       "x", x,
				       "y", y,
				       NULL));
	g_object_set (G_OBJECT (triplet), "window", curve->window, NULL);
	center = gribouy_cursor_new (GRIBOUY_OBJECT (triplet), 0, 0, 0);
	g_object_set (G_OBJECT (triplet),
			"center", center,
			"index", index,
			NULL);

	return triplet;
}

GribouyObject *
gribouy_triplet_new_with_tangents (GribouyObject *curve, gint index,
		gdouble ax, gdouble ay, gdouble x, gdouble y, gdouble px, gdouble py)
{
	GribouyObject *triplet;
	ArtPoint p1, p2;

	triplet = gribouy_triplet_new (curve, index, x, y);

	p1.x = ax;
	p1.y = ay;
	p2.x = px;
	p2.y = py;
	g_object_set (G_OBJECT (triplet),
			"tangent1", &p1,
			"tangent2", &p2,
			NULL);

	return triplet;
}

GribouyObject *
gribouy_triplet_new_with_tangent1 (GribouyObject *curve, gint index,
		gdouble ax, gdouble ay, gdouble x, gdouble y)
{
	GribouyObject *triplet;
	ArtPoint p;

	triplet = gribouy_triplet_new (curve, index, x, y);

	p.x = ax;
	p.y = ay;
	g_object_set (G_OBJECT (triplet), "tangent1", &p, NULL);

	return triplet;
}

GribouyObject *
gribouy_triplet_new_with_tangent2 (GribouyObject *curve, gint index,
		gdouble x, gdouble y, gdouble px, gdouble py)
{
	GribouyObject *triplet;
	ArtPoint p;

	triplet = gribouy_triplet_new (curve, index, x, y);

	p.x = px;
	p.y = py;
	g_object_set (G_OBJECT (triplet), "tangent2", &p, NULL);

	return triplet;
}

/* Destroy handler for Gribouy triplet */
static void
gribouy_triplet_destroy (GtkObject *object)
{
	GribouyTriplet *triplet;

	g_return_if_fail (GRIBOUY_IS_TRIPLET (object));

	triplet = GRIBOUY_TRIPLET (object);

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gribouy_triplet_move_selection (GribouyObject *triplet, gdouble x, gdouble y)
{
	GnomeCanvasGroup *curve;
	GList *lst;
	GribouyObject *c;

	curve = GNOME_CANVAS_GROUP ((GNOME_CANVAS_ITEM (triplet)->parent));
	lst = curve->item_list->next;
	while (lst) {
		c = lst->data;
		lst = lst->next;
		if (c->selected)
			gnome_canvas_item_move (GNOME_CANVAS_ITEM (c), x, y);
	}
}

static void
gribouy_triplet_open_menu (GribouyObject *obj, GdkEvent *event)
{
	GtkItemFactory *factory;
	GtkAccelGroup *accel;
	GtkWidget *menu;
	GtkItemFactoryEntry menu_items[] = {
		{ "/_Link tangents",
			NULL,
			gribouy_triplet_link_tangents_cb, 1,
			"<CheckItem>" },
		{ "/_Erase",
			NULL,
			gribouy_triplet_erase_cb, 1,
			"<StockItem>", GTK_STOCK_CLEAR },
	};

	accel = gtk_accel_group_new ();
	factory = gtk_item_factory_new (GTK_TYPE_MENU,
			"<control>", accel);
	gtk_item_factory_create_items (factory, 2, menu_items, obj);
	menu = gtk_item_factory_get_widget (factory, "<control>");

	if (menu) {
		gtk_check_menu_item_set_active (
				GTK_CHECK_MENU_ITEM (
					gtk_item_factory_get_item (
						factory, "/Link tangents")),
				GRIBOUY_TRIPLET (obj)->linked);
		gtk_widget_show (menu);
		gtk_menu_popup (GTK_MENU (menu),
				NULL, NULL, NULL, NULL,
				2, event->button.time);
	}
}

static void
gribouy_triplet_link_tangents_cb (GribouyTriplet *triplet, guint action, GtkWidget *widget)
{
	GtkCheckMenuItem *item;

	item = GTK_CHECK_MENU_ITEM (widget);
	g_object_set (G_OBJECT (triplet), "linked",
			gtk_check_menu_item_get_active (item), NULL);
}

static void
gribouy_triplet_erase_cb (GribouyTriplet *triplet, guint action,
		GtkWidget *item)
{
	GList *lst;
	GnomeCanvasGroup *curve;
	GribouyObject *trp;
	gint index;

	curve = GNOME_CANVAS_GROUP (GNOME_CANVAS_ITEM (triplet)->parent);
	lst = curve->item_list->next;
	index = 0;
	while (lst) {
		trp = lst->data;
		lst = lst->next;
		if (trp->selected)
			gtk_object_destroy (GTK_OBJECT (trp));
		else
			GRIBOUY_TRIPLET (trp)->index = index ++;
	}
	lst = curve->item_list->next;
	if (lst)
		g_signal_emit_by_name (G_OBJECT (lst->data), "changed", 0);
	else
		gtk_object_destroy (GTK_OBJECT (curve));
}

void
gribouy_triplet_destroy_tangent1 (GribouyTriplet *triplet)
{
	if (triplet->ctrl1) {
		gtk_object_destroy (GTK_OBJECT (triplet->ctrl1));
		triplet->ctrl1 = NULL;
	}
}

void
gribouy_triplet_destroy_tangent2 (GribouyTriplet *triplet)
{
	if (triplet->ctrl2) {
		gtk_object_destroy (GTK_OBJECT (triplet->ctrl2));
		triplet->ctrl2 = NULL;
	}
}
