/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <libxml/parser.h>

#include "gribouy-type1.h"

static guint buffer_size = 0;
static guchar *buffer, *buf;
static int tool;

static char init_xmlfont[] = "<?xml version=\"1.0\"?>\n<Type1Font FontName=\"No-name\" PaintType=\"0\" WMode=\"0\" FontType=\"1\">\n<FontInfo version=\"0.1\" Notice=\"Created with Gribouy\" Copyright=\"Created with Gribouy\" FullName=\"No name\" FamilyName=\"No name\" ItalicAngle=\"0.0\" isFixedPitch=\"false\" UnderlinePosition=\"-96\" UnderlineThickness=\"58\"/>\n<FontBBox v1=\"-115\" v2=\"-240\" v3=\"1151\" v4=\"976\"/>\n<FontMatrix v1=\"0.001\" v2=\"0.0\" v3=\"0.0\" v4=\"0.001\" v5=\"0.0\" v6=\"0.0\"/>\n<Private RD=\"string currentfile exch readstring pop\" ND=\"noaccess def\" NP=\"noaccess put\" BlueScale=\"0.039625\" StdHW=\"68\" StdVW=\"74\" password=\"5839\">\n<BlueValues v1=\"-20\" v2=\"0\" v3=\"547\" v4=\"567\" v5=\"739\" v6=\"759\"/>\n<MinFeature v1=\"16\" v2=\"16\"/>\n</Private>\n</Type1Font>";

static GObjectClass *parent_class = NULL;

xmlDocPtr doc;
xmlNodePtr cur;

static GByteArray *read_file (FILE *handle);
static void gribouy_type1_analyse_file (GribouyType1 *font,
					FILE *handle, gboolean pfa);
static void gribouy_type1_class_init (GribouyType1Class *class);
static void gribouy_type1_init (GribouyType1 *font);
static void gribouy_type1_finalize (GObject *object);
static void gribouy_type1_set_property (GObject *object, guint property_id,
		const GValue *value, GParamSpec *pspec);
static void gribouy_type1_get_property (GObject *object, guint property_id,
					GValue *value, GParamSpec *pspec);
static void gribouy_type1_read_drawing_charstring (GribouyType1 *font,
					guchar *charstring, guint size);
static char *append_encoding (char *str1, xmlNodePtr cur);
static char *gribouy_type1_write_dict (GribouyType1 *font);

enum {
	PROP_0,
	PROP_XMLFONT,
	PROP_CURRENT,
	PROP_SYMBOL,
	PROP_BLUE_VALUES,
	PROP_STANDARD_ENCODING,
};

GType
gribouy_type1_get_type ()
{
	static GType gribouy_type1_type = 0;

	if (!gribouy_type1_type) {
		static const GTypeInfo gribouy_type1_info = {
			sizeof (GribouyType1Class),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_type1_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyType1),
			0,
			(GInstanceInitFunc) gribouy_type1_init,
		};

		gribouy_type1_type = g_type_register_static (G_TYPE_OBJECT, "GribouyType1", &gribouy_type1_info, (GTypeFlags)0);
	}
	return gribouy_type1_type;
}

static void
gribouy_type1_class_init (GribouyType1Class *class)
{
	GObjectClass *g_object_class = (GObjectClass *)class;
	GParamSpec *param_spec;

	parent_class = g_type_class_ref (G_TYPE_OBJECT);

	g_object_class->finalize = gribouy_type1_finalize;
	g_object_class->get_property = gribouy_type1_get_property;
	g_object_class->set_property = gribouy_type1_set_property;

	param_spec = g_param_spec_pointer
		("xmlfont",
		 "", "",
		 (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (g_object_class,
					 PROP_XMLFONT,
					 param_spec);

	param_spec = g_param_spec_pointer
		("symbol",
		 "Current program",
		 "Program of the current symbol",
		 (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (g_object_class,
					 PROP_SYMBOL,
					 param_spec);

	param_spec = g_param_spec_pointer
		("blue_values",
		 "", "",
		 (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (g_object_class,
					 PROP_BLUE_VALUES,
					 param_spec);
		 
	param_spec = g_param_spec_string
		("current",
		 "Current symbol",
		 "Current symbol we are editing",
		 "A",
		 (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (g_object_class,
					 PROP_CURRENT,
					 param_spec);
	param_spec = g_param_spec_boolean
		("standard_encoding",
		 NULL, NULL,
		 TRUE,
		 (G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (g_object_class,
					PROP_STANDARD_ENCODING,
					param_spec);
}

static void
gribouy_type1_finalize (GObject *object)
{
	GribouyType1 *font = GRIBOUY_TYPE1 (object);

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize)(object);
	if (font->xmlfont) {
		xmlFreeDoc (font->xmlfont);
		font->xmlfont = NULL;
	}
	if (font->current) {
		g_free (font->current);
		font->current = NULL;
	}
}

static void
gribouy_type1_set_property (GObject *object, guint property_id,
			   const GValue *val, GParamSpec *pspec)
{
	GribouyType1 *font;
	xmlNodePtr cur;
	GArray *blue_values;
	gchar *prop, *txt;
	int i;

	font = GRIBOUY_TYPE1 (object);

	switch (property_id) {
	case PROP_XMLFONT:
		font->xmlfont = g_value_get_pointer (val);
		break;
	case PROP_CURRENT:
		font->current = g_value_get_string (val);
		break;
	case PROP_SYMBOL:
		if (font->current) {
			GList *prog;

			if (g_hash_table_lookup (font->symbols, font->current))
				g_hash_table_remove (font->symbols, font->current);
			prog = g_value_get_pointer (val);
			if (prog)
				g_hash_table_insert (font->symbols,
						font->current, prog);
		}
		break;
	case PROP_BLUE_VALUES:
		cur = xmlDocGetRootElement (font->xmlfont);
		cur = cur->xmlChildrenNode;
		while (xmlStrcmp (cur->name, (const xmlChar *) "Private"))
			cur = cur->next;

		cur = cur->xmlChildrenNode;
		while (cur && xmlStrcmp (cur->name, (const xmlChar *) "BlueValues"))
			cur = cur->next;

		if (blue_values->len)
			blue_values = g_value_get_pointer (val);
		for (i = 0; i < blue_values->len; i ++) {
			xmlSetProp (cur,
				    prop = g_strdup_printf ("v%d", i + 1),
				    txt = g_array_index (blue_values, int, i));
			g_free (prop);
			g_free (txt);
		}
		break;
	case PROP_STANDARD_ENCODING:
		font->standard_encoding = g_value_get_boolean (val);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
gribouy_type1_get_property (GObject *object, guint property_id,
			   GValue *val, GParamSpec *pspec)
{
	GribouyType1 *font;
	GArray *blue_values;
	gchar *prop, *txt;
	xmlNodePtr cur;
	int i, v;

	font = GRIBOUY_TYPE1 (object);

	switch (property_id) {
	case PROP_XMLFONT:
		g_value_set_pointer (val, font->xmlfont);
		break;
	case PROP_SYMBOL:
		g_value_set_pointer (val, g_hash_table_lookup (font->symbols, font->current));
		break;
	case PROP_CURRENT:
		g_value_set_string (val, font->current);
		break;
	case PROP_STANDARD_ENCODING:
		g_value_set_boolean (val, font->standard_encoding);
		break;
	case PROP_BLUE_VALUES:
		blue_values = g_array_new (FALSE, FALSE, sizeof (int));
		cur = xmlDocGetRootElement (font->xmlfont);
		cur = cur->xmlChildrenNode;
		while (xmlStrcmp (cur->name, (const xmlChar *) "Private"))
			cur = cur->next;

		cur = cur->xmlChildrenNode;
		while (cur && xmlStrcmp (cur->name, (const xmlChar *) "BlueValues"))
			cur = cur->next;

		i = 1;
		txt = NULL;
		do {
			if (txt)
				g_free (txt);
			prop = g_strdup_printf ("v%d", i);
			txt = xmlGetProp (cur, prop);
			if (txt) {
				v = atoi (txt);
				blue_values = g_array_append_val (blue_values, v);
			}
			g_free (prop);
			i ++;
		}
		while (txt);
		g_value_set_pointer (val, blue_values);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static GByteArray *
read_file (FILE *handle)
{
	GByteArray *retval;
	guchar *file;
	guint size;

	retval = g_byte_array_new ();
	file = g_new (guchar, 256);

	do {
		size = fread (file, 1, 256, handle);
		retval = g_byte_array_append (retval, file, size);
	}
	while (size == 256);

	g_free (file);
	return retval;
}

static void
gribouy_type1_init (GribouyType1 *font)
{
	font->xmlfont = xmlParseMemory (init_xmlfont, sizeof (init_xmlfont));
	font->current = NULL;
	font->subroutines = g_ptr_array_new ();
	font->symbols = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, gribouy_art_free);
}

static guchar *
skip_space (guchar *buf) {
	while (*buf == ' ' || *buf == '\t' || *buf == '\n' || *buf == '\r') {
		buf ++;
	}
	return buf;
}

static guchar *
decrypt (guchar *buf, gboolean pfa)
{
	guchar *buf1, *buf2, *buf3;
	gushort eexec_R = 55665;
	int nbi, token;
	guint size;

	buf = skip_space (buf);
	if (pfa) {
		buf1 = buf2 = buf;
		buf1 = skip_space (buf1);
		do {
			while (*buf1 == '\n' || *buf1 == '\r')
				buf1 ++;
			
			if (*buf1 >= '0' && *buf1 <= '9')
				nbi = (*buf1 - '0') << 4;
			else if (*buf1 >= 'a' && *buf1 <= 'f')
				nbi = (*buf1 - 'a' + 10) << 4;
			else if (*buf1 >= 'A' && *buf1 <= 'F')
				nbi = (*buf1 -'A' + 10) << 4;
			else
				break;
			buf1 ++;

			if (*buf1 >= '0' && *buf1 <= '9')
				nbi |= (*buf1 - '0');
			else if (*buf1 >= 'a' && *buf1 <= 'f')
				nbi |= (*buf1 - 'a' + 10);
			else if (*buf1 >= 'A' && *buf1 <= 'F')
				nbi |= (*buf1 -'A' + 10);
			else
				break; /* should not be used */
			buf1 ++;

			*buf2 = (nbi ^ (eexec_R >> 8));
			buf2 ++;
			eexec_R = (nbi + eexec_R) * 52845 + 22719;
		}
		while (1);
		buf += 4;
	}
	else {
		buf3 = buf1 = buf;
		while (*buf == (guchar) '\200') {
			buf ++;

			token = *buf ++;
			size = buf[0]
				| (buf[1] << 8)
				| (buf[2] << 16)
				| (buf[3] << 24);
			buf += 4;

			if (token == 1) {
				memmove (buf1, buf, size);
				buf1 += size;
				buf += size;
			}
			else if (token == 2) {
				buf2 = buf;
				buf += size;
				for ( ; buf2 != buf; buf2 ++) {
					nbi = *buf2;
					*buf1 = (nbi ^ (eexec_R >> 8));
					eexec_R = (nbi + eexec_R) * 52845 + 22719;
					buf1 ++;
				}
			}
			else
				break;
		}
		buf = buf3 + 8;
	}
	return buf;
}

static guchar *
read_double ()
{
	guchar *buf1;

	buf = skip_space (buf);
	buf1 = buf;
	if (*buf == '-')
		buf ++;
	while (*buf >= '0' && *buf <= '9')
		buf ++;
	if (*buf == '.')
		buf ++;
	while (*buf >= '0' && *buf <= '9')
		buf ++;
	return g_strndup (buf1, buf - buf1);
}

static guchar *
read_integer ()
{
	guchar *buf1;

	buf = skip_space (buf);
	buf1 = buf;
	if (*buf == '-')
		buf ++;
	while (*buf >= '0' && *buf <= '9')
		buf ++;
	return g_strndup (buf1, buf - buf1);
}

static gint
read_integer_int ()
{
	gint retval;
	gint sign;

	buf = skip_space (buf);
	if (*buf == '-') {
		sign = -1;
		buf ++;
	}
	else
		sign = 1;
	retval = 0;
	while (*buf >= '0' && *buf <= '9') {
		retval = retval*10 + *buf - '0';
		buf ++;
	}
	if (sign < 0)
		retval = -retval;
	return retval;
}

static guchar *
read_boolean ()
{
	buf = skip_space (buf);

	if (!strncmp (buf, "false", 5)) {
		buf += 5;
		return g_strdup ("false");
	}
	else if (!strncmp (buf, "true", 4)) {
		buf += 4;
		return g_strdup ("true");
	}
	g_warning ("not a boolean\n");
	return NULL;
}

static guchar *
read_string ()
{
	guchar *buf1;
	guchar op, cl;
	guint parenth;

	buf = skip_space (buf);

	buf1 = buf;
	if (*buf1 == '/') {
		do
			buf++;
		while (*buf != ' ' && *buf != '\t' && *buf != '\n' && *buf != '\r');
	}
	else {
		switch (*buf1) {
		case '(':
			op = '(';
			cl = ')';
			break;
		case '[':
			op = '[';
			cl = ']';
			break;
		case '{':
			op = '{';
			cl = '}';
			break;
		default:
			g_warning ("Strange character %c to open a string\n", *buf1);
			return NULL;
		}

		parenth = 1;
		while (parenth) {
			do 
				buf ++;
			while (*buf != op && *buf != cl);
			if (*buf == op)
				parenth ++;
			else
				parenth --;
		}
	}
	buf ++;
	return g_strndup (buf1 + 1, buf - buf1 - 2);
}

static void
read_fontinfo_dict ()
{
	gchar *txt;

	cur = xmlNewChild (cur, NULL, "FontInfo", NULL);
	do {
		if (!strncmp ("/version", buf, 8)) {
			buf += 8;
			xmlNewProp (cur, "version", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/Notice", buf, 7)) {
			buf += 7;
			xmlNewProp (cur, "Notice", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/Copyright", buf, 10)) {
			buf += 10;
			xmlNewProp (cur, "Copyright", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/FullName", buf, 9)) {
			buf += 9;
			xmlNewProp (cur, "FullName", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/FamilyName", buf, 11)) {
			buf += 11;
			xmlNewProp (cur, "FamilyName", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/Weight", buf, 7)) {
			buf += 7;
			xmlNewProp (cur, "Weight", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/ItalicAngle", buf, 12)) {
			buf += 12;
			xmlNewProp (cur, "ItalicAngle", txt = read_double ());
			g_free (txt);
		}
		else if (!strncmp ("/isFixedPitch", buf, 13)) {
			buf += 13;
			xmlNewProp (cur, "isFixedPitch", txt = read_boolean ());
			g_free (txt);
		}
		else if (!strncmp ("/UnderlinePosition", buf, 18)) {
			buf += 18;
			xmlNewProp (cur, "UnderlinePosition", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/UnderlineThickness", buf, 19)) {
			buf += 19;
			xmlNewProp (cur, "UnderlineThickness", txt = read_integer ());
			g_free (txt);
		}
		else
			buf++;
	}
	while (strncmp ("end readonly def", buf, 16));
	buf += 16;
	cur = cur->parent;
}

static void
gribouy_type1_read_main_dict (GribouyType1 *font)
{
	gchar *txt, *nb;

	cur = font->xmlfont->children;
	do {
		if (!strncmp ("/FontInfo", buf, 9)) {
			buf += 9;
			while (strncmp ("begin", buf, 5))
				buf ++;
			buf += 5;
			read_fontinfo_dict ();
		}
		else if (!strncmp ("/FontName", buf, 9)) {
			buf += 9;
			xmlNewProp (cur, "FontName", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/PaintType", buf, 10)) {
			buf += 10;
			xmlNewProp (cur, "PaintType", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/WMode", buf, 6)) {
			buf += 6;
			xmlNewProp (cur, "WMode", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/FontBBox {", buf, 11)) {
			buf += 11;
			cur = xmlNewChild (cur, NULL, "FontBBox", NULL);
			xmlNewProp (cur, "v1", txt = read_integer ());
			g_free (txt);
			xmlNewProp (cur, "v2", txt = read_integer ());
			g_free (txt);
			xmlNewProp (cur, "v3", txt = read_integer ());
			g_free (txt);
			xmlNewProp (cur, "v4", txt = read_integer ());
			g_free (txt);
			cur = cur->parent;
		}
		else if (!strncmp ("/FontType", buf, 9)) {
			buf += 9;
			xmlNewProp (cur, "FontType", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/FontMatrix [", buf, 13)) {
			buf += 13;
			cur = xmlNewChild (cur, NULL, "FontMatrix", NULL);
			xmlNewProp (cur, "v1", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v2", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v3", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v4", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v5", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v6", txt = read_double ());
			g_free (txt);
			cur = cur->parent;
		}
		else if (!strncmp ("/Encoding", buf, 9)) {
			buf += 9;
			buf = skip_space (buf);
			if (!strncmp ("StandardEncoding", buf, 16)) {
				buf += 16;
				font->standard_encoding = TRUE;
//				xmlNewProp (cur, "Encoding", "StandardEncoding");
			}
			else if (!strncmp ("256 array", buf, 9)) {
				font->standard_encoding = FALSE;
//				xmlNewProp (cur, "Encoding", "NonStandard");
				cur = xmlNewChild (cur, NULL, "Characters", NULL);
				while (strncmp ("for", buf, 3))
					buf ++;
				buf += 3;
				buf = skip_space (buf);
				while (!strncmp ("dup ", buf, 4)) {
					buf += 4;
					nb = read_integer ();
					buf = skip_space (buf);
					txt = read_string ();
					xmlNewProp (cur, txt, nb);
					g_print ("code :%s charstring :%s\n", nb, txt);
					g_free (txt);
					g_free (nb);
					while (strncmp ("put", buf, 3))
						buf ++;
					buf += 3;
					buf = skip_space (buf);
				}
				while (strncmp ("def", buf, 3))
					buf ++;
				buf += 3;
				cur = cur->parent;
			}
			else
				g_warning ("Strange encoding !!!\n");
		}
		else if (!strncmp ("/UniqueID", buf, 9)) {
			buf += 9;
			xmlNewProp (cur, "UniqueID", txt = read_integer ());
			g_free (txt);
		}
		else
			buf ++;
	}
	while (strncmp ("currentdict end", buf, 15));
	buf += 15;
}

static void
gribouy_type1_read_drawing_charstring (GribouyType1 *font, guchar *charstring, guint size)
{
	gint *nbi;
	guint i;
	GList *lst = NULL;
	GribouyArt *art;
	guchar *endbuf = buf + (size - 4);

	art = g_new0 (GribouyArt, 1);
	nbi = art->nb;
	i = 0;
	do {
		if (*buf >= 32 && *buf <= 246) {
			g_return_if_fail (i < 6);
			nbi[i++] = *buf - 139;
			buf ++;
		}
		else if (*buf >= 247 && *buf <= 250) {
			g_return_if_fail (i < 6);
			nbi[i++] =  ((*buf - 247) << 8) + *(buf + 1) + 108;
			buf += 2;
		}
		else if (*buf >= 251 && *buf <= 254) {
			g_return_if_fail (i < 6);
			nbi[i++] = - ((*buf - 251) << 8) - *(buf + 1) - 108;
			buf += 2;
		}
		else if (*buf == 255) {
			g_return_if_fail (i < 6);
			nbi[i++] = *(buf + 1) | (*(buf + 2) << 8)
				| (*(buf + 3) << 16)
				| (*(buf + 4) << 24);
			buf += 5;
		}
		else {
			switch (*buf++) {
			case 1:
				art->code = HSTEM;
				break;
			case 3:
				art->code = VSTEM;
				break;
			case 4:
				art->code = VMOVETO;
				break;
			case 5:
				art->code = RLINETO;
				break;
			case 6:
				art->code = HLINETO;
				break;
			case 7:
				art->code = VLINETO;
				break;
			case 8:
				art->code = RRCURVETO;
				break;
			case 9:
				art->code = CLOSEPATH;
				break;
			case 10:
				art->code = CALLSUBR;
				break;
			case 11:
				art->code = RETURN;
				break;
			case 12:
				switch (*buf++) {
				case 1:
					art->code = VSTEM3;
					break;
				case 2:
					art->code = HSTEM3;
					break;
				case 7:
					art->code = SBW;
					break;
				case 16:
					art->code = CALLOTHERSUBR;
					break;
				case 17:
					art->code = POP;
					break;
				case 33:
					art->code = SETCURRENTPOINT;
					break;
				default:
					g_warning ("Not yet implemented...\n");
					exit (1);
				}
				break;
			case 13:
				art->code = HSBW;
				break;
			case 14:
				art->code = ENDCHAR;
				break;
			case 21:
				art->code = RMOVETO;
				break;
			case 22:
				art->code = HMOVETO;
				break;
			case 30:
				art->code = VHCURVETO;
				break;
			case 31:
				art->code = HVCURVETO;
				break;
			default:
				g_warning ("Not yet implemented...\n");
				exit (1);
			}
			lst = g_list_append (lst, art);
			art = g_new0 (GribouyArt, 1);
			nbi = art->nb;
			i = 0;
		}
	}
	while (buf != endbuf);
	g_free (art);
	buf += 2;
	g_hash_table_insert (font->symbols, charstring, lst);
}

static void
gribouy_type1_read_drawing_subr (GribouyType1 *font, guint index)
{
	gint *nbi;
	guint i;
	GList *lst = NULL;
	GribouyArt *art;

	art = g_new0 (GribouyArt, 1);
	nbi = art->nb;
	i = 0;
	do {
		if (*buf >= 32 && *buf <= 246) {
			g_return_if_fail (i < 6);
			nbi[i++] = *buf - 139;
			buf ++;
		}
		else if (*buf >= 247 && *buf <= 250) {
			g_return_if_fail (i < 6);
			nbi[i++] =  ((*buf - 247) << 8) + *(buf + 1) + 108;
			buf += 2;
		}
		else if (*buf >= 251 && *buf <= 254) {
			g_return_if_fail (i < 6);
			nbi[i++] = - ((*buf - 251) << 8) - *(buf + 1) - 108;
			buf += 2;
		}
		else if (*buf == 255) {
			g_return_if_fail (i < 6);
			nbi[i++] = *(buf + 1) | (*(buf + 2) << 8)
				| (*(buf + 3) << 16)
				| (*(buf + 4) << 24);
			buf += 5;
		}
		else {
			switch (*buf++) {
			case 1:
				art->code = HSTEM;
				break;
			case 3:
				art->code = VSTEM;
				break;
			case 4:
				art->code = VMOVETO;
				break;
			case 5:
				art->code = RLINETO;
				break;
			case 6:
				art->code = HLINETO;
				break;
			case 7:
				art->code = VLINETO;
				break;
			case 8:
				art->code = RRCURVETO;
				break;
			case 9:
				art->code = CLOSEPATH;
				break;
			case 10:
				art->code = CALLSUBR;
				break;
			case 11:
				art->code = RETURN;
				break;
			case 12:
				switch (*buf++) {
				case 1:
					art->code = VSTEM3;
					break;
				case 2:
					art->code = HSTEM3;
					break;
				case 7:
					art->code = SBW;
					break;
				case 16:
					art->code = CALLOTHERSUBR;
					break;
				case 17:
					art->code = POP;
					break;
				case 33:
					art->code = SETCURRENTPOINT;
					break;
				default:
					g_warning ("Not yet implemented...\n");
				}
				break;
			case 13:
				art->code = HSBW;
				break;
			case 14:
				art->code = ENDCHAR;
				break;
			case 21:
				art->code = RMOVETO;
				break;
			case 22:
				art->code = HMOVETO;
				break;
			case 30:
				art->code = VHCURVETO;
				break;
			case 31:
				art->code = HVCURVETO;
				break;
			default:
				g_warning ("Not yet implemented...\n");
			}
			lst = g_list_append (lst, art);
			art = g_new0 (GribouyArt, 1);
			nbi = art->nb;
			i = 0;
		}
	}
	while (strncmp ("NP", buf, 2));
	g_free (art);
	buf += 2;
	if (index < font->subroutines->len)
		g_ptr_array_index (font->subroutines, index) = lst;
	else if (index == font->subroutines->len)
		g_ptr_array_add (font->subroutines, lst);
	else {
		g_warning ("Fix me : Problem...");
		exit (1);
	}
}

static void
gribouy_type1_read_charstrings (GribouyType1 *font)
{
	guchar *charstring;
	guint size;
	gushort eexec_R;
	gint nbi, i;

	cur = xmlNewChild (cur, NULL, "CharStrings", NULL);
	do {
		charstring = read_string ();
		xmlNewProp (cur, charstring, NULL);
		size = read_integer_int ();
		while (strncmp ("RD", buf, 2))
			buf ++;
		buf += 3;
		eexec_R = 4330;
		for (i = 0; i < size; i++) {
			nbi = buf[i];
			buf[i] = (nbi ^ (eexec_R >> 8));
			eexec_R = (nbi + eexec_R) * 52845 + 22719;
		}
		buf += 4;
		gribouy_type1_read_drawing_charstring (font, charstring, size);
		buf = skip_space (buf);
	}
	while (strncmp ("end", buf, 3));
	buf += 3;
	cur = cur->parent;
}

static void
gribouy_type1_read_subrs (GribouyType1 *font)
{
	guint index, size, i;
	gint nbi;
	gushort eexec_R;

	do {
		buf = skip_space (buf);
		if (!strncmp ("dup", buf, 3)) {
			buf += 4;
			index = read_integer_int ();
			size = read_integer_int ();
			while (strncmp ("RD", buf, 2))
				buf ++;
			buf += 3;
			eexec_R = 4330;
			for (i = 0; i < size; i++) {
				nbi = buf[i];
				buf[i] = (nbi ^ (eexec_R >> 8));
				eexec_R = (nbi + eexec_R) * 52845 + 22719;
			}
			buf += 4;
			gribouy_type1_read_drawing_subr (font, index);
		}
	}
	while (strncmp ("ND", buf, 2));
	buf += 2;
}

/* static void */
/* read_private_dict () */
/* { */
/* 	gchar *txt; */
/* 	cur = xmlNewChild (cur, NULL, "Private", NULL); */
/* 	do { */
/* 		if (!strncmp ("/RD", buf, 3)) { */
/* 			buf += 3; */
/* 			xmlNewProp (cur, "RD", txt = read_string ()); */
/* 			g_free (txt); */
/* 		} */
/* 		else if (!strncmp ("/ND", buf, 3)) { */
/* 			buf += 3; */
/* 			xmlNewProp (cur, "ND", txt = read_string ()); */
/* 			g_free (txt); */
/* 		} */
/* 		else if (!strncmp ("/NP", buf, 3)) { */
/* 			buf += 3; */
/* 			xmlNewProp (cur, "NP", txt = read_string ()); */
/* 			g_free (txt); */
/* 		} */
/* 		else if (!strncmp ("/UniqueID", buf, 9)) { */
/* 			buf += 9; */
/* 			xmlNewProp (cur, "UniqueID", txt = read_integer ()); */
/* 			g_free (txt); */
/* 		} */
/* 		else if (!strncmp ("/BlueValues [", buf, 13)) { */
/* 			buf += 13; */
/* 			cur = xmlNewChild (cur, NULL, "BlueValues", NULL); */
/* 			xmlNewProp (cur, "v1", txt = read_double ()); */
/* 			g_free (txt); */
/* 			xmlNewProp (cur, "v2", txt = read_double ()); */
/* 			g_free (txt); */
/* 			xmlNewProp (cur, "v3", txt = read_double ()); */
/* 			g_free (txt); */
/* 			xmlNewProp (cur, "v4", txt = read_double ()); */
/* 			g_free (txt); */
/* 			xmlNewProp (cur, "v5", txt = read_double ()); */
/* 			g_free (txt); */
/* 			xmlNewProp (cur, "v6", txt = read_double ()); */
/* 			g_free (txt); */
/* 			cur = cur->parent; */
/* 		} */
/* 		else */
/* 			buf ++; */
/* 	} */
/* 	while (strncmp ("noaccess def", buf, 12)); */
/* 	buf += 12; */
/* 	cur = cur->parent; */
/* } */

static void
gribouy_type1_read_crypt (GribouyType1 *font, gboolean pfa)
{
	guint size;
	gchar *txt;

	buf = decrypt (buf, pfa);
	do {
		if (!strncmp ("/Private", buf, 8)) {
			buf += 8;
			while (strncmp ("begin", buf, 5))
				buf ++;
			buf += 5;
			cur = xmlNewChild (cur, NULL, "Private", NULL);
		}
		else if (!strncmp ("/RD", buf, 3)) {
			buf += 3;
			xmlNewProp (cur, "RD", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/ND", buf, 3)) {
			buf += 3;
			xmlNewProp (cur, "ND", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/NP", buf, 3)) {
			buf += 3;
			xmlNewProp (cur, "NP", txt = read_string ());
			g_free (txt);
		}
		else if (!strncmp ("/UniqueID", buf, 9)) {
			buf += 9;
			xmlNewProp (cur, "UniqueID", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/password", buf, 9)) {
			buf += 9;
			xmlNewProp (cur, "password", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/lenIV", buf, 6)) {
			buf += 6;
			xmlNewProp (cur, "lenIV", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/BlueScale", buf, 10)) {
			buf += 10;
			xmlNewProp (cur, "BlueScale", txt = read_double ());
			g_free (txt);
		}
		else if (!strncmp ("/StdHW [", buf, 8)) {
			buf += 8;
			xmlNewProp (cur, "StdHW", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/StdVW [", buf, 8)) {
			buf += 8;
			xmlNewProp (cur, "StdVW", txt = read_integer ());
			g_free (txt);
		}
		else if (!strncmp ("/BlueValues [", buf, 13)) {
			buf += 13;
			cur = xmlNewChild (cur, NULL, "BlueValues", NULL);
			xmlNewProp (cur, "v1", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v2", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v3", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v4", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v5", txt = read_double ());
			g_free (txt);
			xmlNewProp (cur, "v6", txt = read_double ());
			g_free (txt);
			cur = cur->parent;
		}
		else if (!strncmp ("/MinFeature {", buf, 13)) {
			buf += 13;
			cur = xmlNewChild (cur, NULL, "MinFeature", NULL);
			xmlNewProp (cur, "v1", txt = read_integer ());
			g_free (txt);
			xmlNewProp (cur, "v2", txt = read_integer ());
			g_free (txt);
			cur = cur->parent;
		}
		else if (!strncmp ("/Subrs", buf, 6)) {
			buf += 6;
			size = read_integer_int ();
			while (strncmp ("array", buf, 5))
				buf ++;
			buf += 5;
			font->subroutines = g_ptr_array_sized_new (size);
			gribouy_type1_read_subrs (font);
		}
		else if (!strncmp ("/CharString", buf, 11)) {
			buf += 11;
			while (strncmp ("begin", buf, 5))
				buf ++;
			buf += 5;
			font->symbols = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, gribouy_art_free);
			gribouy_type1_read_charstrings (font);
		}
		else
			buf ++;
	}
	while (strncmp ("mark currentfile closefile", buf, 26));
	buf += 26;
	cur = cur->parent;
}

static void
gribouy_type1_analyse_file (GribouyType1 *font, FILE *handle, gboolean pfa)
{
	guchar *buf1, *buf2;
	gushort eexec_R = 55665;
	gint nbi, token;
	guint size, i;
	GByteArray *tmp;

	tmp = read_file (handle); /* Fill of buffer and buffer_size */
	buffer = tmp->data;
	buffer_size = tmp->len;
	/* After reading the file, decompression... */
	buf = buffer;
	if (!pfa)
		buf += 6;

	while (*buf == '%') { /* We skip the comments */
		do
			buf ++;
		while (*buf != '\n' && *buf != '\r');
		buf ++;
		buf = skip_space (buf);
	}
	do {
		if (!strncmp ("dict begin", buf, 10)) {
			/* Main dictionary */
			buf += 10;
			gribouy_type1_read_main_dict (font);
		}
		else if (!strncmp ("currentfile eexec", buf, 17)) {
			buf += 17;
			gribouy_type1_read_crypt (font, pfa);
		}			
		else if (!strncmp ("cleartomark", buf, 11)) {
			return;
		}
		else
			buf ++;
	}
	while (TRUE);
}

static char *
append_prop1 (char *str1, xmlNodePtr cur, guint *nb, const char *prop, int n, int delim, int end)
{
	gint i;
	char *str, *v[6], *fin;
	char c1, c2;

	cur = cur->xmlChildrenNode;
	while (cur && xmlStrcmp (cur->name, (const xmlChar *) prop))
		cur = cur->next;
	if (!cur)
		return str1;

	switch (end) {
	case 0:
		fin = g_strdup ("");
		break;
	case 1:
		fin = g_strdup ("readonly");
		break;
	}

	switch (delim) {
	case 3:
		c1 = '{';
		c2 = '}';
		break;
	case 4:
		c1 = '[';
		c2 = ']';
		break;
	}
	switch (n) {
	case 2:
		str = g_strdup_printf ("%s/%s %c%s %s%c %s def\n",
				       str1,
				       prop,
				       c1,
				       v[0] = xmlGetProp (cur, "v1"),
				       v[1] = xmlGetProp (cur, "v2"),
				       c2, fin);
		break;
	case 4:
		str = g_strdup_printf ("%s/%s %c%s %s %s %s%c %s def\n",
				       str1,
				       prop,
				       c1,
				       v[0] = xmlGetProp (cur, "v1"),
				       v[1] = xmlGetProp (cur, "v2"),
				       v[2] = xmlGetProp (cur, "v3"),
				       v[3] = xmlGetProp (cur, "v4"),
				       c2, fin);
		break;
	case 6:
		str = g_strdup_printf ("%s/%s %c%s %s %s %s %s %s%c %s def\n",
				       str1,
				       prop,
				       c1,
				       v[0] = xmlGetProp (cur, "v1"),
				       v[1] = xmlGetProp (cur, "v2"),
				       v[2] = xmlGetProp (cur, "v3"),
				       v[3] = xmlGetProp (cur, "v4"),
				       v[4] = xmlGetProp (cur, "v5"),
				       v[5] = xmlGetProp (cur, "v6"),
				       c2, fin);
		break;
	}

	g_free (str1);
	g_free (fin);
	for (i = 0; i < n; i++)
		g_free (v[i]);

	cur = cur->parent;
	*nb = *nb + 1;
	return str;
}

static char *
append_encoding (char *str1, xmlNodePtr cur)
{
	char *str, *str2, *nb;
	xmlAttrPtr prop;

	cur = cur->xmlChildrenNode;
	while (cur && xmlStrcmp (cur->name, (const xmlChar *) "Characters"))
		cur = cur->next;
	if (!cur)
		return str1;

	str = g_strdup_printf ("/Encoding 256 array\n"
				"0 1 255 {1 index exch /.notdef put} for\n");
	prop = cur->properties;
	while (prop != NULL) {
		nb = xmlNodeListGetString(cur->doc, prop->children, 1);
		str2 = g_strdup_printf ("%sdup %s /%s put\n", str, nb, prop->name);
		g_free (str);
		g_free (nb);
		str = str2;
		prop = prop->next;
	}
	str2 = g_strdup_printf ("%s%sreadonly def\n", str1, str);
	g_free (str);
	g_free (str1);
	return str2;
}

static char *
append_prop (char *str1, xmlNodePtr cur, guint *nb, const char *prop, int delim, int end)
{
	char *not, *fin;
	char *str2, *str;

	switch (end) {
	case 0:
		fin = g_strdup ("");
		break;
	case 1:
		fin = g_strdup ("readonly");
		break;
	case 2:
		fin = g_strdup ("executeonly");
		break;
	}
	if ((not = xmlGetProp (cur, prop))) {
		switch (delim) {
		case 0:
			str2 = g_strdup_printf ("/%s %s %s def\n", prop, not, fin);
			break;
		case 1:
			str2 = g_strdup_printf ("/%s (%s) %s def\n", prop, not, fin);
			break;
		case 2:
			str2 = g_strdup_printf ("/%s /%s %s def\n", prop, not, fin);
			break;
		case 3:
			str2 = g_strdup_printf ("/%s {%s} %s def\n", prop, not, fin);
			break;
		}
		str = g_strconcat (str1, str2, NULL);
		g_free (not);
		g_free (fin);
		g_free (str1);
		g_free (str2);
		*nb += 1;
		return str;
	}
	g_free (not);
	g_free (fin);
	g_warning ("Nothing about %s\n", prop);
	return str1;
}

static char *
gribouy_type1_write_dict (GribouyType1 *font)
{
	xmlDocPtr xmlfont;
	char *retval, *str1, *str2, *str, *enc;
	xmlNodePtr cur;
	char *version, *fontname, *v1, *v2, *v3, *v4;
	guint nb;

	xmlfont = font->xmlfont;
	cur = xmlDocGetRootElement (xmlfont);
	fontname = xmlGetProp (cur, "FontName");
	cur = cur->xmlChildrenNode;
	while (xmlStrcmp (cur->name, (const xmlChar *) "FontInfo"))
		cur = cur->next;

	version = xmlGetProp (cur, "version");

	str = g_strdup_printf ("/version (%s) readonly def\n", version);
	nb = 1;
	str = append_prop (str, cur, &nb, "Notice", 1, 1);
	str = append_prop (str, cur, &nb, "Copyright", 1, 1);
	str = append_prop (str, cur, &nb, "FullName", 1, 1);
	str = append_prop (str, cur, &nb, "FamilyName", 1, 1);
	str = append_prop (str, cur, &nb, "Weight", 1, 1);
	str = append_prop (str, cur, &nb, "ItalicAngle", 0, 1);
	str = append_prop (str, cur, &nb, "isFixedPitch", 0, 1);
	str = append_prop (str, cur, &nb, "UnderlinePosition", 0, 1);
	str = append_prop (str, cur, &nb, "UnderlineThickness", 0, 1);

	str1 = g_strdup_printf ("/FontInfo %d dict dup begin\n%send readonly def\n", nb, str);
	g_free (str);

	cur = cur->parent;
	nb = 1;
	str = g_strdup ("");
	str = append_prop (str, cur, &nb, "FontName", 2, 1);
	str = append_prop (str, cur, &nb, "PaintType", 0, 1);
	str = append_prop (str, cur, &nb, "WMode", 0, 1);
	str = append_prop1 (str, cur, &nb, "FontBBox", 4, 3, 1);
	str = append_prop (str, cur, &nb, "FontType", 0, 1);
	str = append_prop1 (str, cur, &nb, "FontMatrix", 6, 4, 1);
	if (font->standard_encoding) {
		str2 = str;
		str = g_strdup_printf ("%s/Encoding StandardEncoding readonly def\n",
				str2);
	}
	else
		str = append_encoding (str, cur);
	
/*	enc = xmlGetProp (cur, "Encoding");
	if (!strcmp ("StandardEncoding", enc)) {
		g_free (enc);
		str = append_prop (str, cur, &nb, "Encoding", 0, 1);
	}
	else {
		g_free (enc);
		str = append_encoding (str, cur);
	}*/
	str = append_prop (str, cur, &nb, "UniqueID", 0, 1);

	retval = g_strdup_printf ("%%!FontType1-1.0: %s %s\n"
				  "%% Generated by Gribouy\n"
				  "%d dict begin\n%s%scurrentdict end\n"
				  "currentfile eexec\n",
				  fontname,
				  version,
				  nb,
				  str1,
				  str);
	g_free (fontname);
	g_free (version);
	g_free (str1);
	g_free (str);

	return retval;
}

static GByteArray *
append_val (GByteArray *routines, int nb)
{
	GByteArray *retval;
	char byte;

	retval = routines;
	if (nb >= -107 && nb <= 107) {
		byte = nb + 139;
		retval = (GByteArray *) (g_array_append_val ((GArray *)retval, byte));
	}
	else if (nb >= 108 && nb <= 1131) {
		byte = ((nb - 108) >> 8) + 247;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = (nb - 108) & 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
	}
	else if (nb >= -1131 && nb <= -108) {
		nb = -nb;
		byte = ((nb - 108) >> 8) + 251;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = (nb - 108) & 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
	}
	else if (nb > 0) {
		byte = 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = nb >> 24;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = (nb & 0xffffff) >> 16;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = (nb & 0xffff) >> 8;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = nb & 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
	}
	else {
		nb = -nb - 1;
		byte = 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = (nb >> 24) ^ 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = ((nb & 0xffffff) >> 16) ^ 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = ((nb & 0xffff) >> 8) ^ 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
		byte = (nb & 255) ^ 255;
		retval = (GByteArray *) (g_array_append_val ((GArray *) retval, byte));
	}
	return retval;
}

static GByteArray *
append_subr (GByteArray *routines, GList *lst)
{
	GList *l;
	GribouyArt *art;
	GByteArray *retval;
	char byte;

	retval = routines;

	for (l = lst; l != NULL; l = g_list_next (l)) {
		art = l->data;
		switch (art->code) {
		case HSTEM:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			byte = 1;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case VSTEM:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			byte = 3;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case VMOVETO:
			retval = append_val (retval, art->nb[0]);
			byte = 4;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case RLINETO:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			byte = 5;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case HLINETO:
			retval = append_val (retval, art->nb[0]);
			byte = 6;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case VLINETO:
			retval = append_val (retval, art->nb[0]);
			byte = 7;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case RRCURVETO:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			retval = append_val (retval, art->nb[2]);
			retval = append_val (retval, art->nb[3]);
			retval = append_val (retval, art->nb[4]);
			retval = append_val (retval, art->nb[5]);
			byte = 8;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case CLOSEPATH:
			byte = 9;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case CALLSUBR:
			retval = append_val (retval, art->nb[0]);
			byte = 10;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case RETURN:
			byte = 11;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case HSBW:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			byte = 13;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;			
		case ENDCHAR:
			byte = 14;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case RMOVETO:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			byte = 21;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case HMOVETO:
			retval = append_val (retval, art->nb[0]);
			byte = 22;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case VHCURVETO:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			retval = append_val (retval, art->nb[2]);
			retval = append_val (retval, art->nb[3]);
			byte = 30;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		case HVCURVETO:
			retval = append_val (retval, art->nb[0]);
			retval = append_val (retval, art->nb[1]);
			retval = append_val (retval, art->nb[2]);
			retval = append_val (retval, art->nb[3]);
			byte = 31;
			retval = (GByteArray *) g_array_append_val ((GArray *) retval, byte);
			break;
		}
	}
	return retval;
}

static GByteArray *
gribouy_encrypt (GByteArray *routines, guint16 r, guint lenIV)
{
	GByteArray *retval;
	char *beg;
	guint16 c1 = 52845;
	guint16 c2 = 22719;
	guchar cipher;
	int i;

	beg = g_new0 (char, (lenIV > 0 ? lenIV:4));
	retval = g_byte_array_prepend (routines, beg, lenIV);
	g_free (beg);

	for (i = 0; i < routines->len; i++) {
		cipher = (g_array_index (retval, char, i) ^ (r >> 8));
		r = (cipher + r) * c1 + c2;
		g_array_index (retval, char, i) = cipher;
	}
	return retval;
}

static void /* GByteArray * */
append_charstring (gpointer key, gpointer value, GByteArray **routines)
{
	GByteArray *rtn;
	GByteArray *retval;
	GByteArray *cs;
	char *str;

	g_return_if_fail (value != NULL);
	rtn = *routines;
	cs = g_byte_array_new ();
	cs = append_subr (cs, (GList *)value);
	cs = gribouy_encrypt (cs, 4330, 4); /* Fix me : lenIV */

	str = g_strdup_printf ("/%s %d RD ", key, cs->len);
	cs = g_byte_array_prepend (cs, str, strlen (str));
	g_free (str);

	str = g_strdup ("ND\n");
	cs = g_byte_array_append (cs, str, strlen (str));
	g_free (str);

	rtn = g_byte_array_append (rtn, cs->data, cs->len);
	g_byte_array_free (cs, TRUE);
	*routines = rtn;
}

static GByteArray *
write_crypt (xmlDocPtr xmlfont, GPtrArray *subrs, GHashTable *symb)
{
	xmlNodePtr cur;
	xmlAttrPtr prop;
	char *str, *str1;
	GByteArray *routines, *retval;
	guint nb;
	int i, lenIV;

	cur = xmlDocGetRootElement (xmlfont)->xmlChildrenNode;
	while (xmlStrcmp (cur->name, (const xmlChar *) "Private"))
		cur = cur->next;

	nb = 0;
	str = g_strdup ("");
	str = append_prop (str, cur, &nb, "RD", 3, 2);
	str = append_prop (str, cur, &nb, "ND", 3, 2);
	str = append_prop (str, cur, &nb, "NP", 3, 2);
	str = append_prop1 (str, cur, &nb, "MinFeature", 2, 3, 0);
	str = append_prop (str, cur, &nb, "password", 0, 0);
	str = append_prop1 (str, cur, &nb, "BlueValues", 6, 4, 0);
	str = append_prop (str, cur, &nb, "UniqueID", 0, 0);
	str = append_prop (str, cur, &nb, "lenIV", 0, 0);
	str1 = g_strdup_printf ("dup /Private %d dict dup begin\n%s",
				nb,
				str);
	g_free (str);

	str = xmlGetProp (cur, "lenIV");
	if (str == NULL)
		lenIV = 4;
	else
		lenIV = atoi (str);

	g_free (str);

	str = str1;

	retval = g_byte_array_new ();

	for (i = 0; i < subrs->len; i++) {
		routines = g_byte_array_new ();
		routines = append_subr (routines, g_ptr_array_index (subrs, i));

		routines = gribouy_encrypt (routines, 4330, lenIV);

		str1 = g_strdup_printf ("dup %d %d RD ", i, routines->len);
		routines = g_byte_array_prepend (routines, str1, strlen (str1));
		g_free (str1);

		str1 = g_strdup ("NP\n");
		routines = g_byte_array_append (routines, str1, strlen (str1));
		g_free (str1);

		retval = g_byte_array_append (retval, routines->data, routines->len);
		g_byte_array_free (routines, TRUE);
	}

	str1 = g_strdup_printf ("/Subrs %d array\n", subrs->len);
	retval = g_byte_array_prepend (retval, str1, strlen (str1));
	g_free (str1);

	str1 = g_strdup ("ND\n");
	retval = g_byte_array_append (retval, str1, strlen (str1));
	g_free (str1);

	routines = g_byte_array_new ();

	g_hash_table_foreach (symb, append_charstring, &routines);

	str1 = g_strdup_printf ("2 index /CharStrings %d dict dup begin\n",
				g_hash_table_size (symb));
	routines = g_byte_array_prepend (routines, str1, strlen (str1));
	g_free (str1);

	str1 = g_strdup ("end\nend\nreadonly put\n"
			 "noaccess put\n"
			 "dup /FontName get exch definefont pop\n"
			 "mark currentfile closefile\n");
	routines = g_byte_array_append (routines, str1, strlen (str1));
	g_free (str1);

	retval = g_byte_array_append (retval, routines->data, routines->len);
	g_byte_array_free (routines, TRUE);
	retval = g_byte_array_prepend (retval, str, strlen (str));
	g_free (str);

	retval = gribouy_encrypt (retval, 55665, 4);

	return retval;
}

int
gribouy_type1_save_file (const char *file, GribouyType1 *font)
{
	FILE *handle;
	char *str;
	GByteArray *str1;
	guint len;
	int i;

	handle = fopen (file, "w");
	if (!handle)
		return 0;

	str = gribouy_type1_write_dict (font);
	len = strlen (str);
	fprintf (handle, "%c%c%c%c%c%c%s",
		 '\200', 1,
		 ((char *)(&len))[0], ((char *)(&len))[1],
		 ((char *)(&len))[2], ((char *)(&len))[3],
		 str);
	g_free (str);

	str1 = write_crypt (font->xmlfont, font->subroutines, font->symbols);
	len = str1->len;

	fprintf (handle, "%c%c%c%c%c%c",
		 '\200', 2,
		 ((char *)(&len))[0], ((char *)(&len))[1],
		 ((char *)(&len))[2], ((char *)(&len))[3]);
	for (i = 0; i < str1->len; i++)
		fprintf (handle, "%c", str1->data[i]);

	len = 524;
	fprintf (handle, "%c%c%c%c%c%c",
		 '\200', 1,
		 ((char *)(&len))[0], ((char *)(&len))[1],
		 ((char *)(&len))[2], ((char *)(&len))[3]);

	for (i = 0; i < 32; i++)
		fprintf (handle, "0000000000000000");
	fprintf (handle, "cleartomark\n%c%c\n", '\200', 3);

	fclose (handle);
	return 1;
}

GObject *
gribouy_type1_new ()
{
	GribouyType1 *font;

	font = g_object_new (gribouy_type1_get_type(), NULL);
	font->standard_encoding = TRUE;
	font->current = NULL;
	return (GObject *)font;
}

GObject *
gribouy_type1_new_from_file (const char *filename)
{
	FILE *handle;
	xmlNodePtr cur;
	GribouyType1 *font;
	gboolean pfa;

	if (filename[strlen(filename) - 1] == 'a')
		pfa = TRUE;
	else if (filename[strlen(filename) - 1] == 'b')
		pfa = FALSE;
	else
		return NULL;

	handle = fopen (filename, "r");
	if (!handle)
		return NULL;

	font = GRIBOUY_TYPE1 (g_object_new (gribouy_type1_get_type(), NULL));
	font->xmlfont = xmlNewDoc ("1.0");
	cur = xmlNewDocNode (font->xmlfont, NULL, "Type1Font", NULL);
	xmlDocSetRootElement (font->xmlfont, cur);

	gribouy_type1_analyse_file (font, handle, pfa);
	fclose (handle);

	return (GObject *)font;
}

void
gribouy_type1_remove (GribouyType1 *font, char *symbol)
{
	if (font->standard_encoding)
		g_warning ("Standard encoding !!!\n");
	else {
		if (g_hash_table_lookup (font->symbols, symbol))
			g_hash_table_remove (font->symbols, symbol);
		else
			g_warning ("No symbol %s\n", symbol);
	}
}
